package com.billpringle.utils.wrputils;

import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.KeyEvent;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintStream;
import java.util.List;
import javax.swing.DefaultListModel;
import javax.swing.JDesktopPane;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JList;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.KeyStroke;
import javax.swing.ScrollPaneConstants;
import javax.swing.SwingWorker;

/**
 * This class rebuilds the zip code database from the
 * US Post Office CityState system.
 * This database is accessed by the WrpValidateCsv class,
 * which validates city / state / zipcode combinations.
 * </p><p>
 * The zip code information can be obtained from the U.S. Post Office.
 * The AIS City/State National information CD is received periodically
 * (monthly or quarterly).  The information is received as a large text
 * file with fixed length records and no end-of-line sequence.
 * The file contains valid zip codes, along with the valid cities for
 * that zip code.
 * The format for this file is described <a href="#fmt">below</a>.
 * </p><p>
 * The zip code validation information is stored in a MySQL database
 * table <i>zipcode</i> with the following fields:
 * </p>
 * <ul>
 * <li>id - unique ID generated by the database (indexed)</li>
 * <li>zipcode - 5 character zip code (indexed)</li>
 * <li>state - 2 character USPS state abbreviation</li>
 * <li>city - the name of a city for this zip code (up to 45 characters)</li>
 * </ul>
 * <p>
 * This class reads the USPS zipcode data file and creates a CSV file
 * that can be used to replace the current database content.
 * The user should perform the following steps:
 * </p>
 * <dl>
 * <dt>Copy the contents of the USPS CD into a directory</dt>
 * <dd>
 * This will create (or replace) the sub-directories: <i>crlf</i>,
 * <i>ctystate</i>, and <i>layout</i>.
 * </dd>
 * <dd>
 * The directory you care about is <i>ctystate</i>;
 * the other directories are not needed.
 * The <i>crlf</i> directory contains a MS-DOS program that can
 * insert CR-LF characters at the end of each line.
 * The <i>layout</i> directory contains documentation on the
 * file format, as well as information on the <i>crlf.exe</i> program.
 * </dd>
 * <dt>Unzip the file <i>ctystate.zip</i>
 * <dd>
 * Within <i>ctystate</i> directory should be a file <i>ctystate.zip</i>,
 * which should be unzipped back into the <i>ctystate</i> directory, 
 * which will create the file
 * <i>ctystate.txt</i>.
 * This is a large text file with fixed-length records and no CR-LF
 * characters, so don't try to read it with a text editor.
 * </dd>
 * <dt>Execute the program extract_citystate</dt>
 * <dd>
 * This is a Windows executable program that executes this class.
 * It was built using the program 
 * <a href="https://sourceforge.net/projects/jsmooth/" target="jsmooth">
 * JSmooth</a>, which creates windows executable wrappers for Java programs.
 * </dd>
 * <dd>
 * The program should open up a window with the message "Log Created"
 * </dd>
 * <dt>Select the Input File</dt>
 * <dd>
 * Select "File / Open USPS Input File" and 
 * navigate to the input file ctystate.txt
 * </dd>
 * <dt>Select the Output File</dt>
 * <dd>
 * Select "File / Open Output/Input CSV File" and 
 * navigate to zipcode.csv (if present) in the original directory. 
 * (you will  need to go up one directory from the input file.)
 * </dd>
 * <dd>
 * If this is the first time you have used this program,
 * or if you created a new directory for this conversion,
 * then the file <i>zipcode.csv</i> won't be there.
 * </dd>
 * <dt>Select <i>Convert Zipcode Information</i></dt>
 * <dd>
 * This option will cause the input file that you selected
 * (probably <i>ctystate/ctystate.txt</i>)
 * to be converted into the output CSV file that you selected
 * (probably <i>zipcode.csv</i>.)
 * </dd>
 * <dd>
 * <b>Note:</b>
 * <br />
 * The name of the CSV file must be the same as the table name
 * in the MySQL database in order for the import to work correctly.
 * </dd>
 * <dt>Run the <i>import.bat</i> program</dt>
 * <dd>
 * The <i>import.bat</i> calls the <i>mysqlimport</i> program
 * to import the output CSV file into the database, replacing
 * the current content of the table.
 * </dd>
 * <dd>
 * This program assumes the output file was <i>zipcode.csv</i>
 * so that it can be inserted into the MySQL database table
 * <i>zipcode</i>.
 * If any other names are used,
 * the command file must be updated.
 * A sample version of import.bat is as follows:
 * </dd>
 * <dd>
 * mysqlimport -h127.0.0.1 -czipcode,state,city -uemail -punisys -L --delete --fields-terminated-by=, --lines-terminated-by="\r\n" email zipcode.csv
 * </dd>
 * </dl>
 * <p>
 * After the above steps have been completed,
 * the zipcode database table will have the latest USPS information.
 * </p><p id="fmt">
 * The USPS AIS City/State National file is a text file with fixed-length
 * records rather than the typical text lines terminated with CR-LF type
 * sequences.  Because of this, normal input text routines such as readLine()
 * can't be used.  This program uses the inner worker class ZipFileReader
 * to extract the zip code data in the background.
 * The input file contains several different record types,
 * but the only type that is used is the "D" type record,
 * which contains the detailed zip code information.
 * </p><p>
 * The format of this file is:
 * </p>
 * <table border="1" cellspacing="0">
 * <tr><td colspan="3">Start and End locations are for substring function</td></tr>
 * <tr><th>Start</th><th>End</th><th>Description</th></tr>
 * <tr><td>0</td><td>1</td><td>
 * Record type.  Type "D" is the type of record that contains 
 * the detailed zipcode/city/state information.
 * </td></tr>
 * <tr><td>1</td><td>6</td><td>The 5-digit zip code</td></tr>
 * <tr><td>13</td><td>41</td><td>A city name for this zip code.
 * There can be multiple records for the same zip code,
 * one for each city associated with the zip code.</td></tr>
 * <tr><td>41</td><td>54</td><td>City abbreviation.
 * This field isn't used often, and when it is,
 * the same abbreviation often appears in another detail
 * record under city name.
 * </td></tr>
 * <tr><td>99</td><td>101</td><td>The two character USPS state abbreviation</td></tr>
 * </table>
 * <p>
 * Although there is a field for abbreviation,
 * normally separate records appear for the zip code,
 * with the city name in one record, and the abbreviation in another record.
 * 
 * 
  <table border="0" cellspacing="0" cellpadding="6" align="center"
  summary="Creative Commons license">
    <tr>
      <td><!-- Creative Commons License -->
      <a rel="license" href=
      "http://creativecommons.org/licenses/by-nc-sa/2.0/" target=
      "CC"><img alt="Creative Commons License" border="0" src=
      "http://creativecommons.org/images/public/somerights20.gif" /></a></td>

      <td><a rel="license" href=
      "http://creativecommons.org/licenses/by-nc-sa/2.0/" target=
      "CC"><img alt="Creative Commons License Symbols" border="0"
      src="cc_lic.gif" /></a> <!-- /Creative Commons License --></td>

      <td>Unless noted otherwise, all materials available for
      download from my site are copyrighted by Bill Pringle, and
      are licensed under a <a rel="license" href=
      "http://creativecommons.org/licenses/by-nc-sa/2.0/" target=
      "CC">Creative Commons License</a>.</td>
    </tr>
  </table>
  
 * @author Bill Pringle
 *
 */

public class WrpBuildZipCodes extends JFrame
	implements ActionListener, ComponentListener
{
	/** sequencing identifier */
	final static long serialVersionUID = 1;

	//////////////////////////////
	//
	// Class constants
	//
	//////////////////////////////
	
	//
	// locations of zipcode fields, extracted with substring
	// The starting location is zero based
	//	The ending location is the character after the end of the field
	//
	/** starting string for zipcode detail record */
	protected static final String USPS_DETAIL = "D";
	/** starting location of zipcode */
	protected static final int USPS_ZIP_START = 1;
	/** ending location of zipcode */
	protected static final int USPS_ZIP_END = 6;
	/** starting name of the city name */
	protected static final int USPS_CITYNAME_START = 13;
	/** ending location of the city name */ 
	protected static final int USPS_CITYNAME_END = 41;
	/** starting location for optional city abbreviation */
	protected static final int USPS_CITYABBR_START = 41;
	/** ending location for optional city abbreviation */
	protected static final int USPS_CITYABBR_END = 54;
	/** starting location for state abbreviation */
	protected static final int USPS_STATE_START = 99;
	/** ending location for state abbreviation */
	protected static final int USPS_STATE_END = 101;
	/** the record size for the input USPS text file */
	protected static final int USPS_RECORD_SIZE = 129;

	///////////////////////////////
	//
	// instance variables
	//
	///////////////////////////////
	/** file chooser */
	protected JFileChooser fc = new JFileChooser();
	/** input file from USPS with zipcode information */
	protected File fileIn;
	/** CSV file converted from input file */
	protected File fileCsv;
	/** reader for input file */
	protected BufferedReader rdr;
	/** print stream for output file */
	protected PrintStream prt;

	///////////////////////////////////
	//
	// Instance GUI variables
	//
	///////////////////////////////////
	/** desktop for main window */
	private JDesktopPane desktop = new JDesktopPane();
	/** panel for scrolling log messages */
	private JPanel logPane = null;
	/** list of log messages */
	private JList logList = null;
	/** list model for log */
	private DefaultListModel logModel = null;
	/** scroll pane for log messages */
	private JScrollPane logScroll = null;


	/**
	 * Default (and only) Constructor
	 * 
	 */
	public WrpBuildZipCodes()
	{
		init(); // initialize object
		buildMenu(); // build the program menu
		buildPage(); // build the window
	}
	
	/**
	 * Initialize object before building menu and page.
	 * 
	 */
	public void init()
	{
		setContentPane(desktop); // attach the desktop to the frame
		logPane = new JPanel(); // create panel for log message
		logModel = new DefaultListModel(); // default list model for log messages
		logList = new JList(logModel); // list for log messages
		logScroll = new JScrollPane(logList); // scrolled pane for log messages
		logScroll.setVerticalScrollBarPolicy(ScrollPaneConstants.VERTICAL_SCROLLBAR_ALWAYS);
	}
	
	/**
	 * Build the menu for the window.
	 * 
	 */
	public void buildMenu()
	{
		JMenuBar menuBar = new JMenuBar(); // top menu bar

		// create File menu
		JMenu menu = new JMenu("File"); // file menu
		menu.setMnemonic(KeyEvent.VK_F); // Alt-F
		String s = "Process USPS ZipCode information"; // tip
		menu.setToolTipText(s);
		menu.getAccessibleContext().setAccessibleDescription(s);
		menuBar.add(menu);		// add to File menu

		// Open input USPS text file
		JMenuItem menuItem = new JMenuItem("Open Input USPS File", KeyEvent.VK_I);
		s = "Select file from USPS containing zipcode information";
		menuItem.setToolTipText(s);
		menu.getAccessibleContext().setAccessibleDescription(s);
		menuItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_I,ActionEvent.ALT_MASK));
		menuItem.setActionCommand("input");
		menuItem.addActionListener(this);
		menu.add(menuItem);

		// Open output CSV text file
		menuItem = new JMenuItem("Open Output/Input CSV File", KeyEvent.VK_V);
		s = "Select input CSV file for Convert, or output CSV file for updating the database";
		menuItem.setToolTipText(s);
		menu.getAccessibleContext().setAccessibleDescription(s);
		menuItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_V,ActionEvent.ALT_MASK));
		menuItem.setActionCommand("csv");
		menuItem.addActionListener(this);
		menu.add(menuItem);

		// Convert input file to output file
		menuItem = new JMenuItem("Convert Zipcode Information", KeyEvent.VK_C);
		s = "Convert zipcode information from input text file to output CSV";
		menuItem.setToolTipText(s);
		menu.getAccessibleContext().setAccessibleDescription(s);
		menuItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_C,ActionEvent.ALT_MASK));
		menuItem.setActionCommand("convert");
		menuItem.addActionListener(this);
		menu.add(menuItem);

		menu.addSeparator();

		// clear the scrolled log
		menuItem = new JMenuItem("Clear Log", KeyEvent.VK_E);
		s = "Clear log messages";
		menuItem.setToolTipText(s);
		menu.getAccessibleContext().setAccessibleDescription(s);
		menuItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_E,ActionEvent.ALT_MASK));
		menuItem.setActionCommand("clear");
		menuItem.addActionListener(this);
		menu.add(menuItem);

		// Exit the program
		menuItem = new JMenuItem("Exit", KeyEvent.VK_X);
		s = "Exit Program";
		menuItem.setToolTipText(s);
		menuItem.getAccessibleContext().setAccessibleDescription(s);
		menuItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_X, ActionEvent.ALT_MASK));
		menuItem.setActionCommand("exit");
		menuItem.addActionListener(this);
		menu.add(menuItem);

		// add menu bar to window
		desktop.getRootPane().setJMenuBar(menuBar); // attach the menu bar
	}

	/**
	 * Build the page contents for the window.
	 * The page for this program consists basically of a scrolled list
	 * containing log messages.
	 * 
	 */
	public void buildPage()
	{
		setTitle("Process USPS ZipCode database"); // set window title
		setSize(600, 400); // set window size
		setDefaultCloseOperation(EXIT_ON_CLOSE); // terminate when window closed
		buildLog(); // initialize the log
		addComponentListener(this); // handle actions
		repaint(); // draw the window
	}

	/**
	 * Build a scrolled pane for log messages.
	 * This method builds a scrolled list for log messages.
	 * 
	 */
	public void buildLog()
	{
		Dimension wdim = this.getSize(); // window size
		Dimension sdim = new Dimension(wdim); // log will be inside the window
		sdim.height -= 70; // log will be a scroll at bottom of screen
		sdim.width -= 10; // width of page
		// logScroll.setHorizontalScrollBarPolicy(ScrollPaneConstants.HORIZONTAL_SCROLLBAR_ALWAYS);
		logPane.add(logScroll); // add the scrolled pane
		logPane.setSize(sdim); // size to fit inside the outer window
		logScroll.setPreferredSize(sdim); // also set preferred size
		// int loc = wdim.height - sdim.height - 20;
		this.add(logPane); // add scrolled pane to window
		logPane.setLocation(0, 0); // position to upper left corner
		addLog("Log created"); // add log message
		logPane.setVisible(true); // make log messages visible
	}

	/**
	 * Add a message to the log.
	 * This method will add the specified message string
	 * to the scrolled log messages.
	 * It will also attempt to reposition the log scroll
	 * so that the most recently added messages are visible.
	 * 
	 * @param msg message to be added to the scroll
	 * 
	 */
	public void addLog(String msg)
	{
		logModel.addElement(msg); // add message to scroll
		int loc = logModel.size() - 1; // location of this message
		if (loc >0)
		{ // more than one message - make latest one visible
			logList.ensureIndexIsVisible(loc); // make this message visible
		}
	}

	/**
	 * Page resized - adjust things.
	 * This method will resize the panel containing the scrolled
	 * log messages.
	 * 
	 */
	public void resizeScroll()
	{
		Dimension wdim = this.getSize(); // window size
		Dimension sdim = logPane.getSize(); // log size
		// adjust size to fit inside window
		sdim.width = wdim.width - 10; // space for scroll bars
		sdim.height = wdim.height - 70; // space for upper and lower window bars
		logPane.setSize(sdim); // set size to fit inside window
		logScroll.setPreferredSize(sdim); // also set preferred size
		// int loc = wdim.height - sdim.height - 70;
		logPane.setLocation(0, 0); // place scroll at upper left corner
	}
	
	/**
	 * React to action command.
	 * This method responds to an event triggered by a user action.
	 * 
	 */
	public void actionPerformed(ActionEvent e)
	{
		String cmd = e.getActionCommand().toLowerCase(); // get command string
		
		if (cmd.equals("input"))
		{ // select input USPS text file
			fileIn = getFile("USPS Text");
		}
		else if (cmd.equals("csv"))
		{ // select output CSV file
			fileCsv = getFile("CSV");
		}
		else if (cmd.equals("convert"))
		{ // convert input USPS file to output CSV file
			convertZipCodes();
		}
		else if (cmd.equals("clear"))
		{ // clear log messages
			logModel.removeAllElements();
			logList = new JList(logModel);
		}
		else if (cmd.equals("exit"))
		{
			exitWindow();
		}
		else
		{
			System.err.println("Invalid action command [" + cmd + "]");
		}
	}
	
	/**
	 * Allow user to select the specified file.
	 * This method brings up a file chooser and
	 * allows the user to select a file.
	 * The string passed as a parameter can be used
	 * to identify to the user what file they are selecting.
	 * 
	 * @param desc a term to specify what file is to be chosen.
	 * @return the selected file.
	 * 
	 */
	public File getFile(String desc)
	{
		File file = null; // clear any past selection
		fc.setDialogTitle("Open " + desc + " file"); // add description as title
		int ret = fc.showOpenDialog(this); // bring up the file chooser for the user
		if (ret == JFileChooser.APPROVE_OPTION) 
		{ // user selected a file
            file = fc.getSelectedFile(); // get the selected file
            addLog(desc + " file is " + file.getPath()); // log selection
        } else 
        { // user canceled the selection
            addLog("User canceled opening " + desc + "file"); // log action
            // file was initialized to null
        }

		return file; // return file, or null if canceled
	}

	/**
	 * Convert input file to output file in the background.
	 * This method creates a ConvertWorker object to perform
	 * the actual conversion in the background.
	 * 
	 */
	public void convertZipCodes()
	{
		ConvertWorker worker = new ConvertWorker(); // create a worker object
		worker.execute(); // execute the worker in the background
	}

	/**
	 * Terminate the program.
	 * This method exits, closing the window.
	 */
	public void exitWindow()
	{
		System.exit(0); // terminate the program
	}
	
	/**
	 * Class to read the input USPS City/State National zip code text file.
	 * This class reads the input zip code text file, which consists
	 * of fixed-length records without any end of line sequence.
	 * The actual format of this file is described 
	 * <a href="WrpBuildZipCodes.html#fmt">in class WrpBuildZipCodes</a>.
	 * </p><p>
	 * This class will read only the detailed zip code records.
	 * 
	 * @author Bill Pringle
	 *
	 */
	class ZipFileReader
	{
		/** input file reader */
		protected FileReader rdr=null;
		/** the most recent error message */
		public String errMsg=null;
		
		/**
		 * Default constructor.
		 * If this constructor is called,
		 * then the input file must be set using the
		 * <a href="#openFile(java.lang.String)"openFile(String)</a> method.
		 * 
		 */
		public ZipFileReader()
		{
			
		}
		
		/**
		 * Constructor with input file name.
		 * This constructor opens the specified input file.
		 * If the file is not found, the error messasge is set.
		 * 
		 * @param fname the name of the USPS input text file
		 */
		public ZipFileReader(String fname)
		{
			try 
			{ // open the input file
				rdr = new FileReader(fname);
			} catch (FileNotFoundException e) 
			{ // file wasn't found - set error message
				errMsg = "ZipReader: File not found exception [" + fname + "]";
				System.err.println(errMsg); // display error message
			}
		}
		
		/**
		 * Constructor with the input file object.
		 * This method opens the input file as a FileReader.
		 * If the file is not found, the error message is set.
		 * 
		 * @param file the input file object for the USPS text file
		 */
		public ZipFileReader(File file)
		{
			try 
			{
				rdr = new FileReader(file); // open as a file reader
			} catch (FileNotFoundException e) 
			{ // ignore exception - set error message
				errMsg = "ZipReader: File not found exception [" + file.getPath() + "]";
				System.err.println(errMsg); // display error message
			}
		}
		
		/**
		 * Open the specified input text file.
		 * This method opens the specified USPS input text file,
		 * returning the status of the open.
		 * 
		 * @param fname the name of the USPS input City/State text file
		 * @return true if the file is opened; false on error
		 */
		public boolean openFile(String fname)
		{
			boolean ret = false; // assume false
			try 
			{
				rdr = new FileReader(fname); // open the specified file
				ret = true; // indicate success
			} catch (FileNotFoundException e) 
			{ // file not found
				errMsg = "Input file [" + fname + "] was not found";
				// ret is already set to false
			}
			return ret; // return result of open operation
		}
		
		/**
		 * Test to see if the input file is currently open.
		 * This method returns the status of the input file.
		 * When the input file is closed, the file object
		 * is set to null in the <a href="#close()">close</a> method.
		 * 
		 * @return true if the file is open; false otherwise
		 */
		public boolean isOpen()
		{
			return (rdr!=null); // true if file is open
		}
		
		/**
		 * Read the next line from the input file.
		 * This method returns the next detailed record.
		 * It will test to determine if the input file
		 * is open.  If so, it will read the next
		 * zip code detail record.
		 * If no detail records remain,
		 * the method will return null.
		 * 
		 * @return the next zip code in input file, or false if none remain
		 * 
		 */
		public WrpZipCode readLine()
		{
			WrpZipCode ret = null; // will be the next zip code (if any)
			if (rdr!=null) ret = nextFileZip(rdr); // file open, read line
			return ret; // return results of read
		}

    	/**
    	 * Read the next detailed zip code line from the raw input file.
    	 * Since the input text file contains fixed-length records
    	 * without any end of line sequence, we must read a fixed number
    	 * of characters and then parse the result.
    	 * 
    	 * @param rdr a file reader opened to the raw citystate.txt input file
    	 * @return a zip code object containing the information from the input line
    	 */
    	public WrpZipCode nextFileZip(FileReader rdr)
    	{
    		WrpZipCode ret = null; // will contain the next zip code record
    		if (rdr==null) return ret; // input file not opened - nothing to return
    		char[] zrec = new char[USPS_RECORD_SIZE]; // input file record
    		String str = null; // will contain the input record
    		//
    		// loop reading the input file until a detail record is found,
    		// then parse the record and return the results
    		//
    		while (true) 
    		{ // look for next detail record
    			int n; // will have how many characters were read
    			try 
    			{ // read the next input record
    				n = rdr.read(zrec); // should always be a full record (129)
    			} catch (IOException e) 
    			{ // I/O error - problems with input file
    				errMsg = "ZipReader.nextFileZip: I/O Error while reading zipcode file [" + e.getMessage() + "]";
    				System.err.println(errMsg); // display error message
    				return null; // corrupt file
    			}
    			if (n != USPS_RECORD_SIZE) 
    			{ // EOF or partial record
    				if (n>0)
    				{ // partial record - should never happen
    					errMsg = "Partial record read - " + n + " instead of "
    						+ USPS_RECORD_SIZE;
    				}
    				return null; // no more zip code records
    			}
    			// record read, check for detail record
    			if (zrec[0]=='D') break; // only parse detail records
    		}
    		// detail record found - parse record
    		/****************************
    		StringBuffer sb = new StringBuffer();
    		for (int i=0; i<zrec.length; i++)
    		{
    			char c = zrec[i];
    			sb.append(c);
    		}
    		str = sb.toString();
    		**************************/
    		str = new String(zrec);
    		// detail record encountered - parse record
    		// int zipcode = 0;
    		// extract fields from record
    		String strZip = str.substring(USPS_ZIP_START, USPS_ZIP_END); // five digit, leading zeros
    		String state = str.substring(USPS_STATE_START, USPS_STATE_END); // state abbreviation
    		String city = str.substring(USPS_CITYNAME_START,
    				USPS_CITYNAME_END).trim(); // city name, trailing spaces trimmed
    		// any abbreviation usually also appears as a city name for the same zip code
    		// duplicates will be resolved later
    		String abbr = str.substring(USPS_CITYABBR_START,
    				USPS_CITYABBR_END).trim(); // city abbreviation, trailing spaces trimmed

    		// no need to validate fields, since file comes from USPS
    		ret = new WrpZipCode(strZip, state, city); // instantiate zip code
    		if (abbr.length() > 0)
    		{ // abbreviation present - add it
    			ret.addCity(abbr); // add abbreviation
    		}
    		return ret; // return zip code object
    	}
    	
    	/**
    	 * Return the latest error message (if any).
    	 * This method will return the latest error message
    	 * and reset the message to null.
    	 * 
    	 * @return the current error message (possibly null)
    	 * 
    	 */
    	public String getErrorMessage()
    	{
    		String ret = errMsg;
    		errMsg = null;
    		return ret;
    	}
		
    	/**
    	 * Close the input file reader.
    	 * This method will return the current input file object
    	 * and free any memory associated with the instance.
    	 * 
    	 * @throws IOException
    	 */
		public void close() throws IOException
		{
			rdr.close(); // close the file
			rdr = null; // reset the reader to null
		}
	}

	/**
	 * This class converts the USPS input text file 
	 * to a CSV text file in the background.
	 * Some methods in this class run in the background and converts the 
	 * USPS input City/State National input file to a CSV file
	 * that can be imported into a database.
	 * </p><p>
	 * This class uses the <a href="WrpBuildZipCodes.ZipFileReader.html">
	 * ZipFileReader</a> class to read the input file,
	 * builds a list of the cities for a given zip code,
	 * and then writes the results to a CSV text file.
	 * The fields of the output CSV file are:
	 * </p>
	 * <ul>
	 * <li>Zip Code</li>
	 * <li>State Abbreviation</li>
	 * <li>City Name</li>
	 * </ul>
	 * 
	 * @author Bill Pringle
	 *
	 */
    private class ConvertWorker extends SwingWorker<Void, String>
    {
    	/**
    	 * Default (and only) constructor.
    	 * This constructor is called to instantiate the class.
    	 * After this is done, the <i>execute()</i> method
    	 * should be called to initiate the conversion.
    	 * 
    	 */
    	public ConvertWorker()
    	{
    		
    	}
    	
    	/**
    	 * This is the main background method to convert records.
    	 * This method calls the method 
    	 * <a href="#convertZipCodes()">convertZipCodes</a> to
    	 * perform the actual conversions.
    	 * 
    	 */
    	@Override
    	protected Void doInBackground()
    	{ // main background routine
    		convertZipCodes(); // convert the input file to the output file
    		return null; // normal return
    	}

    	/**
    	 * Foreground method to
    	 * process the most recent entries that were converted.
    	 * This method runs in the foreground and is called
    	 * periodically to process any entries that were generated
    	 * in the background task.
    	 * </p><p>
    	 * This list passed as an argument is a collection of the
    	 * strings that were passed from the <i>publish()</i>
    	 * background method.
    	 * Rather than list all zip codes that have been processed,
    	 * this method simply displays the most recent zip code
    	 * entry in the scrolled log list.
    	 * 
    	 */
    	protected void process(List<String> str)
    	{
    		int n = str.size() - 1; // location of last element
    		if (n>=0)
    		{ // not empty - log the last entry
    			String s = str.get(n); // get the last entry in the list
    			addLog(s); // log latest entry
    		}
     	}

    	/**
    	 * Convert input file to output CSV file.
    	 * This method validates the necessary parameters,
    	 * and then reads the input file, building a list
    	 * of cities associated with each zip code,
    	 * and creates a CSV text file containing the zip code
    	 * information.
    	 * 
    	 */
    	public void convertZipCodes()
    	{
    		if (fileIn == null)
    		{ // no input file defined
    			addLog("Input text file not specified");
    			return; // abort
    		}
    		if (fileCsv == null)
    		{ // no output file defined
    			addLog("Output CSV file not specified");
    			return; // abort
    		}

    		// open input file, read and process zipcode detail records
    		ZipFileReader zrdr=null; // will read the input file
    		zrdr = new ZipFileReader(fileIn); // create input file reader
    		if (zrdr==null || !zrdr.isOpen())
    		{ // problems with input file
    			addLog("Can't open input USPS text file [" + fileIn.getPath() + "]");
    			return; // abort
    		}

    		// create output CSV text file
    		try {
    			prt = new PrintStream(fileCsv); // create output file
    		} catch (FileNotFoundException e) {
    			addLog("Can't open output file [" + fileCsv + "], msg=[" + e.getMessage() + "]");
    			return; // abort
    		}
    		
    		// initialize loop variables
    		int num=0; // number of records written
    		String cZip = ""; // current zip code value
    		WrpZipCode currZip = null; // current zip code object

    		//
    		// input and output files are set up correctly
    		// loop reading the input file and create output
    		// records for each zip code, along with all the
    		// valid cities associated with each zip code.
    		//
    		try {
    			while (true)
    			{
    				WrpZipCode zip = zrdr.readLine(); // get next detail record
    				if (zip==null) break; // end of file
    				String strZip = zip.getZipcode(); // get this zip code value

    				if (!strZip.equals(cZip))
    				{ // new zipcode - write old one (if present)
    					if (currZip!=null)
    					{ // previous zip code exists, write it out
        					writeZipCode(currZip, prt); // write output lines
        					num += currZip.getNumberCities(); // number of records written
    					}
    					// set loop variables for future iterations
    					currZip = zip; // save current zip code object
    					cZip = strZip; // save current zip code value
    				}
    				else
    				{ // same zipcode - add city
    					// add any new cities to the current zip code
    					addCities(currZip, zip); // merge lists of cities
    				}
    			}
    			// done with input file - clean up
    			writeZipCode(currZip, prt); // write last zipcode
    			zrdr.close(); // close input file
    			prt.close(); // close output file
    		} catch (IOException e) { // problems
    			addLog("I/O Exception while reading input file [" + e.getMessage() + "]");
    		}
    		publish(num + " entries written."); // report results

    	}
    	
    	/**
    	 * Add the cities in the zipcode <i>zip</i> to the current zipcode <i>czip</i>.
    	 * This method will add any new cities to the current zipcode object.
    	 * Since a detail record can contain an abbreviation, and since
    	 * that abbreviation can also appear in a later detail record,
    	 * this method will check for duplicates, and only add
    	 * new city entries.
    	 * 
    	 * @param czip the current zipcode
    	 * @param zip the zipcode to add to the current zipcode
    	 * 
    	 */
    	protected void addCities(WrpZipCode czip, WrpZipCode zip)
    	{
    		for (int i=0; i<zip.getNumberCities(); i++)
    		{ // loop through cities for the new zip code object
    			String city = zip.getCity(i); // get the city name
    			if (!czip.hasCity(city))
    			{ // city is not present, add it
    				czip.addCity(city);
    			} // if city already defined, it is not added
    		}
    	}
    	
    	/**
    	 * Write CSV lines to the output file for this zip code.
    	 * This method will write a CSV line for each city associated
    	 * with this zip code entry.
    	 * This method assumes that a zip code will never span a state,
    	 * or at least that a given zip code object will only be for
    	 * a single state.
    	 * This method publishes each zip code that it writes,
    	 * although the foreground method <i>process()</i> will
    	 * only log the bottom of the lists that it receives.
    	 * 
    	 * @param zip the zip code object
    	 * @param prt the output CSV file
    	 */
    	protected void writeZipCode(WrpZipCode zip, PrintStream prt)
    	{
    		if (zip==null) return; // nothing to write
    		String strZip = zip.getZipcode(); // get the zip code value
    		String strState = zip.getState(); // get the state for this zip code
    		for (int i=0; i<zip.getNumberCities(); i++)
    		{ // loop through the cities for this zip code
    			String strCity = zip.getCity(i); // get city name
    			String s = strZip + "," + strState + "," + strCity; // build output line
    			prt.println(s); // print line
    			publish("ZipCode [" + s + "]"); // publish results - see process()
    		}
    	}

    }
    
	// ComponentListener routines

    /** called when window is hidden */
   public void componentHidden(ComponentEvent e)
    {
    }
    
   /** called when window is moved */
    public void componentMoved(ComponentEvent e)
    {
    }
    
    /**
     * called when the window is resized.
     * This method resizes the log window to fit in the new window.
     * 
     */
    public void componentResized(ComponentEvent e)
    {
    	resizeScroll();
    }
    
    /** called when the window becomes visible */
    public void componentShown(ComponentEvent e)
    {
    }

    /**
     * Main driver for this class.
     * This program will instantiate the class
     * and run it.
     * 
	 * @param args not used
	 */
	public static void main(String[] args) {
        java.awt.EventQueue.invokeLater(new Runnable() {
            public void run() {
                new WrpBuildZipCodes().setVisible(true);
            }
        });
	}

}
