package com.billpringle.utils.wrputils;

/*
 * WrpCsv.java
 *
 * Created on March 18, 2007, 9:07 PM
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

import java.io.*;
import java.util.*;

/**
 *
 * This class encapsulates a parser for CSV files.
 * This class can also be used to read CSV data from
 * a buffered input stream, such as a file.
 * </p><p>
 * The input string is separated into fields 
 * by any of the specified delimiters, 
 * which defaults to a comma.
 * </p><p>
 * This implementation is based on a similar
 * routine that I wrote in C, which was in turn based loosely on
 * Kernighan and Pike's "The Practice of Programming."
 * The C version I wrote, and this Java version
 * are both available under a Creative Commons license
 * (<a href="http://creativecommons.org/licenses/by-nc-sa/2.0/" target="ccl">
 * Attribution, Non-Commercial, and Share-Alike)</a>).
 * </p><p>
 * This class is normally called from other classes,
 * but contains a test driver <a href="#main(java.lang.String[])">main</a> routine
 * for testing.
 *</p>
 *
 *
  <table border="0" cellspacing="0" cellpadding="6" align="center"
  summary="Creative Commons license">
    <tr>
      <td><!-- Creative Commons License -->
      <a rel="license" href=
      "http://creativecommons.org/licenses/by-nc-sa/2.0/" target=
      "CC"><img alt="Creative Commons License" border="0" src=
      "http://creativecommons.org/images/public/somerights20.gif" /></a></td>

      <td><a rel="license" href=
      "http://creativecommons.org/licenses/by-nc-sa/2.0/" target=
      "CC"><img alt="Creative Commons License Symbols" border="0"
      src="cc_lic.gif" /></a> <!-- /Creative Commons License --></td>

      <td>Unless noted otherwise, all materials available for
      download from my site are copyrighted by Bill Pringle, and
      are licensed under a <a rel="license" href=
      "http://creativecommons.org/licenses/by-nc-sa/2.0/" target=
      "CC">Creative Commons License</a>.</td>
    </tr>
  </table>
  
 * @author Bill Pringle
 * 
 */
public class WrpCsv {
    //
	// default limits, which can be overridden by the user
	//
	/** 
	 * Default maximum number of fields.
	 * This is the largest number of fields
	 * that the CSV structure can handle.
	 * It can be overridden using the 
	 * <a href="#setMaxfields(int)"><b>setMaxFields</b></a> method,
	 * which will increase the maximum number of fields
	 * 
	 */
    private static int defaultMaxFields = 100;
    /** 
     * Default maximum size of the input line.
     * This is the longest size of the input line
     * can be handled.
     * This can be overridden by using the <b>setMaxLine</b> method.
     */
    private static int defaultMaxLine = 4096;
    /** 
     * Default collection of delimiters.
     * Any character within this string will act as a delimiter.
     * Normally this string consists of a single character,
     * but it is possible to use more than one character as a delimeter.
     * If more than one character appears in the delimiter string,
     * then any of those characters will be treated as a delimiter.
     * In other words, you cannot define a combination of strings to form
     * a single delimiter - each character would act as a delimiter.
     */
    private static String defaultDelims = ",";
    
    //
    // class level variables
    //
    /** Maximum input line size (not used) */
    protected int maxLine=defaultMaxLine;
    /** Input line to be split into fields */
    protected String parseLine;    
    /** List of fields parsed from input line */
    protected Vector<String> fields=null;
    /** Number of fields found in the input line */
    protected Integer numFields;  
    /** 
     * Maximum number of fields expected.
     * Parsing will terminate when this many fields are encountered.
     * If more fields are found, the will be contained
     * within the last field parsed.
      */
    protected Integer maxFields=defaultMaxFields;  
    /** Minimum number of fields expected (not used) */
    protected Integer minFields=0;  
    /** 
     * Possible field separators.
     * Any of the characters within this string
     * will terminate a field. 
     */
    protected String delims=defaultDelims;  
    /** 
     * Error message for last error.
     * This value can be retrieved if an error status
     * is returned by using the 
     * <a href="#getErrorMessage()">getErrorMessage</a> method.
     * 
     */
    protected String errorMessage;    
    
    /**
     * Default constructor.
     * </p><p>
     * This constructor must be used if the default
     * values are not appropriate.
     * The client should use this constructor,
     * and then the appropriate setter routines to modify
     * the default values.
     * </p><p>
     * If this constructor is used,
     * one of the <b>parseString</b> methods
     * must be used to parse the string.
     * 
     */
    public WrpCsv() {
        init();
    }

    /**
     * Creates a new instance of WrpCsv 
     * and parses the specified string into fields.
     * </p><p>
     * This method will use the default values and delimiters
     * to parse the fields.
     * If different delimiters are to be used,
     * the client can use the
     * <a href="#WrpCsv(String, String)">WrpCsv(String, String)</a>
     * constructor.
     * 
     * @param str string to be parsed
     * 
     */
    public WrpCsv(String str)
    {
        parseString(str);       // parse the string
    }
    
    /**
     * Creates a new instance of WrpCsv 
     * and parses the specified string into fields,
     * using the specified delimiter(s).
     * </p><p>
     * This constructor creates a new instance of WrpCsv and parses the string
     * into fields using the specified set of delimiters,
     * using the default values.
     * 
     * @param str input string to parse
     * @param delstr field delimiters
     * 
     */
    public WrpCsv(String str, String delstr)
    {
        setDelims(delstr);
        parseString(str);       // parse the string
    }
    
    /////////////////////////////////////////////////////////////
    //
    // Public Methods
    //
    ////////////////////////////////////////////////////////////
    
    /**************************************************************
     *              Set/Get Class variables
     *************************************************************/
    /** Get the maximum line size (not really used) */
    public int getMaxLine() { return maxLine; }
    /** Set the maximum size of the input line */
    public void setMaxLine(int val) { if (val>0) maxLine = val; }
    
    /**
     * Set field delimiters for current CSV.
     * The default delimiter is a comma (","). 
     * More than one string is allowed (e.g., "|/" bars or slashes);
     * either character will delimit the fields.
     * If the argument is null or empty, this call is ignored.
     * 
     * @param str
     */
    public void setDelims(String str)
    {
        if (str!=null && str.length() > 0) delims = str; // valid delimiters
    }

    /**
     * Return current delimiters.
     * 
     * @return string of delimiters
     */
    public String getDelims()
    {
        return delims;  // return current delimiters
    }

    /**
     * Return the most recent error message.
     * 
     * @return the message
     */
    public String getErrorMessage()
    {
        return errorMessage;
    }
    
    /**
     * Clear the current error message.
     * This method can be called by the client
     * after retrieving the most recent error message.
     * 
     */
    public void clearErrorMessage()
    {
    	errorMessage = null;
    }

    /**
     * Return the current maximum number of fields.
     * 
     * @return max. number of fields
     */
    public int getMaxFields() { return maxFields; }

    /**
     * Set the maximum number of fields.
     * If the parser contains more than the maximum
     * number of fields, parsing terminates and
     * the last field contains all remaining fields
     * from the string.
     * If <i>val</i> is non-positive, this call is ignored. 
     * 
     * @param val new max. number of fields
     */
    public void setMaxFields(int val)
    { // validate parameter - ignore invalid values
        if (val > 0) maxFields = val;
    }
    
    /**
     * Main parsing routine - parse string into fields.
     * </p><p>
     * This method drives the parsing, by 
     * scanning the input line and calling
     * either quotedField or nonquotedField
     * to extract the actual field.
     * </p><p>
     * Upon return, each field is stored in the
     * <a href="#fields">fields</a> vector, and can be accessed
     * using the <a href="#getField(int)">getField()</a> method.
	 *
     * @param str string to parse
     * @return number of fields found
     * 
     */
    public int parseString(String str)
    {
    	if (str==null) return 0;
        // reset class variables for new string
        init();                 // just in case not already called

        // parse entire string into individual fields
        parseLine = str;     // string to parse
        int sloc = 0;            // starting location in input string
        int eloc;               // will be ending location of current field
        int len = parseLine.length();   // max. chars in line
        if (len==0) return 0; // empty line
        String fld;             // current field

        /**
         * This routine parses the string manually
         * StringToken and the split method of String aren't appropriate
         * since two double quotes are replaced by a single double quote,
         * and the delimiters can be any character in the <b>delims</b> string.
         * </p><p>
         * We parse the string to all but the last field.
         * The last field is handled separately,
         * since it will be the rest of the string,
         * regardless of any delimiters.
         * 
         */
        while ((sloc < len) && (numFields < (maxFields-1)))
        { // scan line and break into fields, up to the maximum
        	// if field starts with quote, scan for closing quote;
        	// else scan for delimiter
            if (parseLine.charAt(sloc)=='"')
            { // quoted field - scan for closing quote
                sloc++;     // actual start of next field (ignoring starting quote)
                eloc = quotedField(parseLine, sloc);
                if (eloc < 0)
                { // rest of line - no closing quote
                	fld = copyField(parseLine.substring(sloc)); // field is rest of line
                	eloc = parseLine.length(); // length will be one more than offset
                }
                else if (eloc >= len)
                { // rest of string
                	fld = copyField(parseLine.substring(sloc));
                }
                else
                { // end of field found - use it
                    fld = copyField(parseLine.substring(sloc, eloc));
                    if (eloc < len) eloc++;     // skip past the closing quote
                }
            }
            else
            { // not quoted field - search for next delimiter
                eloc = nonquotedField(parseLine, sloc); // search for terminating delimiter
                if (eloc < 0) 
                { // delimiter not found - use rest of string
                    fld = parseLine.substring(sloc); // field is rest of string
                    eloc = parseLine.length(); // this will terminate loop
                }
                else 
                { // delimiter found - extract it and point to next character 
                    fld = parseLine.substring(sloc, eloc); // extract field
                    sloc = eloc + 1;    // start of next field
                }
            }
            // store any field we found
            // if the line was empty, numFields should be -1
            if (numFields >=0 && numFields < maxFields)
            { // field found - store it and reset start character offset
                fields.insertElementAt(fld, numFields++);
                sloc = eloc+1;    // start of next field
            }
        }

        /**
         * We have processed all but the last field.
         * If any of the string hasn't been parsed yet,
         * it will be treated as the last field.
         */
        if (sloc < len && numFields==(maxFields-1))
        { // some of the string not parsed, store is as the final field
            if (numFields >=0 && numFields < fields.size())
            { // store the rest of the string as the last field
                fields.insertElementAt(parseLine.substring(sloc), numFields++);
            }
        }
        // if line ends with delimiter, then the last field is emptye
        // unless we hit the maximum number of fields, since
        // the entire string was used as the last field
        Character c = str.charAt(str.length()-1); // last character
        if (delims.indexOf(c) >=0 && numFields<maxFields)
        { // insert an empty field as the last field
        	fields.insertElementAt("", numFields++); // empty string, not null
        }
        return numFields; // return the number of fields that were found
    }

    /**
     * Parse string using specified delimiters.
     * </p><p>
     * The delimiters are saved, and will be used 
     * for all future calls unless they are reset.
     * 
     * @param str string to parse
     * @param delims delimiter characters to use
     * @return number of fields found
     * 
     */
    public int parseString(String str, String delims)
    { // set delims, and then parse string
        setDelims(delims); // change the delimiters
        return parseString(str); // call the mail parsing routine
    }

    /**
     * Read a line from an input stream and parse it.
     * </p><p>
     * This method can be used to parse each line in a file.
     * Once the CSV is created, by calling this method,
     * you will read and parse the next line in the input file.
     * </p><p>
     * After making this call, the client can call the various
     * get routines to retrieve the actual fields.
     * 
     * @param inp BufferedReader input file
     * @return number of fields parsed
     * 
     */
    public int readLine(BufferedReader inp)
    	throws IOException
    {
        String str = inp.readLine(); // read the next line
        return parseString(str); // call the main parser to parse the line
    }

    /** Return the current number of fields. */
    public int getNumFields() { return numFields; }
    /** Return the original line. */
    public String getLine() { return parseLine; }
    /** Get the specific field. */
    public String getField(int loc)
    { // validate argument, and return null if invalid
        if (loc < 0 || loc >= numFields) return null; // invalid argument
        return (String)fields.elementAt(loc); // return requested field
    }
    
    /**
     * Set the specified field.
     * </p><p>
     * This method sets the specified field to the specified string,
     * and returns the old value if any; otherwise null.
     * 
     * @param loc index of element to be set (zero based)
     * @param fld field value to be set
     * @return the old value if any; null otherwise
     */
    public String setField(int loc, String fld)
    {
    	String ret = null;

    	if (loc<0) return null; // invalid field
    	
    	/*
    	 * The position <i>loc</i> can be withing the existing vector
    	 * or beyond it.
    	 * If within the vector, we will return the old value.
    	 * If beyond the existing vector, we need to increase
    	 * the size of the existing vector.
    	 */
    	if (loc<numFields)
    	{ // loc in vector, save old value
    		ret = fields.get(loc); // old value
    	}
    	else
    	{ // set will increase the number of fields
    		numFields = loc+1;
    	}
    	
    	/*
    	 * If the current capacity of the vector is smaller
    	 * than the new element, we must increase the capacity
    	 * of the existing vector.
    	 * For efficiency reasons, we will increase the capacity
    	 * well beyond what is needed so far.
    	 */
    	if (loc>=fields.capacity())
    	{ // this will increase the size of the vector
    		fields.ensureCapacity(loc+100); // make room for more
    	}
    	
    	/*
    	 * The capacity is now correct, but the size of
    	 * the vector can still be too small.
    	 * Unlike capacity, the size of the vector
    	 * should reflect the number of elements
    	 * that actually are stored in the vector.
    	 * We will set the size to one greater than <i>loc</i>.
    	 * If the size is increased by more than one element,
    	 * the vector will be filled with null entries
    	 * between the old size and new size.
    	 */
    	if (loc >= fields.size())
    	{
    		fields.setSize(loc+1);    		
    	}
    	
    	// ready to actually set the element
    	fields.set(loc, fld); // store new value
    	return ret; // return old value, if any
    }
    
    /**
     * Export CSV as an output string.
     * </p><p>
     * This method will create a string suitable for a CSV file
     * containing the information that has been stored.
     * If the <i>quoteAll</i> flag is set, all fields
     * will be quoted.
     * If not, only those fields containing a delimiter
     * character will be quoted.
     * 
     * @param quoteAll
     * @return the CSV string
     */
    public String export(boolean quoteAll)
    {
    	StringBuffer sb = new StringBuffer();
    	String quote = "";
    	for (int i=0; i<numFields; i++)
    	{ // loop through fields
    		if (i>0) sb.append(","); // next field
    		String s = fields.get(i); // get field
    		if (quoteAll || containsDelim(s))
    		{
    			quote = "\"";
    		}
    		else
    		{
    			quote = "";
    		}
    		sb.append(quote+s+quote); // append field
    	}
    	
    	String ret = sb.toString();
    	return ret;
    }
    
    /**
     * Export the CSV as a string, quoting only when needed.
     * </p><p>
     * This method is calls <i>export(<b>boolean</b> quoteAll)</i>
     * with the <i>quoteAll</i> set to false,
     * indicating that only the fields containing a delimiter character
     * should be quoted.
     * 
     * @return the CSV string
     */
    public String export()
    {
    	return export(false);
    }

    ////////////////////////////////////////////////////////
    //
    // Private methods
    //
    /////////////////////////////////////////////////////////

    /**
     * Test if string contains a delimiter.
     * </p><p>
     * This method determines if the specified string
     * contains any of the delimiters that have been defined.
     * This class allows more than one character to be treated
     * as a delimiter. When the parser encounters any of the
     * possible delimiters, the current field is ended. 
     * 
     * @param s the string to test
     * @return true if any delimiters found; false otherwise
     */
    private boolean containsDelim(String s)
	{
    	int cnt = 0; // number of delimiters found so far
    	for (int i=0; i<delims.length(); i++)
    	{ // check for each delimiter
    		char c = delims.charAt(i); // a delimiter
    		if (s.indexOf(c)>=0) cnt++; // found a delimiter
    	}
    	return (cnt>0); // contains delimiters
	}

	/**
     * Initializes class variables.
     * </p><p>
     * This method is called by the constructors
     * and the <b>parseString()</b> methods
     * to initialize all class variables.
     * 
     */
    private void init()
    {
        parseLine = null; // no string to parse yet
        if (fields==null) fields = new Vector<String>(maxFields); // create array for fields
        numFields = 0; // no fields yet
        errorMessage = null;  // no error message yet
    }
    
    /**
     * Return index location of the closing quote for the current quoted field.
     * This method is called when a field starts with a double quote.
     * </p><p>
     * Internal double quotes are indicated by two double quotes in a row,
     * which is treated as a single character.
     * If not closing quote is found, the location past the end
     * of the string is returned.
     * This means that the value returned can be used as an argument
     * for the <b>substring</b> method to extract the field.
     * 
     * @param str string being parsed
     * @param sloc offset to start searching (starting with zero)
     * @return offset of closing quote (starting with zero)
     * 
     */
    private int quotedField(String str, int sloc)
    { // return offset to end of string
        int eloc = str.indexOf('"', sloc); // find closing quote
        if (eloc < 0) return eloc;    // missing closing quote
        if (eloc>=str.length()-1) return eloc;	// closing quote is next-to-last char
        // quote found before end of line - check for two double quotes
        if (str.charAt(eloc+1)!='"') return eloc;   // no embedded quotes
        // embedded quotes - find actual end of field
        eloc++;     // position of second quote
        //
        // At least one embedded quote found - check for others
        // search until we find a double-quote that isn't followed by another
        //
        while (eloc < str.length())
        { // find end double quote
            eloc = str.indexOf('"', eloc+1); // next quote
            if (eloc < 0) return str.length()+1; // missing closing quote
            // if next character is also a quote, skip both
            if (eloc<str.length()-1 && str.charAt(eloc+1) != '"') return eloc;
            // second quote found - skip it
            eloc++;
        }
        // should never happen
        return str.length()+1;
    }

    /**
     * Return the offset for the terminating delimiter for a non-quoted field.
     * </p><p>
     * This method will scan for the first occurrence of
     * any of the currently defined delimiter characters.
     * 
     * @param str string being parsed
     * @param sloc offset of where to start searching
     * @return offset of delimiter
     * 
     */
    private int nonquotedField(String str, int sloc)
    {
        char c; // temporary variable for examining string
        for (int i=sloc; i<str.length(); i++)
        { // find one of the defined delimiters
            c = str.charAt(i); // get next character
            if (delims.indexOf(c)>=0)
            { // found delimiter
                return i;   // location of terminating delimiter
            }
        }
        // no delimiter, rest of string is field
        return -1;      // end of string
    }

    /**
     * Copy string, transforming consecutive double quotes into one.
     * This method is used to extract the actual field,
     * allowing for consecutive double quotes, which are used
     * to indicate that the field contains a double quote.
     * Since this method is only called when dealing with quoted fields,
     * any double quote we encounter should be followed by a second one.
     * Only the first double quote will be copied.
     * 
     * @param str field to copy
     * @return copy of string
     * 
     */
    private String copyField(String str)
    {
        StringBuffer sb = new StringBuffer(); // will contain the returned field
        for (int i=0; i<str.length(); i++)
        { // loop through actual field
            char c=str.charAt(i); // get current character
            if (c=='"')
            { // found double quote - should be two of them
                if (i<str.length()-1 && str.charAt(i+1)=='"')
                { // two double quotes - ignore first one
                    i++;    // skip past first quote
                }
            }
            sb.append(c);
        }
        return sb.toString();   // return field
    }
    
    /* (non-Javadoc)
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
    	StringBuffer sb = new StringBuffer();
    	
    	for (int i=0; i<numFields; i++)
    	{
    		if (i>0) sb.append(","); // add comma between fields
    		String fld = fields.get(i);
    		if (fld!=null) sb.append("\"" + fld + "\"");
    	}
    	
    	String ret = sb.toString();
    	return ret;
    }
     
    /**
     * Test driver.
     * 
     * @param args not used
     * 
     */
     public static void main(String args[]) {
         WrpCsv csv = new WrpCsv();
         System.out.println("Enter delimiters: ");
         InputStreamReader rdr = new InputStreamReader(System.in);
         BufferedReader kbd = new BufferedReader(rdr);
         String delims, str;
         try
         {
             delims = kbd.readLine();
             csv.setDelims(delims);
         }
         catch (IOException e)
         {
            System.out.println("I/O error" + e.toString());
         }
         while (true)
         {
             System.out.println("Enter string to parse, or ENTER to quit: ");
             try
             {
                 str = kbd.readLine();
                 if (str.length()==0) break;    // done
                 int nflds = csv.parseString(str);
                 System.out.println(nflds + " fields parsed.\n");
                 for (int i=0; i<nflds; i++)
                 {
                     String s = csv.getField(i);
                     System.out.println(i + ">[" + s + "]<");
                 }
                 System.out.println("toString: [" + csv.toString() + "]");
                 String errmsg = csv.getErrorMessage();
                 if (errmsg!=null)
                 {
                	 System.err.println("*** Error *** [" + errmsg + "]");
                 }
             }
             catch (IOException e)
             {
                 System.out.println("I/O Error encountered " + e);
             }
         }
     }
}
