package com.billpringle.utils.wrputils;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Date;
import java.util.Vector;

/**
 * This class provides a wrapper for a database connection.
 * It is expected that clients will create their own classes
 * that extend this class.
 * The client class would then provide the connection information
 * for their specific database.
 * This would allow client programs to connect to the database
 * without having to provide the connection information.
 * A class extension can be done as follows:
 * </p>
 * <code>
 * <pre>
 * public class EMailDb extends WrpDb
 * {
 * 	public static String strDatabase = new String("<i>dbname</i>");
 *  public static String strHost = new String("<i>hostname</i>");
 * 	public static String strUser = new String("<i>username</i>");
 * 	public static String strPassword = new String("<i>password</i>");
 * 	
 * 	public EMailDb() throws ClassNotFoundException, IllegalAccessException, InstantiationException, SQLException
 * 	{
 * 		super(strDatabase, strHost, strUser, strPassword);
 * 	}
 * }
 * </pre>
 * </code>
 * <p>
 * By default, a MySQL database connection is established
 * with the following parameters:
 * </p>
 * <ul>
 * <li>Database name &mdash; db</li>
 * <li>Host name &mdash; localhost</li>
 * <li>User name &mdash; sa</li>
 * <li>User password is an empty string</li>
 * </ul>
 * <p>
 * To establish a connection using different parameters
 * use one of two means:
 * <dl>
 * <dd><code>WrpDb db = WrpDb(dbname, hostname, username, password);</code></dd>
 * </dl>
 * 
 * or:
 * 
 * <dl>
 * <dd><code>WrpDb db = WrpDb(false);</code></dd>
 * <dd><code>setDbName(dbname);</code></dd>
 * <dd><code>setHostName(hostname)</code></dd>
 * <dd><code> ... </code></dd>
 * <dd><code>openConnection();</code></dd>
 * </dl>
 * 
 * </p><p>
 * A connection can also be established to a non-MySQL database:
 * <dl>
 * <dd><code>WrpDb db = WrpDb(false);</code></dd>
 * <dd><code>setConnectString(connectionstring);</code></dd>
 * <dd><code>openConnection();</code></dd>
 * </dl>
 * </p>
 * 
 * 
  <table border="0" cellspacing="0" cellpadding="6" align="center"
  summary="Creative Commons license">
    <tr>
      <td><!-- Creative Commons License -->
      <a rel="license" href=
      "http://creativecommons.org/licenses/by-nc-sa/2.0/" target=
      "CC"><img alt="Creative Commons License" border="0" src=
      "http://creativecommons.org/images/public/somerights20.gif" /></a></td>

      <td><a rel="license" href=
      "http://creativecommons.org/licenses/by-nc-sa/2.0/" target=
      "CC"><img alt="Creative Commons License Symbols" border="0"
      src="cc_lic.gif" /></a> <!-- /Creative Commons License --></td>

      <td>Unless noted otherwise, all materials available for
      download from my site are copyrighted by Bill Pringle, and
      are licensed under a <a rel="license" href=
      "http://creativecommons.org/licenses/by-nc-sa/2.0/" target=
      "CC">Creative Commons License</a>.</td>
    </tr>
  </table>
  
 * @author Bill Pringle
 *
 */
public class WrpDb {
	/** The serial version ID */
	final static long serialVersionUID = 1;

	// Database variables
	/** 
	 * The database name.
	 * This variable is used to build the connection string
	 * by the method <a href="#buildConnectString()">buildConnectString()</a>.
	 * 
	 */
	public String strDatabase = new String("db");
	/** 
	 * The host  name 
	 * This variable is used to build the connection string
	 * by the method <a href="#buildConnectString()">buildConnectString()</a>.
	 * 
	 */
	public String strHost = new String("localhost");
	/** 
	 * The user name 
	 * This variable is used to build the connection string
	 * by the method <a href="#buildConnectString()">buildConnectString()</a>.
	 * 
	 */
	public String strUser = new String("sa");
	/** 
	 * The user password 
	 * This variable is used to build the connection string
	 * by the method <a href="#buildConnectString()">buildConnectString()</a>.
	 * 
	 */
	public String strPassword = new String("");
	/**
	 * The connection String.
	 * This string is the actual connection string
	 * used to open the database.
	 * In most cases, this string is built using the
	 * <a href="#buildConnectString()">buildConnectString()</a>
	 * method, which is called by most constructors.
	 * It can also be set by the client directly after
	 * using the <a href="#WrpDb(java.lang.Boolean)">boolean constructor</a>.
	 * 
	 */
	public String strConnect = new String("jdbc:mysql://" + strHost + ":3306/" + strDatabase + "?user=" + strUser + "&password=" + strPassword);
	/**
	 * Actual database connection.
	 * This object is initialized during the
	 * <a href="#openConnection()">openConnection()</a> method,
	 * and are used by various methods.
	 * The connection is closed by the
	 * <a href="#closeConnection()">closeConnection()</a> method.
	 * </p><p>
	 * This connection is declared public, so that the client
	 * can use the object directly.
	 * This can result in an unstable condition,
	 * and so the client should use all caution,
	 * and assume all responsibility for this action.
	 * 
	 */
	public Connection conn=null;	// database connection
	/**
	 * The actual Statement object.
	 * This object is initialized during the
	 * <a href="#openConnection()">openConnection()</a> method,
	 * and are used by various methods.
	 * The Statement is closed by the
	 * <a href="#closeConnection()">closeConnection()</a> method.
	 * </p><p>
	 * This connection is declared public, so that the client
	 * can use the object directly.
	 * This can result in an unstable condition,
	 * and so the client should use all caution,
	 * and assume all responsibility for this action.
	 */
	public Statement stmt=null;		// statement
	/**
	 * The character to be used for quoting strings
	 * in queries.
	 * </p><p>
	 * This character is declared public, and can be set using
	 * the <a href="#setSqlQuote(java.lang.Character)">setSqlQuote()</a>
	 * method.
	 * 
	 */
	public Character SqlQuote='\'';	// enclosing character for string values
	/**
	 * A single quote character (').
	 * </p><p>
	 * This variable can be used to modify the quote character
	 * that will be used to quote strings in queries
	 * using the <a href="#setSqlQuote(java.lang.Character)">setSqlQuote()</a> method.
	 * 
	 */
	public static Character SQLQUOTE_SINGLE = '\'';
	/**
	 * A double quote character (").
	 * </p><p>
	 * This variable can be used to modify the quote character
	 * that will be used to quote strings in queries
	 * using the <a href="#setSqlQuote(java.lang.Character)">setSqlQuote()</a> method
	 * 
	 */
	public static Character SQLQUOTE_DOUBLE = '"';
	
	/** driver string for MySQl */
	public static final String MYSQL = "com.mysql.jdbc.Driver";
	/** driver string for PostGres */
	public static final String POSTGRES = "org.postgresql.Driver";
	
	/**
	 * Driver class.
	 * The default is mysql.  If you are using a different driver,
	 * override this value in the class you use to extend this class.
	 */
	public String strDriver = MYSQL;
	
	/**
	 * This is the default constructor.
	 * It will open a connection using the default parameters.
	 * It is expected that this constructor will not be used often.
	 * 
	 * @throws ClassNotFoundException
	 * @throws IllegalAccessException
	 * @throws InstantiationException
	 * @throws SQLException
	 */
	public WrpDb()
		throws ClassNotFoundException, IllegalAccessException, 
		InstantiationException, SQLException
	{
		openConnection(); // open the connection using the default parameters
	}
	
	/**
	 * This constructor opens the database using specified parameters.
	 * This method creates a new instance of the class,
	 * and opens the database connection using the
	 * specified parameters.
	 * The connection assumes a MySQL database.
	 * 
	 * @param dbName the name of the database.
	 * @param hostName the name of the database server.
	 * @param userName is the user name for the database connection
	 * @param userPassword is the user password for the database connection
	 * @throws SQLException 
	 * @throws InstantiationException 
	 * @throws IllegalAccessException 
	 * @throws ClassNotFoundException 
	 * 
	 */
	public WrpDb(String dbName, String hostName, String userName, String userPassword) 
		throws ClassNotFoundException, IllegalAccessException, InstantiationException, SQLException
	{
		this.strDatabase = dbName; // set the database name
		this.strHost = hostName; // set the host name
		this.strUser = userName; // set the user  name
		this.strPassword = userPassword; // set the user password
		buildConnectString(); // build a MySQL connection string
		openConnection(); // open the connection using the specified parameters
	}

	/**
 * This constructor opens the database using specified parameters.
	 * This method creates a new instance of the class,
	 * and opens the database connection using the
	 * specified parameters.
	 * The connection assumes a MySQL database.
	 * 
	 * @param dbName the name of the database.
	 * @param hostName the name of the database server.
	 * @param userName is the user name for the database connection
	 * @param userPassword is the user password for the database connection
	 * @param strDriver driver signature (default is com.mysql.jdbc.Driver)
	 * @throws SQLException 
	 * @throws InstantiationException 
	 * @throws IllegalAccessException 
	 * @throws ClassNotFoundException 
	 * 
	 */
	public WrpDb(String dbName, String hostName, String userName, String userPassword, String strDriver) throws ClassNotFoundException, IllegalAccessException, InstantiationException, SQLException
	{
		this.strDatabase = dbName; // set the database name
		this.strHost = hostName; // set the host name
		this.strUser = userName; // set the user  name
		this.strPassword = userPassword; // set the user password
		this.strDriver = strDriver;
		buildConnectString(); // build a MySQL connection string
		openConnection(); // open the connection using the specified parameters
	}

	/**
	 * This constructor will open a database using the specified connection string.
	 * This constructor allows the client to specify the exact database connection
	 * parameters.
	 * This is most useful if the client needs a non-standard database connection.
	 * 
	 * @param strConnect connection string
	 * @param strDriver driver class
	 * @throws ClassNotFoundException
	 * @throws IllegalAccessException
	 * @throws InstantiationException
	 * @throws SQLException
	 */
	public WrpDb(String strConnect, String strDriver)
		throws ClassNotFoundException, IllegalAccessException, 
		InstantiationException, SQLException
	{
		this.strConnect = strConnect; // set the connection string
		this.strDriver = strDriver; // set driver class
		openConnection(); // open the database using this connection string
	}
	
	/**
	 * This constructor won't open the connection if the parameter is false.
	 * If the parameter is true, then this constructor is identical
	 * to the default constructor.
	 * </p><p>
	 * This constructor can be used to create the object
	 * but not open the connection.  The various parameters
	 * that make up the connection string can be specified
	 * using the various setter routines,
	 * (including specifying the connection string directly).
	 * Once complete, the database connection can then be opened
	 * using the <a href="#openConnection()">openConnection()</a>.
	 * 
	 * @param open if true, the connection is opened; otherwise it isn't
	 * @throws ClassNotFoundException
	 * @throws IllegalAccessException
	 * @throws InstantiationException
	 * @throws SQLException
	 */
	public WrpDb(Boolean open)
		throws ClassNotFoundException, IllegalAccessException, 
		InstantiationException, SQLException
	{
		if (open) openConnection(); // open connection is requested
	}
	
	/**
	 * Rebuild the connection string using the specified database name.
	 * This methodd changes the database name, 
	 * and then rebuilds the connection string.
	 * 
	 * @param dbName the name of the database to be opened
	 */
	public void setDbName(String dbName)
	{
		this.strDatabase = dbName; // set the database name
		buildConnectString(); // rebuild connection string
	}
	
	/**
	 * Rebuild the connection string using the specified host name.
	 * This method changes the host name and then
	 * rebuilds the connection string.
	 * 
	 * @param hostName the name of the MySql database host machine
	 * 
	 */
	public void setHostName(String hostName)
	{
		this.strHost = hostName; // set the host name
		buildConnectString(); // rebuild the connection string
	}
	
	/**
	 * Rebuild the connection string using the specified user name.
	 * This method changes the user name and then rebuilds
	 * the connection string.
	 * 
	 * @param userName the name of the user within the connection string
	 */
	public void setUserName(String userName)
	{
		this.strUser = userName; // set the user name
		buildConnectString(); // rebuild the connection string
	}
	
	/**
	 * Rebuild the connection string using the specified password.
	 * This method changes the user password, and then
	 * rebuilds the connection string.
	 * 
	 * @param password the user password
	 */
	public void setPassword(String password)
	{
		this.strPassword = password; // set the user password
		buildConnectString(); // rebuild the connection string
	}

	/**
	 * Rebuild the connection string using the previously
	 * specified components to the string.
	 * </p><p>
	 * This method is called by all of the setter methods,
	 * so the user should never need to call this method directly
	 * 
	 */
	public void buildConnectString()
	{ 
		// build connection string using pre-defined parameters
		if (strDriver.equals(MYSQL))
		{ // mysql
			strConnect = new String("jdbc:mysql://" 
					+ strHost + ":3306/" + strDatabase 
					+ "?user=" + strUser + "&password=" + strPassword);
		}
		else if (strDriver.equals(POSTGRES))
		{ // postgres
			strConnect = new String("jdbc:postgresql://"
				+ strHost + "/" + strDatabase
				+ "?user=" + strUser 
				+ "&password=" + strPassword);
		}
	}

	/**
	 * Manually set the connection string.
	 * Calling this method only makes sense if the
	 * <a href="#WrpDb(java.lang.Boolean)">boolean constructor</a> is used.
	 * If any of the connection-related setters are called after this method,
	 * the connection string will be rebuilt, overwriting this value.
	 * </p><p>
	 * This method can be used to establish a connection string
	 * for a non-MySQL database.
	 * For MySQL databases, the client can set database, host name,
	 * user name, and password with a constructor and/or setter routines.
	 * 
	 * @param str the connection string to be used to open the database.
	 */
	public void setConnectString(String str)
	{
		this.strConnect = str; // set the connection string
	}
	
	/**
	 * Get the current connection string.
	 * </p><p>
	 * This method returns the current connection string.
	 * 
	 * @return the current connection string.
	 */
	public String getConnectString()
	{
		return strConnect; // return the connection string
	}
	
	/**
	 * Establish a connection to the database.
	 * </p><p>
	 * This method assumes that the
	 * <a href="#strConnectString">connection string</a>
	 * has been set up correctly.
	 * 
	 */
	public void openConnection()
		throws ClassNotFoundException, IllegalAccessException, 
		InstantiationException, SQLException
	{
		Class.forName(strDriver).newInstance(); // get driver class
		conn = DriverManager.getConnection(strConnect); // get a connection
		// create statement that will be used during the connection session
		stmt = conn.createStatement(ResultSet.TYPE_SCROLL_INSENSITIVE, ResultSet.CONCUR_UPDATABLE);
	}
	
	/**
	 * Close database connection.
	 * This method will close any statement object as well as the connection object.
	 * 
	 * @throws SQLException 
	 * 
	 */
	public void closeConnection() throws SQLException
	{
		// close connection, including the statement if needed
		
		// test for statement
		if (stmt!=null)
		{ // statement created - try to close it
				stmt.close(); // close the statement
			stmt = null; // reset statement to prevent future use
		}
		
		// test for connection
		if (conn!=null)
		{ // connection created - try closing it
				conn.close(); // close connection
			conn = null; // reset connection to prevent future use
		}
	}
	
	/**
	 * Close the database connection, suppressing any exceptions.
	 * This method will close connection and suppress
	 * any exception that might be raised.
	 * To detect exceptions, the client can call
	 * the <a href="#closeConnection()">closeConnection</a> method.
	 * 
	 */
	public void close()
	{
		try {
			closeConnection();
		} catch (SQLException e) {
			System.err.println("SQL Exception while closing the database [" + e.getMessage() + "]");
		} // close the database connection
	}
	
	/**
	 * Execute the specified SQL query and return the result of the query.
	 * This method should only be used for those queries that will return a result.
	 * 
	 * @param query the SQL query to execute
	 * @return the ResultSet from the query (e.g., the records matching the query)
	 * @throws SQLException
	 * 
	 */
	public ResultSet executeQuery(String query) throws SQLException
	{
		if (stmt==null) return null; // no statement - ignore call
		ResultSet rs = stmt.executeQuery(query); // execute the query
		return rs; // return the results
	}
	
	/**
	 * Test if more results sets remaining.
	 * This method will test if any more result sets have been returned
	 * by the previous query (probably a stored procedure.)
	 * If an SQLException occurs, this method will return false.
	 * 
	 * @return true if more result sets; false otherwise
	 */
	public boolean getMoreResults()
	{
		if (stmt==null) return false; // no statement - ignore call
		try
		{
			return stmt.getMoreResults();
		} catch (SQLException e)
		{
			return false;
		}
	}

	/**
	 * Test if more results sets remaining, specifying what to do with current Result Set.
	 * This method will test if any more result sets have been returned
	 * by the previous query (probably a stored procedure.)
	 * The constant <i>current</i> indicates disposition of the
	 * current record set using the method 
	 * getResultSet: Statement.CLOSE_CURRENT_RESULT, 
	 * Statement.KEEP_CURRENT_RESULT, or Statement.CLOSE_ALL_RESULTS
	 * </p><p>
	 * If an SQLException occurs, this method will return false.
	 * See java.sql.Statement.getMoreResults.
	 * 
	 * @param current constant defined in Statement
	 * @return result set or null if none remain
	 */
	public boolean getMoreResults(int current)
	{
		if (stmt==null) return false; // no statement - ignore call
		try
		{
			return stmt.getMoreResults(current);
		} catch (SQLException e)
		{
			return false;
		}
	}
	
	/**
	 * Get next result set.
	 * This method can be used when more than one record set is
	 * returned from a query (probably a stored procedure).
	 * The normal way to use this would be:
	 * <br />
	 * if (db.getMoreResults()) rs = db.getResultSet();
	 * <br />
	 * 
	 * @return the record set; null in case of error
	 */
	public ResultSet getResultSet()
	{
		if (stmt==null) return null;
		try
		{
			return stmt.getResultSet();
		} catch (SQLException e)
		{
			return null;
		}
	}
	
	/**
	 * Execute the specified SQL update query and return the result.
	 * This method should only be used for those queries that do not
	 * return a result set, but rather the number of records affected.
	 * 
	 * @param query the update query to be executed
	 * @return the result of the query - the number of records affected
	 * @throws SQLException
	 * 
	 */
	public int executeUpdate(String query) throws SQLException
	{
		if (stmt == null) return -1;	// error - no statement
		int ret = stmt.executeUpdate(query); // execute update
		return ret; // return number of records
	}
	
	/**
	 * Return the column for the specified ResultSet.
	 * This method will return a vector containing the field
	 * values for the specified column in each row of the result set.
	 * 
	 * @param rs the ResultSet from an SQL selection query
	 * @param colName the name of the column to be retrieved
	 * @return a string vector containing the data stored in the column
	 * @throws SQLException
	 * 
	 */
	public Vector<String> getColumn(ResultSet rs, String colName) throws SQLException
	{
		if (rs==null) return null;	// no ResultSet
		Vector<String> ret = new Vector<String>(); // create vector for return value
		while (rs.next())
		{ // loop through the result set, getting the column value
			String s = rs.getString(colName); // get the column value
			ret.add(s); // add to the return vector
		}
		
		// finished processing the result set
		return ret; // return the column values
	}
	
	/**
	 * Make sure string is safe for SQL insert.
	 * SqlQuote defines which quote character is used for strings:
	 * either single (SQLQUOTE_SINGLE) or double (SQLQUOTE_DOUBLE) quotes.
	 * SqlQuote should be set correctly before this routine is called.
	 * </p><p>
	 * This method will do the following:
	 * </p>
	 * <ul>
	 * <li>Convert null pointers to empty strings</li>
	 * <li>Change any single/double quotes to double/single quotes</li>
	 * </ul>
	 * <p>
	 * It is the client's responsibility to call this method.
	 * This method should be used for each query before it is executed
	 * that involves user data in any form.
	 * In addition to preventing accidental data corruption (e.g., embedded
	 * quotes within a field value), it will also help prevent insertion attacks.
	 * 
	 * @param s candidate query string
	 * @return cleaned query string, or an empty string (if null)
	 * 
	 */
	public String sqlSafe(String s)
	{
		if (s==null || s.length()==0) return ""; // no data - return empty string
		StringBuffer sb = new StringBuffer(); // will contain the cleaned string
		for (int i=0; i<s.length(); i++)
		{ // loop through the query
			char c = s.charAt(i); // get character
			// if a character matches the quote character,
			// change it to the other quote character
			if (c==SqlQuote)
			{ // matches quote - replace with other quote character
				if (SqlQuote==SQLQUOTE_SINGLE) c = SQLQUOTE_DOUBLE;
				else c = SQLQUOTE_SINGLE;
			}
			sb.append(c); // add (possibly updated) character to resulting string
		}
		String ret = sb.toString(); // build the safe query string
		return ret; // return safe query string
	}
	
	/**
	 * Render date value as string for insert into database.
	 * </p><p>
	 * This method encloses quotes around any date value,
	 * or a null string if date is null.
	 * This method can be used along with SqlSafe when
	 * building a query.
	 * 
	 * @param d date value
	 * @return quoted date or null
	 */
	public String SqlDate(Date d)
	{
		String ret;
		
		if (d==null)
		{ // null date, return null string
			ret = null;
		}
		else
		{ // date present, place in quotes
			ret = "'" + d + "'";
		}
		
		return ret;
	}

	/**
	 * Return what type of quote is being used for queries.
	 * </p><p>
	 * This method can be used to determine how to handle embedded quotes.
	 * For example, if the query has literals enclosed in single quotes,
	 * then an apostrophe can cause problems.
	 * Likewise, if the query has literals enclosed in double quotes,
	 * then a double-quoted substring must be handled differently
	 * than if it were enclosed in single quotes 
	 * 
	 * @return the sqlQuote character
	 * 
	 */
	public Character getSqlQuote() {
		return SqlQuote; // return the current quote character
	}

	/**
	 * Specify what quote character should be used
	 * to enclose literals.
	 * Typically, this variable is set to either a
	 * <a href="#SQLQUOTE_SINGLE">single quote</a>
	 * or a
	 * <a href="#SQLQUOTE_DOUBLE">double quote</a>.
	 * The client can pass any character to this method,
	 * but care should be taken when doing this.
	 * 
	 * @param sqlQuote the sqlQuote to set
	 * 
	 */
	public void setSqlQuote(Character sqlQuote) {
		SqlQuote = sqlQuote; // set the current quote character
	}
}
