package com.billpringle.utils.wrputils;


import java.awt.Container;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintStream;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Date;
import java.util.Vector;

import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.JTextField;
import javax.swing.KeyStroke;
import javax.swing.ScrollPaneConstants;
import javax.swing.SpringLayout;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.TableColumn;

import com.billpringle.utils.wrputils.WrpValidateCsv.ValidationRuntimeException;



/**
 * This class allows a user edit validation rules that are stored
 * in a database.
 * </p><p>
 * In addition to editing the rules, the user can also compile
 * them, and even execute them to validate a specified CSV line
 * or all of the lines within a specified CSV file. 
 * 
 * @author pringlb
 *
 */

public class WrpEditValidationRules extends JFrame
	implements ActionListener, MouseListener, ComponentListener
{
	/** unique serializing identifier */
	final static long serialVersionUID = 1;

	/** the connection to the database */
	protected WrpDb db = null;
	/** the unique line number for this record */
	protected int csvLine=1;
	/** 
	 * The CSV string that can be validated.
	 * </p><p>
	 * When the "Execute" button is clicked,
	 * this string will be validated,
	 * provided the user hasn't specified a CSV file
	 * to validate.
	 * 
	 */
	protected WrpCsv csv=null;
	/** 
	 * The file containing CSV records that can be validated.
	 * </p><p>
	 * If the user has specified a CSV file,
	 * it will be validated when the user clicks
	 * on the "Execute" button.
	 * 
	 */
	protected File csvFile=null;
	/** the database table name containing validation rules */
	protected String ValTbl = "newvalidationrules";
	/** CSV validation object */
	protected WrpValidateCsv Val = new WrpValidateCsv();
	/** vector of validation rules */
	protected Vector<WrpValidateCsv.ValidationRule> validations=null;
	/** vector indicating what rules have been modified */
	protected Vector<Boolean> modified=null;
	/** The number of rules read from the database */
	protected int rulesRead=0;
	/** 
	 * The width of the rules columns that was set by the user.
	 * </p><p>
	 * When rules are reloaded, these column widths will be retained.
	 * 
	 */
	protected int[] colWidth = {0,0,0,0,0,0,0,0,0,0,0};
	
	// variables for test compilation and execution
	/** number of compile errors (-1 means no compile) */
	protected int compileErrors=-1;
	/** number of run-time errors (-1 means no execution) */
	protected int execErrors=-1;
	
	/** list of valid zip code combinations */
	protected WrpZipCode zipCodes = null;
	
	//
	// GUI components
	//
	/** pop up menu */
	JPopupMenu popup=null;
	/** panel at top of page */
	JPanel pnlTop=null;
	/** wrapper panel to prevent top of page resizing */
	JPanel pnlTopPage=null;
	/** label for title at top of page */
	JLabel lblTitle=null;
	/** table of validation rules */
	JTable tblRules=null;
	/** optional lower sequence number to load a subset of the validation rules */
	JTextField txtLowSeq=null;
	/** optional upper sequence number to load a subset of the validation rules */ 
	JTextField txtHighSeq=null;
	/** optional field to indicate stored procedure */
	JTextField txtTable=null;
	/** button to load validation rules */
	JButton btnGet=null;
	/** button to save validation rules */
	JButton btnSave=null;
	/** button to compile the validation rules */
	JButton btnCompile=null;
	/** button to print validation rules */
	JButton btnPrint=null;
	/** 
	 * Button to specify a CSV string to parse
	 */
	JButton btnCsv=null;
	/** 
	 * Button to specify a CSV file to parse 
	 */
	JButton btnFile=null;
	/**
	 * Button to validate one or more CSV strings
	 * </p><p>
	 * If the user has specified a CSV file,
	 * then all the lines in that file are validated
	 * using the current set of validation rules.
	 * If no CSV file is specified, then the specified
	 * CSV string is validated instead.
	 * 
	 */
	JButton btnExecute=null;
	/** button to renumber statements */
	JButton btnRenum=null;
	/** button to terminate the program */
	JButton btnQuit=null;
	/** scroll panel for the validation rules */
	JPanel pnlScroll=null;
	/** scroll pane for the validation rules */
	JScrollPane scroll=null;
	/** size of the top portion of the page */
	Dimension topSize = null;
	/** a file chooser to select the CSV file */
	JFileChooser fc = new JFileChooser();
	
	/**
	 * Default (and only) constructor
	 */
	public WrpEditValidationRules(WrpDb db)
	{
		/* create a vector to store valildation rules */
		validations = new Vector<WrpValidateCsv.ValidationRule>();
		setSize(600, 300); // default window size
		buildMenu(); // build the window menu system
		buildPage(); // create and arrange the GUI components 
		setTitle("Edit Known Shipper Validation Rules"); // set frame title
		setDefaultCloseOperation(EXIT_ON_CLOSE); // terminate when window is closed

		/*
		 * Open a database connection that will be used to load
		 * and possibly save the validation rules.
		 */
		try
		{
			this.db = db;
			zipCodes = new WrpZipCode(db);
			Val.setZipCodes(zipCodes);
		} catch (Exception e)
		{
			JOptionPane.showMessageDialog(
					this, 
					"Problems accessing the Database! [" + e.getMessage() + "]", 
					"Error on Open", 
					JOptionPane.ERROR_MESSAGE);
			System.exit(2);
		}
		// pack();
		repaint(); // display the page
		topSize = pnlTop.getSize(); // save initial size of top panel
	}
	
	/**
	 * Creates the main menu (e.g., File)
	 * 
	 */
	public void buildMenu()
	{
		JMenuBar menuBar = new JMenuBar();
		JMenu menu = new JMenu("File");
		menu.setMnemonic(KeyEvent.VK_F);
		menu.getAccessibleContext().setAccessibleDescription("Create subwindow to display message or attachment");
		menuBar.add(menu);		// add to File menu

		JMenuItem menuItem = new JMenuItem("Print");
		menuItem.getAccessibleContext().setAccessibleDescription("Print Rules");
		menuItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_P, ActionEvent.ALT_MASK));
		menuItem.setActionCommand("print");
		menuItem.addActionListener(this);
		menu.add(menuItem);

		menu.addSeparator();

		menuItem = new JMenuItem("Exit");
		menuItem.getAccessibleContext().setAccessibleDescription("Exit Program");
		menuItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_X, ActionEvent.ALT_MASK));
		menuItem.setActionCommand("exit");
		menuItem.addActionListener(this);
		menu.add(menuItem);

		setJMenuBar(menuBar);

	}
	
	/**
	 * Creates the actual page content
	 * 
	 */
	public void buildPage()
	{
		// setLayout(new GridLayout(0,1));
        Container contentPane = getContentPane();
        SpringLayout layout = new SpringLayout();
        contentPane.setLayout(layout);
		setLayout(new SpringLayout());
		pnlTop = new JPanel();
		// pnlTop.setBackground(Color.yellow);
		pnlTop.setLayout(new BoxLayout(pnlTop, BoxLayout.PAGE_AXIS));
		JPanel pnl = new JPanel();
		JLabel lbl = new JLabel("Select validation rules - enter range of line numbers, or blank for all rules");
		pnl.add(lbl);		
		pnlTop.add(pnl);

		pnl = new JPanel();
		lbl = new JLabel("Seq from ");
		pnl.add(lbl);
		txtLowSeq = new JTextField(8);
		pnl.add(txtLowSeq);
		lbl = new JLabel(" to ");
		pnl.add(lbl);
		txtHighSeq = new JTextField(8);
		pnl.add(txtHighSeq);
		lbl = new JLabel(" Validation Table: ");
		pnl.add(lbl);
		txtTable = new JTextField(16);
		txtTable.setText(ValTbl);
		pnl.add(txtTable);
		pnlTop.add(pnl);
		
		// row of buttons for action
		pnl = new JPanel();
		// pnl.setBackground(Color.red);
		btnGet = new JButton("Get");
		btnGet.setToolTipText("Load validation rules from the database");
		btnGet.addActionListener(this);
		pnl.add(btnGet);

		btnSave = new JButton("Save");
		btnSave.setToolTipText("Save changes to the database.  (You will need to recompile)");
		btnSave.addActionListener(this);
		pnl.add(btnSave);

		btnCsv = new JButton("Csv");
		btnCsv.setToolTipText("Specify a single CSV line to be executed");
		btnCsv.addActionListener(this);
		pnl.add(btnCsv);
		
		btnFile =  new JButton("File");
		btnFile.setToolTipText("Specify a CSV file to be loaded for execution");
		btnFile.addActionListener(this);
		pnl.add(btnFile);

		btnCompile = new JButton("Compile");
		btnCompile.setToolTipText("Compile the current set of rules.");
		btnCompile.addActionListener(this);
		pnl.add(btnCompile);

		btnExecute = new JButton("Execute");
		btnExecute.setToolTipText("Execute the rules against the file (if specified) or the CSV line");
		btnExecute.addActionListener(this);
		pnl.add(btnExecute);

		btnRenum = new JButton("Renumber");
		btnRenum.setToolTipText("Renumber the statements");
		btnRenum.addActionListener(this);
		pnl.add(btnRenum);

		btnQuit = new JButton("Quit");
		btnQuit.setToolTipText("Close window and terminate program");
		btnQuit.addActionListener(this);
		pnl.add(btnQuit);

		pnlTop.add(pnl);

		pnl = new JPanel();
		// pnl.setBackground(Color.magenta);
		lblTitle = new JLabel();
		pnl.add(lblTitle);
		pnlTop.add(pnl);

		// try to prevent top panel from resizing

		/**************************
		pnlTopPage = new JPanel();
		pnlTopPage.setLayout(new BoxLayout(pnlTopPage, BoxLayout.PAGE_AXIS));
		// pnlTopPage.setBackground(Color.blue);
		pnlTopPage.add(pnlTop);
		pg.add(pnlTopPage);
		********************/
		add(pnlTop);
		layout.putConstraint(SpringLayout.WEST, pnlTop, 5, SpringLayout.WEST, contentPane);
		layout.putConstraint(SpringLayout.NORTH, pnlTop, 5, SpringLayout.NORTH, contentPane);

		// add(pnlTop);

		// table of rules
		validations = new Vector<WrpValidateCsv.ValidationRule>(); // new list
		modified = new Vector<Boolean>();	// was entry modified

		pnlScroll = new JPanel();
		// pnlScroll.setLayout(new BoxLayout(pnlScroll, BoxLayout.PAGE_AXIS));
		updateTable();

		add(pnlScroll);
		layout.putConstraint(SpringLayout.WEST, pnlScroll, 5, SpringLayout.WEST, contentPane);
		layout.putConstraint(SpringLayout.NORTH, pnlScroll, 5, SpringLayout.SOUTH, pnlTop);

		// add(pnlPage);
		
		addComponentListener(this); // monitor resize
	}
	
	/**
	 * Save any changes to the current rules to the database.
	 * </p><p>
	 * This routine loops through the validation rules
	 * and updates any modified rule to the database.
	 * It will also insert any new validation rules
	 * that were added since the previous load.
	 * It will also preserve the width of the columns in the form
	 * 
	 */
	public void updateTable()
	{ // current field changed - display new table
		
		if (tblRules!=null)
		{ // table already exists, save column widths
			for (int i=0; i<tblRules.getColumnCount(); i++)
			{
				TableColumn tc = tblRules.getColumnModel().getColumn(i);
				int w = tc.getWidth();
				colWidth[i] = w;
			}
		}
		
		lblTitle.setText("Validation Rules for " + ValTbl + ".  To delete, set seq to zero");
		RuleModel rmod = new RuleModel();
		tblRules = new JTable(rmod);
		
		// restore previous column sizes
		if (colWidth[0]>0)
		{ // not first time
			for (int i=0; i<tblRules.getColumnCount(); i++)
			{
				TableColumn tc = tblRules.getColumnModel().getColumn(i);
				tc.setPreferredWidth(colWidth[i]);
			}
		}

		// tblRules.setAutoResizeMode(JTable.AUTO_RESIZE_NEXT_COLUMN);
		scroll = new JScrollPane(tblRules);
		scroll.setHorizontalScrollBarPolicy(ScrollPaneConstants.HORIZONTAL_SCROLLBAR_ALWAYS);
		scroll.setVerticalScrollBarPolicy(ScrollPaneConstants.VERTICAL_SCROLLBAR_ALWAYS);
		pnlScroll.removeAll();
		pnlScroll.add(scroll);
		resizeScroll();
	}
	
	/**
	 * Resize the window components
	 * </p><p>
	 * This method is called when the window is resized
	 * or the list changes
	 * 
	 */
	public void resizeScroll()
	{
		Container contentPane = getContentPane();
		SpringLayout layout = (SpringLayout) contentPane.getLayout();
		layout.putConstraint(SpringLayout.WEST, pnlScroll, 5, SpringLayout.WEST, contentPane);
		layout.putConstraint(SpringLayout.NORTH, pnlScroll, 5, SpringLayout.SOUTH, pnlTop);

		Dimension gdim = getSize();	// size of window
		// adjust for top panel
		Dimension wdim = new Dimension(gdim);
		wdim.width -= 20;
		wdim.height -= 250;
		Dimension tdim = pnlTop.getSize();
		if (tdim.height==0) return;		// not created yet
		if (topSize.height==0) topSize = tdim;
		tdim.height = topSize.height;	// keep same height
		pnlTop.setPreferredSize(tdim);
		tdim = pnlTop.getSize();
		wdim.height = gdim.height - tdim.height - 75;
		wdim.width -= 10;
		scroll.setPreferredSize(wdim);
		tblRules.setSize(wdim);
		validate();
	}
	
	/**
	 * Get the name of the table containing the validation rules
	 *  
	 * @return the name of the database table
	 */
	public String getValidationTable()
	{
		return ValTbl;
	}
	
	/**
	 * Update the name of the database table containing validation rules
	 * 
	 * @param valtbl the name of the database table
	 * 
	 */
	public void setValidationTable(String valtbl)
	{
		this.ValTbl = valtbl;
	}
	
	/**
	 * Retrieve the rules for the current field from the database
	 * </p><p>
	 * This method calls resizeScroll() to resize the page if needed
	 * 
	 */
	public void loadRules()
	{
		compileErrors = -1; // reset compiled results
		if (saveChanges()) return;	// user canceled save
		String range = "";
		String s1 = txtLowSeq.getText();
		String s2 = txtHighSeq.getText();
		if (s1.length()>0 || s2.length()>0)
		{ // range entered - use it
			if (s1.length()>0 && s2.length()>0)
			{ // upper and lower range entered
				range = " where seq >= " + s1 + " and seq <= " + s2;
			}
			else if (s1.length()>0)
			{ // only lower limit entered
				range = " where seq >= " + s1;
			}
			else
			{ // only upper limit entered
				range = " where seq <= " + s2;
			}
		}
		String sTbl = txtTable.getText();
		if (sTbl!=null && sTbl.length()>0) ValTbl = sTbl;
		String query = "select * from " + ValTbl 
		+ range + " order by seq";
		// System.out.println("Query: " + query);
		try
		{
			ResultSet rs = db.executeQuery(query);
			validations = new Vector<WrpValidateCsv.ValidationRule>(); // new list
			modified = new Vector<Boolean>();	// was entry modified
			int i;
			String s;
			while (rs.next())
			{ // add entries to list
				WrpValidateCsv.ValidationRule v = Val.newRule();
				i = rs.getInt("ID");
				v.setId(i);
				i = rs.getInt("seq");
				v.setSeq(i);
				s = rs.getString("label");
				v.setLabel(s);
				s = rs.getString("invar");
				v.setInVar(s);
				s = rs.getString("cmd");
				v.setCmd(s);
				s = rs.getString("arg1");
				v.setArg1(s);
				s = rs.getString("arg2");
				v.setArg2(s);
				s = rs.getString("outvar");
				v.setOutVar(s);
				s = rs.getString("result");
				v.setResult(s);
				s = rs.getString("msg");
				v.setMsg(s);
				s = rs.getString("GoTo");
				v.setGoTo(s);
				s = rs.getString("comment");
				v.setComment(s);
				validations.add(v);	// add to list
				Boolean b = new Boolean(false);
				modified.add(b); // not modified yet
			}
			rulesRead = validations.size();	// how many rules were read
			updateTable();
		} catch (SQLException e)
		{
			System.err.println("SQL Exception retrieving validation rules" 
					+ ", e=" + e);
			e.printStackTrace();
			JOptionPane.showMessageDialog(
					this,
					"Exception encounterd when retrieving validation rules",
					"Error on Query",
					JOptionPane.ERROR_MESSAGE
					);
		}
	}
	
	/**
	 * Get the current collection of validation rules
	 * 
	 * @return list of validation rules
	 */
	public Vector<WrpValidateCsv.ValidationRule> getRules() 
	{ 
		return validations; 
	}
	
	/**
	 * Define the current collection of validation rules
	 * 
	 * @param rules a vector of validation rules
	 * 
	 */
	public void setRules(Vector<WrpValidateCsv.ValidationRule> rules)
	{
		validations = rules;
		int n = validations.size(); // number of rules
		modified = new Vector<Boolean>(n); // create modified vector
		for (int i=0; i<n; i++)
		{
			Boolean b = new Boolean(false);
			modified.set(i, b);
		}
	}
	
	/**
	 * If changes were not saved, ask user if they should be changed
	 * </p><p>
	 * This routine is called each time the user requests a save.
	 * If any changes were made but not saved,
	 * the user is asked if the changes should be save
	 * using a Confirmation Dialog box.
	 * The possible user responses are:
	 * </p>
	 * <ul>
	 * <li>Yes - save the changes and get the requested rules</li>
	 * <li>No - don't save the changes and get the requested rules</li>
	 * <li>Cancel - don't perform the get</li>
	 * </ul>
	 * 
	 * @return true if the Get request should be canceled; otherwise false.
	 * 
	 */
	public boolean saveChanges()
	{
		boolean ret = false;	// normal return - do save
		
		/**
		 * Loop through the modified vector looking for any rule
		 * that was changed.
		 * When the first change is encountered, the user is asked
		 * if they want to save their changes before the new
		 * rules are loaded.
		 * 
		 */
		for (int i=0; i<modified.size(); i++)
		{ // see if any fields have been modified
			if (modified.get(i))
			{ // something modified - ask user if we should save
				/**
				 * If the user answers "Yes", then the changes are saved.
				 * If the user answers "No", the changes are discarded.
				 * If the user answers "Cancel", the Get request is cancelled.
				 * 
				 */
				int opt = JOptionPane.showConfirmDialog(this, "Save changes?");
				if (opt == JOptionPane.YES_OPTION)
				{ // save before get
					saveRules();
				} else if (opt==JOptionPane.CANCEL_OPTION)
				{ // don't do save
					ret = true;
				}
				// if user clicked 'No', then nothing to do
				break;
			}
		}
		return ret;	// normal return - do save
	}
	
	/**
	 * Save any new or changed rules
	 * 
	 */
	public void saveRules()
	{
		int nrules = validations.size();
		if (nrules<=0) return;	// nothing to save
		int i=-1;
		int nupd = 0;
		try
		{ // update or insert rules as needed
			for (i=0; i<rulesRead; i++)
			{ // update any modified rule
				if (!modified.get(i)) continue; // not modified
				WrpValidateCsv.ValidationRule v = validations.get(i);	// get entry
				int num = v.getSeq();	// see if deleted
				if (num<=0) deleteRule(v);
				else updateRule(v);
				nupd++;
				modified.set(i, false);	// reset modified flag
			}
			
			// insert any new rules
			for (i=rulesRead; i<nrules; i++)
			{
				WrpValidateCsv.ValidationRule v = validations.get(i);
				insertRule(v);
				nupd++;
				modified.set(i, false);	// reset modified flag
			}
			rulesRead = nrules;	// reset how many read
			JOptionPane.showMessageDialog(
					this,
					nupd + " rules saved",
					"Update complete",
					JOptionPane.INFORMATION_MESSAGE
					);
		} catch (SQLException e)
		{
			System.err.println("SQLException saving rule, e=[" + e + "]");
			e.printStackTrace();
			JOptionPane.showMessageDialog(
					this,
					"Error when updating rules",
					"Update Error",
					JOptionPane.ERROR_MESSAGE
					);
		}
	}
	
	/**
	 * Insert a new rule into the database
	 * 
	 * @param v the valiation rule to be inserted
	 * @throws SQLException
	 */
	public void insertRule(WrpValidateCsv.ValidationRule v) throws SQLException
	{ // add new rule
		StringBuffer sb = new StringBuffer(
				"insert into " + ValTbl 
				+ " (seq, label, invar, cmd, arg1, arg2, "
				+ "outvar, result, msg, goto, comment) values (");
		int num = v.getSeq();
		sb.append(num);
		String s = v.getLabel();
		sb.append(",'" + s + "'");
		s = v.getInVar();
		sb.append(",'" + s + "'");
		s = v.getCmd();
		sb.append(",'" + s + "'");
		s = v.getArg1();
		sb.append(",'" + s + "'");
		s = v.getArg2();
		sb.append(",'" + s + "'");
		s = v.getOutVar();
		sb.append(",'" + s + "'");
		s = v.getResult();
		sb.append(",'" + s + "'");
		s = db.sqlSafe(v.getMsg());
		sb.append(",'" + s + "'");
		s = v.getGoTo();
		sb.append(",'" + s + "'");
		s = db.sqlSafe(v.getComment());
		sb.append(",'" + s + "'");
		sb.append(")");
		String query = sb.toString();
		// System.out.println("Query=[" + query + "]");
		db.executeUpdate(query);
	}
	
	/**
	 * Update the specified validation rules
	 * </p><p>
	 * The unique row identifier (primary key) 
	 * from the database is used
	 * to update this specific rule in the database.
	 *  
	 * @param v the validation rule to be updated
	 * @throws SQLException
	 */
	public void updateRule(WrpValidateCsv.ValidationRule v) throws SQLException
	{ // update existing rule
		StringBuffer sb = new StringBuffer(
				"update " + ValTbl);
		int num = v.getSeq();
		sb.append(" set seq=" + num);
		String s = v.getLabel();
		sb.append(", label='" + s + "'");
		s = v.getInVar();
		sb.append(", invar='" + s + "'");
		s = v.getCmd();
		sb.append(", cmd='" + s + "'");
		s = v.getArg1();
		sb.append(", arg1='" + s + "'");
		s = v.getArg2();
		sb.append(", arg2='" + s + "'");
		s = v.getOutVar();
		sb.append(", outvar='" + s + "'");
		s = v.getResult();
		sb.append(", result='" + s + "'");
		s = db.sqlSafe(v.getMsg());
		sb.append(", msg='" + s + "'");
		s = v.getGoTo();
		sb.append(", goto='" + s + "'");
		s = db.sqlSafe(v.getComment());
		sb.append(", comment='" + s + "'");
		num = v.getId();
		sb.append(" where id=" + num);
		
		String query = sb.toString();
		// System.out.println("Query=[" + query + "]");
		db.executeUpdate(query);
	}
	
	/**
	 * Delete the specified validation rule from the database
	 * 
	 * @param v the validation rule to be deleted
	 * @throws SQLException
	 */
	public void deleteRule(WrpValidateCsv.ValidationRule v) throws SQLException
	{ // delete existing rule
		int num = v.getId();
		String query = "delete from " + ValTbl + " where id=" + num;
		// System.out.println("Query=[" + query + "]");
		db.executeUpdate(query);
	}
	
	/**
	 * Internal class that is the model for the rules table
	 * 
	 * @author pringlb
	 *
	 */
	private class RuleModel extends AbstractTableModel
	{
		final static long serialVersionUID = 1;

		private String[] colNames = {
				"Seq", "Label", "InVar", "Cmd", 
				"arg1", "arg2", "OutVar", "Result",
				"Msg", "GoTo", "Comment"};
		private WrpValidateCsv.ValidationRule newValidation = Val.newRule(); // used for inserts
		
		/**
		 * Default constructor
		 * 
		 */
		public RuleModel()
		{
			
		}
		
		public int getColumnCount() { return 11; }
		public int getRowCount() { return validations.size() + 1; }
		public String getColumnName(int col) { return colNames[col]; }
		public boolean isCellEditable(int row, int col) { return true; }
		
		public Class<?> getColumnClass(int col)
		{
			if (col==0)
			{ // sequence number
				return new Integer(0).getClass();
			}
			else if (col==1)
			{ // statement label
				return new String().getClass();
			}
			else if (col==2)
			{ // input variable
				return new String().getClass();
			}
			else if (col==3)
			{ // command
				return new String().getClass();
			}
			else if (col==4)
			{ // argument #1
				return new String().getClass();
			}
			else if (col==5)
			{ // argument #2
				return new String().getClass();
			}
			else if (col==6)
			{ // output variable
				return new String().getClass();
			}
			else if (col==7)
			{ // validation results
				return new String().getClass();
			}
			else if (col==8)
			{ // output message
				return new String().getClass();
			}
			else if (col==9)
			{ // GoTo destination
				return new String().getClass();
			}
			else if (col==10)
			{ // comment
				return new String().getClass();
			}
			else return null;
		}
		
		public Object getValueAt(int row, int col)
		{
			WrpValidateCsv.ValidationRule v; // point to requested validation
			if (row<validations.size()) v=validations.get(row);
			else 
			{ // pointing to blank entry
				v = newValidation; // point to expansion entry
			}
				
			// return the attribute associated with the requested column
			if (col==0)
			{ // sequence number
				int seq = v.getSeq();
				return seq;
			}
			else if (col==1)
			{ // label
				String fps = v.getLabel();
				if (fps==null) fps="";
				return fps;
			}
			else if (col==2)
			{ // start of substring (if used)
				String invar = v.getInVar();
				if (invar==null) invar="";
				return invar;
			}
			else if (col==3)
			{ // command
				String cmd = v.getCmd();
				if (cmd==null) cmd="";
				return cmd;
			}
			else if (col==4)
			{ // end of substring (if used)
				String arg1 = v.getArg1();
				if (arg1==null) arg1="";
				return arg1;
			}
			else if (col==5)
			{ // command
				String arg2 = v.getArg2();
				if (arg2==null) arg2="";
				return arg2;
			}
			else if (col==6)
			{ // output value
				String outvar = v.getOutVar();
				if (outvar==null) outvar="";
				return outvar;
			}
			else if (col==7)
			{ // result string
				String res = v.getResult();
				if (res==null) res="";
				return res;
			}
			else if (col==8)
			{ // message to be returned
				String msg = v.getMsg();
				if (msg==null) msg="";
				return msg;
			}
			else if (col==9)
			{ // goto destination
				String dest = v.getGoTo();
				if (dest==null) dest="";
				return dest;
			}
			else if (col==10)
			{ // comment
				String cmt = v.getComment();
				if (cmt==null) cmt="";
				return cmt;
			}
			else return null; // should never happen
		}
		
		public void setValueAt(Object o, int row, int col)
		{
			WrpValidateCsv.ValidationRule v;
			if (row<validations.size())
			{ // existing entry being modified
				 v = validations.get(row);
				 modified.set(row, true);
			}
			else 
			{ // creating new entry
				v = newValidation;
				validations.add(v); // add it to list
				newValidation = Val.newRule(); //new expansion
				Boolean b = new Boolean(true);
				modified.add(b);
			}

			if (col==0)
			{ // sequence number
				int seq;
				if (o==null) seq=-1;
				else seq = (Integer)o;
				v.setSeq(seq);
			}
			else if (col==1)
			{ // statement label
				String s = (String)o;
				if (s==null) s ="";
				v.setLabel(s);
			}
			else if (col==2)
			{ // input variable
				String s = (String)o;
				if (s==null) s = "";
				v.setInVar(s);
			}
			else if (col==3)
			{ // command
				String s = (String)o;
				if (s==null) s ="";
				v.setCmd(s);
			}
			else if (col==4)
			{ // arg1
				String s = (String)o;
				if (s==null) s = "";
				v.setArg1(s);
			}
			else if (col==5)
			{ // arg2
				String s = (String)o;
				if (s==null) s = "";
				v.setArg2(s);
			}
			else if (col==6)
			{ // output variable
				String s = (String)o;
				if (s==null) s = "";
				v.setOutVar(s);
			}
			else if (col==7)
			{ // validation result
				String s = (String)o;
				if (s==null) s = "";
				v.setResult(s);
			}
			else if (col==8)
			{ // output message
				String s = (String)o;
				if (s==null) s = "";
				v.setMsg(s);
			}
			else if (col==9)
			{ // goto destination
				String s = (String)o;
				if (s==null) s = "";
				v.setGoTo(s);
			}
			else if (col==10)
			{ // comment
				String s = (String)o;
				if (s==null) s = "";
				v.setComment(s);
			}
			else
			{
				System.err.println("setValueAt - invalid column [" + col + "]");
			}
		}
	}
	
	/**
	 * Select a CSV file to load for validation.
	 * </p><p>
	 * This method uses an input dialog box to obtain a CSV line
	 * which can be parsed and validated when the user
	 * clicks on the "Execute" button.
	 * 
	 */
	public void loadCsv(int CsvLine)
	{
		String str = JOptionPane.showInputDialog(this, "Enter CSV line to validate");
		if (str!=null)
		{
			csv = new WrpCsv(str);
			Val.setCsv(csv);
			Val.setCsvLine(CsvLine++);
		}
	}
	
	/**
	 * Select the CSV file to be validated.
	 * </p><p>
	 * This method uses a file chooser to obtain a CSV line
	 * which can be parsed and validated when the user
	 * clicks on the "Execute" button.
	 * 
	 */
	public void selectOpenFile()
	{
		csvFile = null;
		int ret = fc.showOpenDialog(this);
		if (ret == JFileChooser.APPROVE_OPTION)
		{ // file selected
			csvFile = fc.getSelectedFile();
		}
	}
	
	protected File selectSaveFile()
	{
		File f = null;
		int ret = fc.showSaveDialog(this);
		if (ret == JFileChooser.APPROVE_OPTION)
		{ // file selected
			f = fc.getSelectedFile();
		}
		return f;
	}
	
	/**
	 * Renumber the statements.
	 * </p><p>
	 * This method is called if the user clicks on the
	 * "Renumber" button.
	 * It will ask the user for an increment, and then
	 * renumber all of the statements using that increment.
	 * 
	 */
	public void renumberStatements()
	{
		String str = JOptionPane.showInputDialog(this, "Enter increment value, or -1 to cancel");
		if (str!=null)
		{
			try
			{
				int incr = Integer.parseInt(str);
				if (incr <1)
				{
					JOptionPane.showMessageDialog(this, "Increment must be positive", "Request Ignored", JOptionPane.ERROR_MESSAGE);
				}
				else
				{
					renumberStatements(incr);
				}
			} catch (NumberFormatException e)
			{
				JOptionPane.showMessageDialog(this, "Invalid increment value");
			}
		}
		
	}
	
	/**
	 * Renumber statements using the specified increment.
	 * 
	 * @param incr increment value for renumbering statements
	 */
	private void renumberStatements(int incr)
	{
		int num = incr; // starting number
		for (int i=0; i<validations.size(); i++)
		{ // renumber statements
			WrpValidateCsv.ValidationRule rule = validations.get(i); // get rule
			String lbl = rule.getLabel();
			if (lbl!=null && lbl.length()>0) num += incr; // extra incr before labels
			rule.setSeq(num); // change sequence number
			num += incr; // increment to next sequence number
			modified.set(i, true); // mark as modified
		}
		JOptionPane.showMessageDialog(this, "Last sequence number is " + num + ".  Don't forget to Save Rules");
	}
	
	/**
	 * Compile current set of rules, and display results.
	 * </p><p>
	 * If any error messages were generated, display them as well.
	 * 
	 */
	public void compileRules()
	{
		Val.setRules(validations); // set current rules
		compileErrors = Val.compileRules();
		if (compileErrors==0)
		{
			JOptionPane.showMessageDialog(this, "No compile errors!");
		}
		else
		{
			WrpScrollFrame frame = new WrpScrollFrame();
			frame.setTitle("Error messages for validation rules");
			Vector<String> msgs = Val.getErrMsgs();
			frame.setScrollData(msgs);
			frame.toFront();
			frame.setVisible(true);
		}
	}
	
	/**
	 * Execute the rules using the specified CSV string or file.
	 * </p><p>
	 * If the rules haven't been compiled yet,
	 * or any compile-time errors were found,
	 * then nothing is done.
	 * </p><p>
	 * Assuming the rules were compiled without errors,
	 * then either the CSV string or the CSV file are
	 * validated.
	 * If a CSV has been specified, then it is validated;
	 * otherwise the specified CSV string is validated.
	 * If neither were specified, an error message is displayed.
	 * </p><p>
	 * If any run-time errors are found, report them
	 * using a wrpScrollFrame
	 * 
	 */
	public void executeRules() throws ValidationRuntimeException
	{
		if (compileErrors<0)
		{
			JOptionPane.showMessageDialog(this, "You must first compile the rules before executing them");
		}
		else if (compileErrors>0)
		{
			JOptionPane.showMessageDialog(this, "Can't execute with compilation errors");
		}
		else if (csvFile!=null)
		{ // run validation against entire CSV file
			validateFile(csvFile);
		}
		else if (csv==null)
		{
			JOptionPane.showMessageDialog(this, "You must specify a string or file to validate");
		}
		else
		{
			int ret = Val.validate(csvLine++);
			int stat = Val.getStat();
			String s = WrpValidateCsv.recStat(stat);
			String msg;
			if (stat<=WrpValidateCsv.REC_WARN) 
			{
				msg = "\nNormalized string: " + Val.getNormString();
			}
			else
			{
				msg = "";
			}
			JOptionPane.showMessageDialog(this, "Found " + ret + " run-time errors.  Record status is " + s
					+ msg);
			Vector<String> msgs = Val.getErrMsgs(); // get any messages
			if (msgs.size()> 0)
			{
				WrpScrollFrame frame = new WrpScrollFrame();
				frame.setTitle("Validation Execution Error Messages");
				Vector<String> errMsgs = Val.getErrMsgs();
				frame.setScrollData(errMsgs);
				frame.toFront();
				frame.setVisible(true);
			}
		}
	}
	
	/**
	 * Validate a CSV file
	 * </p><p>
	 * Read each line from the file and run the validation commands
	 * against each line.
	 * 
	 * @param f the CSV file to be validated
	 */
	public void validateFile(File f)
	{
		BufferedReader br=null;
		try {
			br = new BufferedReader(new FileReader(f));
		} catch (FileNotFoundException e) 
		{
			JOptionPane.showMessageDialog(this, "File not found!");
		}
		Vector<String> msgs = new Vector<String>();
		int fileStat = WrpValidateCsv.REC_GOOD; // assume file is good
		String st;
		int lnum = 0;
		int nerr = 0;
		try {
			while ((st = br.readLine())!=null)
			{
				if (st.length()==0) continue; // ignore blank lines
				csv = new WrpCsv(st);
				Val.setCsv(csv);
				Val.setCsvLine(++lnum);
				// Val.executeRules();
				Val.validate();
				int stat = Val.getStat(); // record status
				if (stat>fileStat) fileStat=stat;
				int n = Val.getRunErrors(); // number of run-time errors
				nerr += n; // update number of errors
				Vector<String> m = Val.getErrMsgs(); // get any messages
				if (m.size() > 0)
				{ // messages for record - append to file messages
					if (msgs.size()>0) msgs.add("==============="); // separate from previous messages
					for (int i=0; i<m.size(); i++)
					{ // append record messages to file messages
						String s = m.get(i);
						msgs.add(s);
					}
				}
			}
			// file read and validated - display results
			String msg = nerr + " errors encounted.  File status is " + WrpValidateCsv.recStat(fileStat);
			JOptionPane.showMessageDialog(this, msg);
			if (msgs.size()>0)
			{ // display error messages in a scroll frame
				WrpScrollFrame frame = new WrpScrollFrame();
				frame.setTitle("Validation Execution Error Messages for file [" 
						+ f.getName() + "]");
				frame.setScrollData(msgs);
				frame.toFront();
				frame.setVisible(true);
			}
		} catch (ValidationRuntimeException e) {
			JOptionPane.showMessageDialog(this, "I/O error while reading file");
			e.printStackTrace();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

	}

	/**
	 * Allow the user to print the current rules.
	 * This method will create an HTML document containing
	 * the current rules formatted in an HTML table.
	 * 
	 */
	public void printRules()
	{
		File f = selectSaveFile(); // get the output file
		if (f==null) return; // no file specified
		try {
			PrintStream prt = new PrintStream(f);
			buildHtmlHeader(prt); // start off the HTML file
			for (int i=0; i<validations.size(); i++)
			{
				WrpValidateCsv.ValidationRule v = validations.get(i);	// get entry
				printHtmlRule(v, prt); // print rule within table
			}
			buildHtmlFooter(prt); // finish up the HTML file
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	
	/**
	 * Build the beginning of an HTML file for printing rules.
	 * 
	 * @param prt the output stream for the HTML file
	 * 
	 */
	public void buildHtmlHeader(PrintStream prt)
	{
		prt.println("<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">");
		prt.println("<html>");
		prt.println("<head>");
		prt.println("<meta http-equiv=\"Content-Type\" content=\"text/html; charset=iso-8859-1\" />");
		prt.println("</head>");
		prt.println("<body>");
		prt.println("<h1>Validation Rules for " + txtTable.getText() 
				+ " on " + new Date()
				+ "</h1>");
		prt.println("<table border='1' cellspacing='0'>");
		prt.println("<tr><th>Seq</th><th>Label</th><th>InVar</th><th>Cmd</th>"
				+ "<th>arg1</th><th>arg2</th><th>OutVar</th>"
				+ "<th>Result</th><th width='30%'>Message</th><th>GoTo</th><th width='20%>Comment</th></tr>"
				);

	}

	/**
	 * Create the bottom part of the HTML rules file.
	 * 
	 * @param prt the output stream for the HTML rules file
	 * 
	 */
	public void buildHtmlFooter(PrintStream prt)
	{
		prt.println("</table>");
		prt.println("</body>");
		prt.println("</html>");
	}
	
	/**
	 * Format a rule for insertion into the HTML table.
	 * 
	 * @param v the validation rule to be printed in a table
	 * @param prt the output stream for the HTML rules table
	 * 
	 */
	public void printHtmlRule(WrpValidateCsv.ValidationRule v, PrintStream prt)
	{
		prt.println("<tr><td>" + v.getSeq() + "</td><td>"
				+ nbsp(v.getLabel()) + "</td><td>"
				+ nbsp(v.getInVar()) + "</td><td>"
				+ nbsp(v.getCmd()) + "</td><td>"
				+ nbsp(v.getArg1()) + "</td><td>"
				+ nbsp(v.getArg2()) + "</td><td>"
				+ nbsp(v.getOutVar()) + "</td><td>"
				+ nbsp(v.getResult()) + "</td><td>"
				+ nbsp(v.getMsg()) + "</td><td>"
				+ nbsp(v.getGoTo()) + "</td><td>"
				+ nbsp(v.getComment()) + "</td></tr>"
				);
	}
	
	/**
	 * Substitute empty strings with &amp;nbsp;
	 * This method is used when filling data cells within a column.
	 * If an empty string appears, then the table cell can look strange,
	 * but if it has a non-breaking space character, it looks just fine.
	 * 
	 * @param str the string to test
	 * @return returns the string, or NBSP if null or an empty string
	 */
	public String nbsp(String str)
	{
		String ret = str;
		if (ret==null) ret = "";
		if (ret.length() > 0) return ret;
		return "&nbsp";
	}
	
	//
	// Listener routines
	//
	
	// ActionListener
	public void actionPerformed(ActionEvent e)
	{
		String cmd = e.getActionCommand().toLowerCase();
		if (cmd.equals("get"))
		{
			loadRules();
		}
		else if (cmd.equals("save"))
		{
			saveRules();
		}
		else if (cmd.equals("compile"))
		{
			compileRules();
		}
		else if (cmd.equals("csv"))
		{ // specify CSV line (turn off CSV file, if any)
			loadCsv(csvLine++);
			csvFile = null;
		}
		else if (cmd.equals("file"))
		{ // specify CSV file (turn off CSV line)
			selectOpenFile();
		}
		else if (cmd.equals("execute"))
		{
			executeRules();
		}
		else if (cmd.equals("renumber"))
		{
			renumberStatements();
		}
		else if (cmd.equals("print"))
		{
			printRules();
		}
		else if (cmd.equals("quit") || cmd.equals("exit"))
		{
			System.exit(0);
		}
		else
		{
			System.err.println("Invalid action command [" + cmd + "]");
		}
	}
	
	// mouse listener routines
	
	public void mouseClicked(MouseEvent e)
	{
		if (e.getClickCount() == 2)
		{ // double clicked on list - display that message
			// viewValidations();	// display selected messages
		}
	}
	
	public void mouseEntered(MouseEvent e)
	{
		
	}
	
	public void mouseExited(MouseEvent e)
	{
		
	}
	
	public void mousePressed(MouseEvent e)
	{
		ifPopup(e);
	}
	
	public void mouseReleased(MouseEvent e)
	{
		ifPopup(e);
	}
	
	public void ifPopup(MouseEvent e)
	{
		if (e.isPopupTrigger()) 
		{
            popup.show(e.getComponent(),
                       e.getX(), e.getY());
		}
	}
	
	// ComponentListener routines
    public void componentHidden(ComponentEvent e)
    {
    }
    
    public void componentMoved(ComponentEvent e)
    {
    }
    
    public void componentResized(ComponentEvent e)
    {
    	resizeScroll();
    }
    
    public void componentShown(ComponentEvent e)
    {
    }
    
    /**
     * This is a sample driver for this program.  
     * It is expected that the user will create their own main program,
     * using their own database rather than the example UtilDb.
     * 
     * @param args not used
     */
    public static void main(String args[]) {
    	if (args.length==0)
    	{
            java.awt.EventQueue.invokeLater(new Runnable() {
                public void run() {
                	try
					{
						UtilDb db = new UtilDb();
						new WrpEditValidationRules(db).setVisible(true);
					} catch (ClassNotFoundException e)
					{
						System.err.println("Class not Found [" + e.getMessage() + "]");
						e.printStackTrace();
					} catch (IllegalAccessException e)
					{
						System.err.println("Illegal Access Exception [" + e.getMessage() + "]");
						e.printStackTrace();
					} catch (InstantiationException e)
					{
						System.err.println("Instatiation Exception [" + e.getMessage() +"]");
						e.printStackTrace();
					} catch (SQLException e)
					{
						System.err.println("SQL Exception [" + e.getMessage() + "]");
						e.printStackTrace();
					}
                }
            });
    	}
    	else
    	{
    		System.err.println("No command arguments accepted!");
    		System.exit(1);
    	}
    }

}
