package com.billpringle.utils.wrputils;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;

/**
 * This class encapsulates a record using fixed length fields.
 * The WrpCSV class handles CSV formatted fields, while this class
 * handles files with fixed length columns. 
 * 
  <table border="0" cellspacing="0" cellpadding="6" align="center"
  summary="Creative Commons license">
    <tr>
      <td><!-- Creative Commons License -->
      <a rel="license" href=
      "http://creativecommons.org/licenses/by-nc-sa/2.0/" target=
      "CC"><img alt="Creative Commons License" border="0" src=
      "http://creativecommons.org/images/public/somerights20.gif" /></a></td>

      <td><a rel="license" href=
      "http://creativecommons.org/licenses/by-nc-sa/2.0/" target=
      "CC"><img alt="Creative Commons License Symbols" border="0"
      src="cc_lic.gif" /></a> <!-- /Creative Commons License --></td>

      <td>Unless noted otherwise, all materials available for
      download from my site are copyrighted by Bill Pringle, and
      are licensed under a <a rel="license" href=
      "http://creativecommons.org/licenses/by-nc-sa/2.0/" target=
      "CC">Creative Commons License</a>.</td>
    </tr>
  </table>
  * @author Bill Pringle
 *
 */
public class WrpFixedCols
{
	/** left alignment flag */
	public static final int ALIGN_LEFT = 1;
	/** right alignment flag */
	public static final int ALIGN_RIGHT = 2;
	/** center alignment flag */
	public static final int ALIGN_CENTER = 3;
	
	/** string buffer containing record data */
	private StringBuffer sbuf;
	/** alignment flag */
	private int alignmentFlag = 1;
	/** padding character */
	private char padChar = ' ';
	
	/**
	 * Create empty fixed record of a given size.
	 * 
	 * @param recsize the record size
	 */
	public WrpFixedCols(int recsize)
	{
		init(recsize);
	}
	
	/**
	 * Create fixed size record and initialize to the specified string.
	 * 
	 * @param str the record data
	 */
	public WrpFixedCols(String str)
	{
		sbuf = new StringBuffer(str);
	}
	
	/**
	 * Create new record and set alignment and padding.
	 * This method creates a new record of the specified size,
	 * alignment, and padding character.
	 * 
	 * @param recsize the size of the record
	 * @param alignFlg the alignment for new fields
	 * @param padchar the padding character
	 */
	public WrpFixedCols(int recsize, int alignFlg, char padchar)
	{
		this.alignmentFlag = alignFlg;
		this.padChar = padchar;
		init(recsize);
	}
	
	/**
	 * Initialize an empty WrpFixedCols object.
	 * 
	 * @param recsize the record size of the object
	 */
	private void init(int recsize)
	{
		sbuf = new StringBuffer(recsize);
		for (int i=0; i<recsize; i++)
		{
			sbuf.append(padChar); // initialize with all blanks
		}
	}
	
	/**
	 * Reset (clear) the record.
	 * This method resets the record to all blanks.
	 * Any previous information about the record is lost.
	 * 
	 * @param recsize the size of the record
	 */
	public void reset(int recsize)
	{
		init(recsize);
	}
	
	/**
	 * Reset (clear) record and initialize it with the specified string.
	 * Any previous information about the record is lost.
	 * 
	 * @param str the data for the record
	 */
	public void reset(String str)
	{
		sbuf = new StringBuffer(str);
	}
	
	/**
	 * Return substring of data record.
	 * 
	 * @param start first column of field (zero based)
	 * @param end first column past end of field (zero based)
	 * @return the substring from column start through column end-1
	 */
	public String substring(int start, int end)
	{
		if (start>=end) return null;
		if (end>sbuf.length()) return null;
		String ret = sbuf.substring(start, end);
		return ret;
	}
	
	/**
	 * Replace the existing substring from start to end with
	 * the specified string.
	 * The replacing string must be the same size as the substring
	 * it replaces.
	 * 
	 * @param start start of substring (zero based)
	 * @param end the first column past the end of the field (zero based)
	 * @param str the replacement string
	 * @return the string that was replaced, or null on error
	 */
	public String replace(int start, int end, String str)
	{
		if (start>=end) return null; // arguments in wrong order
		if (end>sbuf.length()) return null; // past end of record
		if (str.length()!=(end-start)) return null; // wrong size
		String ret = sbuf.substring(start, end); // what will be returned
		StringBuffer sb = sbuf.replace(start, end, str); // replace substring
		if (sb != sbuf) sbuf = sb; // should never happen
		return ret;		
	}
	
	/**
	 * Create a fixed-length field.
	 * 
	 * @param fldSize size of field
	 * @param str initial value of the field
	 * @param alignFlg alignment flag
	 * @param padchar padding character, in case string is shorter than fldSize
	 * @return a string as specified by the above parameters
	 */
	public static String buildField(int fldSize, String str, int alignFlg, char padchar)
	{
		StringBuffer ret = new StringBuffer();
		int strLen = str.length(); // length of field
		if (strLen == fldSize)
		{ // no padding needed
			ret.append(str);
		}
		else if (strLen > fldSize)
		{ // too long - truncate
			ret.append(str.substring(0, fldSize));
		}
		else
		{ // too short - add padding
			int padLen = fldSize - strLen; // amount of padding
			
			if (alignFlg == ALIGN_RIGHT)
			{ // add pad character before string value
				for (int i=0;i<padLen; i++)
				{
					ret.append(padchar);
				}
				ret.append(str);
			}
			else if (alignFlg == ALIGN_CENTER)
			{ // add padding before and after string
				int p1 = padLen / 2; // padding before
				int p2 = padLen - p1; // padding after
				for (int i=0; i<p1; i++)
				{
					ret.append(padchar);
				}
				ret.append(str);
				for (int i=0; i<p2; i++)
				{
					ret.append(padchar);
				}
			}
			else
			{ // align left 
				ret.append(str); // start with string
				for (int i=0; i<padLen; i++)
				{ // add padding character
					ret.append(padchar);
				}
			}
		}
		
		return ret.toString();
	}
	
	/**
	 * Build fixed length field, using alignment and padding defaults.
	 * This method will create a fixed length field of the size
	 * and content as specified, with the default alignment and padding.
	 * 
	 * @param fldSize
	 * @param str
	 * @return a string of the specified length and content
	 */
	public String buildField(int fldSize, String str)
	{
		return buildField(fldSize, str, alignmentFlag, padChar);
	}
	
	/**
	 * Build left aligned field.
	 * The field will be padded with spaces, if needed.
	 * @param fldSize size of field
	 * @param str string value of field
	 * @return a left aligned string (space padded)
	 */
	public String buildFieldLeft(int fldSize, String str)
	{
		return buildField(fldSize, str, ALIGN_LEFT, ' ');
	}

	/**
	 * Build right aligned field.
	 * The field will be padded with zeros, if needed.
	 * 
	 * @param fldSize size of field
	 * @param str string value of field
	 * @return a right aligned string (zero padded)
	 */
	public String buildFieldRight(int fldSize, String str)
	{
		return buildField(fldSize, str, ALIGN_RIGHT, '0');
	}
	
	/**
	 * Set the value of a field and return the old value.
	 * This method assigns a new value to the specified field
	 * and returns the old value.
	 * 
	 * @param beg index of start of field
	 * @param end index of first character past the field
	 * @param fld new value for field
	 * @param alignFld alignment flag
	 * @param padchar padding character
	 * @return the old value of the field
	 */
	public String setField(int beg, int end, String fld, int alignFld, char padchar)
	{
		if (beg<0 || end < 1) return null; // invalid arguments
		int fldSize = end - beg;
		if (fldSize < 0) return null;
		String f = buildField(fldSize, fld, alignFld, padchar);
		String ret = replace(beg, end, f);
		return ret; // the old value of the field
	}
	
	/**
	 * Set new value for field, returning the old value.
	 * This method assigns a new value to the specified field
	 * and returns the old value.
	 * The default alignment and padding will be used.
	 * 
	 * @param beg index of start of field
	 * @param end index of first character past the field
	 * @param fld new value for field
	 * @return the old value of the field
	 */
	public String setField(int beg, int end, String fld)
	{
		return setField(beg, end, fld, alignmentFlag, padChar);
	}

	/**
	 * Get specified field.
	 * This method returns the specified substring.
	 * This method is essentially the same as the substring method,
	 * but the terminology better suits the concept of fields.
	 * 
	 * @param beg first column of field (zero based)
	 * @param end first column past end of field (zero based)
	 * @return the substring from column start through column end-1
	 */
	public String getField(int beg, int end)
	{
		return substring(beg, end);
	}
	
	/**
	 * Return the length of the record.
	 * 
	 * @return the length of the record
	 */
	public int length()
	{
		return sbuf.length();
	}
	
	/**
	 * Get the numeric alignment flag from the string value.
	 * 
	 * @param flg string containing 'L', 'R', or 'C'
	 * @return the numeric equivalence of the alignment string
	 */
	public static int getAlignmentFlag(String flg)
	{
		int ret = ALIGN_LEFT;
		char alignChar = 'L';

		if (flg!=null)
		{
			String s = flg.toUpperCase();
			alignChar = s.charAt(0);
		}

		switch(alignChar)
		{
		case 'L': ret = ALIGN_LEFT; break;
		case 'R': ret = ALIGN_RIGHT; break;
		case 'C': ret = ALIGN_CENTER; break;
		}

		return ret;
	}
	
	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	public String toString()
	{
		return sbuf.toString();
	}
	
	/**
	 * Dump contents of record.
	 * White space characters are replaced by up arrows (^).
	 * 
	 * @return the record with white space translated to up arrows
	 */
	public String dump()
	{
		return dump(sbuf.toString());
	}
	
	/**
	 * Dump the specified string.
	 * Substitute  up arrows (^) for white space.
	 * 
	 * @param str string to dump
	 * @return the string with white space translated to up arrows
	 */
	public String dump(String str)
	{
		if (str==null) return "[null]";
		StringBuffer sb = new StringBuffer();
		for (int i=0; i<str.length(); i++)
		{
			char c = str.charAt(i);
			if (Character.isWhitespace(c)) c = '^';
			sb.append(c);
		}
		return sb.toString();
	}
	
    /**
     * Get the default alignment flag.
     * 
	 * @return the alignmentFlag
	 */
	public int getAlignmentFlag()
	{
		return alignmentFlag;
	}

	/**
	 * Set the default alignment flag.
	 * 
	 * @param alignmentFlag the alignmentFlag to set
	 */
	public void setAlignmentFlag(int alignmentFlag)
	{
		this.alignmentFlag = alignmentFlag;
	}

	/**
	 * Get the default padding character.
	 * 
	 * @return the padChar
	 */
	public char getPadChar()
	{
		return padChar;
	}

	/**
	 * Set the default padding character.
	 * 
	 * @param padChar the padChar to set
	 */
	public void setPadChar(char padChar)
	{
		this.padChar = padChar;
	}

	/**
     * Test driver.
     * 
     * @param args not used
     */
    public static void main(String args[])
    {
    	/** input stream reader */
        InputStreamReader rdr = new InputStreamReader(System.in);
        /** input keyboard reader */
        BufferedReader kbd = new BufferedReader(rdr);
        /** work string */
        String str;
        /** record size */
        int recsize;
        /** exit status */
        int status = -1; // should never happen
        /** work object */
        WrpFixedCols cols=null;
        /** help string for initializing record */
        String helpInit = "When asked for record size, you may enter:\n\t= the size of the record,\n\t= zero to initialize the record to string,\n\t=or a negative number to quit";
        /** help string for field commands */
        String helpFields = "The following commands are available:\n\t> set|beg|end|str|\n\t> get|beg|end|\n\t> align|L, R, or C|\n\t> pad|pad char|\n\t> dump\n\t> reset\n\t> quit";
        /** user command */
        String cmd = "";
        /** work CSV object */
		WrpCsv csv = new WrpCsv("","|"); // for parsing user commands
    	
    	/*
    	 * Loop creating records
    	 */
    	System.out.println(helpInit);
        while (true)
        {
        	try
			{
				System.out.print("Record size: ");
				str = kbd.readLine();
				if (str.startsWith("?"))
				{
					System.out.println(helpInit);
					continue;
				}
				recsize = Integer.parseInt(str); // convert to integer
				if (recsize < 0)
				{ // quit
					status = 0;
					break;
				}
				else if (recsize==0)
				{
					System.out.println("Enter initial string:");
					str = kbd.readLine();
					cols = new WrpFixedCols(str);
				}
				else if (recsize > 0)
				{
					cols = new WrpFixedCols(recsize);
				}
				
				if (cols==null)
				{
					System.err.println("Internal error: cols not initialized");
					status = 2;
					break;
				}
				/*
				 * Loop manipulating record
				 */
				System.out.println(helpFields);
				while (true)
				{
					System.out.println("Enter command: ");
					str = kbd.readLine();
					csv = new WrpCsv(str + "||||", "|"); // make sure enough fields
					cmd = csv.getField(0).toLowerCase();
					if (cmd.startsWith("?"))
					{
						System.out.println(helpFields);
					}
					else if (cmd.startsWith("pad"))
					{
						str = csv.getField(1); // padding character
						char c = str.charAt(0);
						cols.setPadChar(c);
						System.out.println("Padding character now [" + cols.getPadChar() + "]");
					}
					else if (cmd.startsWith("align"))
					{
						str = csv.getField(1).toUpperCase();
						int af = 0;
						if (str.startsWith("L")) af = ALIGN_LEFT;
						else if (str.startsWith("R")) af = ALIGN_RIGHT;
						else if (str.startsWith("C")) af = ALIGN_CENTER;
						if (af==0)
						{ // invalid alignment
							System.out.println("Invalid alignment.  Enter L, R, or C");
						}
						else
						{ // set alignment flag
							cols.setAlignmentFlag(af);
						}
					}
					else if (cmd.startsWith("get"))
					{ // get specified field
						String s1 = csv.getField(1); // begin index
						String s2 = csv.getField(2); // end index
						int beg=-1;
						int end=-2;
						try
						{
							beg = Integer.parseInt(s1);
							end = Integer.parseInt(s2);
							str = cols.getField(beg, end);
							System.out.println("field is [" + str + "], [" + cols.dump(str) + "]");
						} catch (Exception e)
						{
							System.out.println("Invalid numbers: [" + beg + ", [" + end + "]");
						}
					}
					else if (cmd.startsWith("set"))
					{
						String s1 = csv.getField(1); // begin index
						String s2 = csv.getField(2); // end index
						try
						{
							int beg = Integer.parseInt(s1);
							int end = Integer.parseInt(s2);
							String s3 = csv.getField(3); // new value
							String s4 = cols.setField(beg, end, s3);
							System.out.println("Old value was [" + s4 + "], [" + cols.dump(s4) + "]");
							System.out.println("Record now [" + cols.dump() + "]");
						} catch (Exception e)
						{
							System.out.println("Invalid number(s): [" + s1 + "], [" + s2 + "]");
						}
					}
					else if (cmd.startsWith("quit"))
					{
						break;
					}
					else if (cmd.startsWith("reset"))
					{ // start new record
						break;
					}
					else if (cmd.startsWith("dump"))
					{
						str = cols.dump();
						System.out.println("Record is [" + str + "]");
					}
					else
					{
						System.out.println("Invalid command [" + cmd + "]");
						continue;
					}
				}
			} catch (NumberFormatException e)
			{
				recsize = 0;
			} catch (IOException e)
			{
				System.err.println("I/O Exception [" + e.getMessage() + "]");
				break;
			}
			if (cmd.startsWith("quit"))
			{ // quit
				status = 0;
				break;
			}
        }
		System.exit(status);
    }

}
