package com.billpringle.utils.wrputils;

import java.math.BigDecimal;
import java.text.NumberFormat;

/**
 * Class to encapsullate money values.
 * 
 * This class represents money values.
 * The money values are stored internally
 * as BigDecimal variables.
 * 
 * All BigDecimal numbers are created using
 * Strings in the constructors to avoid loss
 * of precision if converted to double.
 * 
 * @author Bill Pringle
 *
 */
public class WrpMoney
{
	/** the actual amount of money */
	private BigDecimal amount;
	/** scale - the number of digits right of decimal place */
	private int scale=2;
	/** method of rounding */
	private int rounding = BigDecimal.ROUND_HALF_DOWN;

	/**
	 * Default constructor
	 * 
	 */
	public WrpMoney()
	{
		
	}
	
	/**
	 * Constructor using a string
	 * 
	 * @param val the value of the amount
	 */
	public WrpMoney(String val) throws NumberFormatException
	{
		String s = val;
		if (s==null || s.length()==0) s = "0";
		Character ch = s.charAt(0);
		if (ch=='$') s = s.substring(1);
		amount = new BigDecimal(s);
	}
	
	/**
	 * Constructor using a BigDecimal value
	 * 
	 * @param val the value of the amount
	 */
	public WrpMoney(BigDecimal val)
	{
		amount = new BigDecimal(val.toPlainString());
	}
	
	/**
	 * Calculate the total of two money values
	 * 
	 * Calculate the sum of this amount and
	 * the amount passed.
	 * Return the sum as a new object
	 * 
	 * @param amt contains the amount to add to this amount
	 * @return a new WrpMoney object with the sum
	 */
	public WrpMoney add(WrpMoney amt)
	{
		BigDecimal sum = amount.add(amt.getAmount());
		WrpMoney ret = new WrpMoney(sum);
		return ret;
	}
	
	/**
	 * Calculate the current value plus the given value
	 * 
	 * This method calculates the sum of the current
	 * money value plus the specified amount
	 * and returns the result as a new object
	 * 
	 * @param val the value to add
	 * @return a new WrpMoney object with the sum
	 */
	public WrpMoney add(BigDecimal val)
	{
		WrpMoney v = new WrpMoney(val);
		return add(v);
	}
	
	/**
	 * Increment the current amount by the specified amount
	 * 
	 * This method increments the money value by the
	 * amount stored in the argument
	 * 
	 * @param amt contains the amount to be added
	 */
	public void incrementBy(WrpMoney amt)
	{
		BigDecimal sum = amount.add(amt.getAmount());
		amount = sum;
	}
	
	/**
	 * Increment the money value by the amount specified
	 * 
	 * This method will increment the internal amount of money
	 * by the specified value in the argument
	 * 
	 * @param val the amount to add to the money amount
	 */
	public void incrementBy(BigDecimal val)
	{
		BigDecimal v = amount.add(val);
		amount = v;
	}
	
	/**
	 * Subtract two money values
	 * 
	 * Calculate the difference between the current amount
	 * and the amount passed.
	 * Return the difference as a new object.
	 * 
	 * @param amt contains the amount to subtract from this amount
	 * @return a new WrpMoney object with the difference
	 */
	public WrpMoney subtract(WrpMoney amt)
	{
		BigDecimal diff = amount.subtract(amt.getAmount());
		WrpMoney ret = new WrpMoney(diff);
		return ret;
	}
	
	/**
	 * Calculate the difference of two money values
	 * 
	 * This method subtracts the specified amount 
	 * from the amount of this object
	 * and returns a new object with that difference
	 * 
	 * @param val the amount to subtract from the current value
	 * @return a new object with the difference
	 */
	public WrpMoney subtract(BigDecimal val)
	{
		WrpMoney v = new WrpMoney(val);
		return subtract(v);
	}
	
	/**
	 * Decrement the money amount by the specified value.
	 * 
	 * This method will decrement the amount of this object
	 * by the value contained in the argument
	 * 
	 * @param amt contains the amount to subtract
	 */
	public void decrementBy(WrpMoney amt)
	{
		BigDecimal diff = amount.subtract(amt.getAmount());
		amount = diff;
	}
	
	/**
	 * Compute the product of two amounts.
	 * 
	 * This method calculates the product of
	 * this amount with the specified amount.
	 * (probably an amount and a rate)
	 * and returns the result as a new object
	 * 
	 * @param amt contains the amount to multiply this by
	 * @return a new WrpMoney object with the product
	 */
	public WrpMoney multiply(WrpMoney amt)
	{
		BigDecimal prod = amount.multiply(amt.getAmount());
		WrpMoney ret = new WrpMoney(prod);
		return ret;
	}
	
	/**
	 * Multiply the current money amount by the specified value
	 * 
	 * This method will multiply the value of this object
	 * by the specified amount
	 * and return the result in a new object.
	 * 
	 * @param amt the amount to multiply the amount of this object
	 * @return a new object with the product
	 */
	public WrpMoney multiply(BigDecimal amt)
	{
		WrpMoney v = new WrpMoney(amt);
		return multiply(v);
	}
	
	/**
	 * Multiply the current amount by the specified amount
	 * 
	 * This method will multiply the money amount
	 * of this object by the value stored in the parameter
	 * 
	 * @param amt contains the amount to be multiplied by
	 */
	public void multiplyBy(WrpMoney amt)
	{
		BigDecimal v = amount.multiply(amt.getAmount());
		amount = v;
	}
	
	/**
	 * Multiply the current money value by the specified value
	 * 
	 * This method multiplies the current money amount
	 * by the amount specified in the parameter
	 * 
	 * @param val the amount to multiply this money amount
	 */
	public void multiplyBy(BigDecimal val)
	{
		BigDecimal v = amount.multiply(val);
		amount = v;
	}
	
	/**
	 * Divide the current amount by the specified amount
	 * 
	 * This method will divide the amount of this object
	 * by the amount stored in the parameter
	 * and return a new object with this amount
	 * 
	 * @param amt amount to divide this money amount by
	 * @return new object with result in amount
	 */
	public WrpMoney divide(WrpMoney amt)
	{
		BigDecimal v = amount.divide(amt.getAmount(), scale, rounding);
		WrpMoney ret = new WrpMoney(v);
		return ret;
	}
	
	/**
	 * Divide money amount by specified value
	 * 
	 * This method divides the amount of the current object
	 * by the specified value
	 * and returns a new object with that value
	 * 
	 * @param val the value to divide by
	 * @return a new object with the result
	 */
	public WrpMoney divide(BigDecimal val)
	{
		BigDecimal v = amount.divide(val);
		WrpMoney ret = new WrpMoney(v);
		return ret;
	}
	
	/**
	 * Divide the current amount by the specified amount
	 * 
	 * This method will divide the amount of the current object
	 * with the amount specified in the parameter
	 * and udpate the current amount with the result
	 * 
	 * @param amt the amount to divide by
	 */
	public void divideBy(WrpMoney amt)
	{
		BigDecimal v = amount.divide(amt.getAmount());
		amount = v;
	}
	
	/**
	 * Divide the current amount by the specified value
	 * 
	 * This method divides the amount in the current object
	 * with the parameter value
	 * and updates the amount to that result
	 * 
	 * @param val the amount to divide by
	 */
	public void divideBy(BigDecimal val)
	{
		BigDecimal v = amount.divide(val);
		amount = v;
	}
	
	/**
	 * Format the current amount
	 * 
	 * This method returns a string with the current
	 * value formatted for currency.
	 * <.p><p>
	 * <b>Note:</b><br />
	 * Due to the need to convert from BigDecimal to double
	 * when using the NumberFormat Currency methods,
	 * some precision might be lost in the formatted string,
	 * but the actual amount is unaffected.
	 * 
	 * @return the amount in currency format
	 */
	public String format()
	{
		NumberFormat n = NumberFormat.getCurrencyInstance();
		double d = amount.doubleValue(); // may lose precision
		String ret = n.format(d);
		return ret;
	}
	
	/**
	 * @return the amount
	 */
	public BigDecimal getAmount()
	{
		return amount;
	}

	/**
	 * @param amount the amount to set
	 */
	public void setAmount(BigDecimal amount)
	{
		this.amount = amount;
	}

	/**
	 * @return the scale
	 */
	public int getScale()
	{
		return scale;
	}

	/**
	 * @param scale the scale to set
	 */
	public void setScale(int scale)
	{
		this.scale = scale;
	}

	/**
	 * @return the rounding
	 */
	public int getRounding()
	{
		return rounding;
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString()
	{
		if (amount==null) return ""; // no value
		else return amount.toPlainString(); // no exponent
	}

	/**
	 * @param rounding the rounding to set
	 */
	public void setRounding(int rounding)
	{
		this.rounding = rounding;
	}

	/**
	 * @param args
	 */
	public static void main(String[] args)
	{
		// TODO Auto-generated method stub

	}

}
