package com.billpringle.utils.wrputils;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;

/**
 * This class translates content strings into pattern strings,
 * reflecting the type of characters in the original strings.
 * </p><p>
 * This class allows the user to determine the general pattern
 * of a data string, such as all numeric, alphabetic, etc.
 * Pattern indicators include the following:
 * </p>
 * <ul>
 * <li>9 &ndash; numeric digit</li>
 * <li>a &ndash; alphabetic letter</li>
 * <li>A &ndash; upper case letter (if case differentiation requested</li>
 * <li>_ &ndash; (underscore) - white space character</li>
 * <li>- &ndash; (dash) delimiter (period, comma, semicolon, dash, etc.)</li>
 * <li>? &ndash; any other character</li>
 * </ul>
 * <p> 
 * This class can also be used to remove any characters that
 * doesn't match a given pattern.  The following types of patterns
 * are supported.  The following codes are case-insensitive.
 * </p>
 * <ul>
 * <li>~ &ndash; patterns starting with a tilde are trimmed.
 * 			a pattern string must follow this character</li>
 * <li>p &ndash; Default pattern - represent all characters in input string</li>
 * <li>x &ndash; Build pattern for only alphanumeric characters, discarding all others</li>
 * <li>f &ndash; Return the original input string (after trimming)</li>
 * <li>c &ndash; Case-sensitive pattern: 'A' for upper case, and 'a' for lower case</li>
 * <li>u &ndash; Translate all letters into upper case</li>
 * <li>9 &ndash; Numeric only - replace all digits with a '9', ignoring all others</li>
 * <li>z<i>N</i> &ndash; Zero filled pattern.  Same as '9' pattern, 
 * 			but if less than <i>N</i> digits, 
 * 			add '9' characters until the pattern has a length of <i>N</i></li>
 * <li>n<i>N</i> &ndash; Numeric characters only - copy all digits from the input string, 
 * 			discarding anything other than a digit.
 * 			If the resulting string is less than <i>N</i> digits,
 * 			then append enough leading zeros to make it that long.</li>
 * <li>a &ndash; Alphabetic only.  Copy all letters from the source string,
 * 			discarding all others.</li>
 * <li>^ &ndash; Allow control characters (not implemented)</li>
 * </ul>
 * 
 * 
  <table border="0" cellspacing="0" cellpadding="6" align="center"
  summary="Creative Commons license">
    <tr>
      <td><!-- Creative Commons License -->
      <a rel="license" href=
      "http://creativecommons.org/licenses/by-nc-sa/2.0/" target=
      "CC"><img alt="Creative Commons License" border="0" src=
      "http://creativecommons.org/images/public/somerights20.gif" /></a></td>

      <td><a rel="license" href=
      "http://creativecommons.org/licenses/by-nc-sa/2.0/" target=
      "CC"><img alt="Creative Commons License Symbols" border="0"
      src="cc_lic.gif" /></a> <!-- /Creative Commons License --></td>

      <td>Unless noted otherwise, all materials available for
      download from my site are copyrighted by Bill Pringle, and
      are licensed under a <a rel="license" href=
      "http://creativecommons.org/licenses/by-nc-sa/2.0/" target=
      "CC">Creative Commons License</a>.</td>
    </tr>
  </table>
  
 * @author Bill Pringle
 *
 */
public class WrpPattern {
	/** input string - output pattern based on this string */
	private String inString=null;
	/** pattern key - what kind of pattern */
	private String pattern="p"; // default pattern

	/**
	 * Default constructor
	 * 
	 */
	public WrpPattern()
	{
		init();
	}
	
	/**
	 * Constructor using default pattern and specified string
	 * 
	 * @param str
	 */
	public WrpPattern(String str)
	{
		init();
		applyPattern(str);
	}
	
	/**
	 * Initialization routine.
	 * This method clears the input string
	 * and the output pattern string.
	 * 
	 */
	private void init()
	{
		inString=null;
		pattern="p";
	}
	
	/**
	 * Define the default pattern.
	 * This method stores the specified pattern
	 * for later use.
	 * 
	 */
	public void setPattern(String pattern)
	{
		this.pattern = pattern;
	}
	/**
	 * Apply the default pattern to the specified string.
	 * 
	 * @param str the string for which a pattern is to be built
	 * @return a string containing the pattern of the input string using the specified pattern code
	 * 
	 */
	public String applyPattern(String str)
	{
		String ret = applyPattern(str, pattern);
		return ret;
	}
	
	/**
	 * Apply the default pattern to the specified string.
	 * 
	 * @param patt the pattern to be applied
	 * @return the result of the pattern applied to the default input string
	 */
	public String patternString(String patt)
	{
		String ret = applyPattern(inString, patt);
		return ret;
	}
	
	/**
	 * Apply the previously specified pattern against
	 * the specified field string.
	 *  
	 * @param fld the input string
	 * @param patt the pattern code to be used in builing the resulting pattern
	 * @return the resulting pattern for the input string using the current pattern
	 * 
	 */
	public static String applyPattern(String fld, String patt)
	{
		if (patt.length()==0) return ""; // no pattern
		StringBuffer sb = new StringBuffer();
		// switches to determine what goes into pattern
		Boolean alpha=false, digit=false, 
			space=false, delim=false, 
			trim=false;
		Boolean casePatt=false;

		// examine the pattern code to determine how the pattern should be built
		String p = patt.toLowerCase();
		// trim code must be at start of pattern
		// If found, save it and examine the rest of the pattern code
		if (p.charAt(0)=='~')
		{
			if (p.length()>1)
			{ // trim before pattern
				p = p.substring(1);
			}
			else
			{ // invalid - pattern must follow trim
				System.err.println("Warning: Pattern must follow trim (~) [" + p + "] ~f assumed");
				return fld.trim();	// just trim field
			}
		}
		// pattern is normally a single character,
		// but can be followed by a qualifier (e.g., z5)
		if (p.equals("p"))
		{ // normal pattern - include all characters
			alpha = digit = space = delim = true;
		}
		else if (p.equals("x"))
		{ // alpha-numeric only
			alpha = digit = true;
		}
		else if (p.equals("f"))
		{ // no pattern - just return original field
			return fld;
		}
		else if (p.equals("c"))
		{ // case-sensitive pattern
			alpha = digit = space = delim = true;
			casePatt = true;
		}
		else if (p.equals("u"))
		{ // convert to upper case
			return fld.toUpperCase();
		}
		else if (p.equals("9") || p.startsWith("z"))
		{ // only digits
			digit = true;
		}
		else if (p.startsWith("n"))
		{ // field contents - digits only
			for (int i=0; i<fld.length(); i++)
			{
				char c = fld.charAt(i);	// get character
				if (Character.isDigit(c))
				{
					sb.append(c);
				}
			}
			if (p.length()>1)
			{ // zero fill numeric value
				try
				{ // get number of digits
					int len = Integer.parseInt(p.substring(1));
					while (sb.length()<len)
					{ // add leading zero
						sb.insert(0, '0');
					}
				} catch (NumberFormatException e)
				{
					System.err.println("applyPattern: Invalid number of digits [" + p + "]");
				}
			}
			String ret = sb.toString();
			return ret;
		}
		else if (p.equals("a"))
		{ // copy contents - only alpha chars
			for (int i=0; i<fld.length(); i++)
			{
				char c = fld.charAt(i);	// get character
				if (Character.isLetter(c))
				{
					sb.append(c);
				}
			}
			String ret = sb.toString();
			return ret;
		}
		else if (p.equals("^"))
		{ // allow control characters
			alpha = digit = space = delim = true;
		}
		else
		{ // invalid pattern code
			System.err.println("Invalid pattern code [" + patt + "]");
			return fld; // don't do anything
		}
		
		// process each character and replace with desired types
		String s;
		if (trim)
		{
			s = fld.trim();
		}
		else
		{
			s = fld;
		}
		for (int i=0; i<s.length(); i++)
		{
			char c = s.charAt(i);	// get character
			if (Character.isLetter(c))
			{ // letter
				if (alpha) 
				{
					if (casePatt)
					{ // differentiate between upper and lower case
						if (Character.isUpperCase(c))
						{
							sb.append("A");
						}
						else
						{
							sb.append("a");
						}
					}
					else
					{
						sb.append("a");
					}
				}
			}
			else if (Character.isDigit(c))
			{ // digit
				if (digit) sb.append("9");
			}
			else if (Character.isWhitespace(c))
			{ // white space
				if (space) sb.append("_");
			}
			else
			{ // must be delimiter
				if (delim) sb.append("-");
			}
		}
		
		// check for leading zero pattern
		if (p.startsWith("z"))
		{
			try
			{
				int z = Integer.parseInt(p.substring(1));
				while (z > sb.length()) sb.append("9");
			} catch (NumberFormatException e)
			{
				System.err.println("Invalid field length for z pattern [" + patt + "]");
			}
		}
		
		String ret = sb.toString();
		return ret;
	}
	
	// get/set routines
	/**
	 * Set the default string.
	 * 
	 * @param str the default input string
	 */
	public void setString(String str) { this.inString = str; }
	
	/**
	 * Get the default input string.
	 * 
	 * @return the default input string
	 */
	public String getString() { return this.inString; }
	
	/**
	 * Get the default pattern.
	 * 
	 * @return the default pattern
	 */
	public String getPattern() { return this.pattern; }
	
	/**
	 * Test driver.
	 * 
	 * @param args not used
	 */
	public static void main(String[] args)
	{
		WrpPattern patt = new WrpPattern();
		
		String inStr, pattStr;
		System.out.println("Enter pattern string: ");
        InputStreamReader rdr = new InputStreamReader(System.in);
        BufferedReader kbd = new BufferedReader(rdr);
        try
        {
            pattStr = kbd.readLine();
            while (true)
            { // get strings and turn them into patterns
            	System.out.println("Enter string: ");
            	inStr = kbd.readLine();
            	String p = patt.applyPattern(inStr);
            	System.out.println("String [" + inStr + "] maps to pattern [" + p + "]");
            	if (p.equals(pattStr))
            	{
            		System.out.println(" -- patterns match");
            	}
            	else
            	{
            		System.out.println(" ** patterns don't match");
            	}
            }
        }
        catch (IOException e)
        {
           System.out.println("I/O error" + e.toString());
        }

	}

}
