package com.billpringle.utils.wrputils;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Timestamp;

/**
 * This class encapsulates a record within a ResultSet.
 * It can be used with
 * {@link com.billpringle.utils.wrputils.WrpResultSetRow}
 * to create a JTable.
 * 
 *   <table border="0" cellspacing="0" cellpadding="6" align="center"
summary="Creative Commons license">
<tr>
  <td><!-- Creative Commons License -->
  <a rel="license" href=
  "http://creativecommons.org/licenses/by-nc-sa/2.0/" target=
  "CC"><img alt="Creative Commons License" border="0" src=
  "http://creativecommons.org/images/public/somerights20.gif" /></a></td>

  <td><a rel="license" href=
  "http://creativecommons.org/licenses/by-nc-sa/2.0/" target=
  "CC"><img alt="Creative Commons License Symbols" border="0"
  src="cc_lic.gif" /></a> <!-- /Creative Commons License --></td>

  <td>Unless noted otherwise, all materials available for
  download from my site are copyrighted by Bill Pringle, and
  are licensed under a <a rel="license" href=
  "http://creativecommons.org/licenses/by-nc-sa/2.0/" target=
  "CC">Creative Commons License</a>.</td>
</tr>
</table>
 * 
 * @author Bill Pringle
 *
 */
public class WrpResultSetRow
{
	/** parent result set vector */
	public WrpResultSetVector vec;
	/** number of columns */
	public int ncols;
	/** names of columns */
	public String[] colNames;
	/** data types of columns */
	public String[] colTypes;
	/** vector for columns */
	public String[] colValues;
	
	/**
	 * Normal constructor.
	 * The parent vector will include this object
	 * 
	 * @param vec parent vector
	 * @throws SQLException 
	 */
	public WrpResultSetRow(WrpResultSetVector vec) throws SQLException
	{
		// get information from parent vector
		this.vec = vec; // save parent vector
		colNames = vec.colNames; // column headers
		colTypes = vec.colTypes; // column types
		
		// save the columns as strings
		buildColumns();
	}

	/**
	 * Extract the columns for this record and convert to an array of strings.
	 * The index for result strings start with 1, not 0.
	 * @throws SQLException 
	 * 
	 */
	private void buildColumns() throws SQLException
	{
		if (colNames==null) ncols = 0;
		else
		{
			ncols = colNames.length; // number of columns
		}
		colValues = new String[ncols]; // column values
		ResultSet resultSet = vec.getResultSet();
		
		// loop through columns and convert to strings
		for (int i=0; i<ncols; i++)
		{
			String st = colTypes[i]; // type of field
			if (st.equals("int"))
			{ // integer
				int n = resultSet.getInt(i+1);
				colValues[i] = n + ""; // store as string
			}
			else if (st.equals("varchar"))
			{ // string
				String s = resultSet.getString(i+1);
				colValues[i] = s;
			}
			else if (st.equals("datetime"))
			{
				Timestamp ts = resultSet.getTimestamp(i+1);
				colValues[i] = ts + "";
			}
			else if (st.equals("float"))
			{
				float f = resultSet.getFloat(i+1);
				colValues[i] = f + "";
			}
			else if (st.equals("int unsigned"))
			{
				int n = resultSet.getInt(i+1);
				colValues[i] = n + "";
			}
			else
			{
				System.err.println("WrpResultSetRow: invalid column type [" + st + "]");
			}
		}
	}
	
	/**
	 * Get the column value as a string
	 * 
	 * @param ndx index of column
	 * @return column value as a string
	 */
	public String getColString(int ndx)
	{
		if (colValues==null) return null; // no data yet
		String ret = null;
		if (ndx >=0 && ndx < colValues.length)
		{
			ret = colValues[ndx];
		}
		return ret;
	}

	/**
	 * @return the vec
	 */
	public WrpResultSetVector getVec()
	{
		return vec;
	}

	/**
	 * @param vec the vec to set
	 */
	public void setVec(WrpResultSetVector vec)
	{
		this.vec = vec;
	}

	/**
	 * @return the ncols
	 */
	public int getNcols()
	{
		return ncols;
	}

	/**
	 * @param ncols the ncols to set
	 */
	public void setNcols(int ncols)
	{
		this.ncols = ncols;
	}

	/**
	 * @return the colNames
	 */
	public String[] getColNames()
	{
		return colNames;
	}

	/**
	 * @param colNames the colNames to set
	 */
	public void setColNames(String[] colNames)
	{
		this.colNames = colNames;
	}

	/**
	 * @return the colTypes
	 */
	public String[] getColTypes()
	{
		return colTypes;
	}

	/**
	 * @param colTypes the colTypes to set
	 */
	public void setColTypes(String[] colTypes)
	{
		this.colTypes = colTypes;
	}

	/**
	 * @return the colValues
	 */
	public String[] getColValues()
	{
		return colValues;
	}

	/**
	 * @param colValues the colValues to set
	 */
	public void setColValues(String[] colValues)
	{
		this.colValues = colValues;
	}
	
	
}
