package com.billpringle.utils.wrputils;

import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.util.Vector;

import javax.swing.JFrame;
import javax.swing.JList;
import javax.swing.JScrollPane;
import javax.swing.WindowConstants;

/**
 * Create a scrolled window containing the specified content.
 * This class can be used to display such things as
 * a help messages, search results, etc.
 * The resulting frame is independent, even if called from an
 * JInternalFrame class.
 * </p>
 * 
 * 
  <table border="0" cellspacing="0" cellpadding="6" align="center"
  summary="Creative Commons license">
    <tr>
      <td><!-- Creative Commons License -->
      <a rel="license" href=
      "http://creativecommons.org/licenses/by-nc-sa/2.0/" target=
      "CC"><img alt="Creative Commons License" border="0" src=
      "http://creativecommons.org/images/public/somerights20.gif" /></a></td>

      <td><a rel="license" href=
      "http://creativecommons.org/licenses/by-nc-sa/2.0/" target=
      "CC"><img alt="Creative Commons License Symbols" border="0"
      src="cc_lic.gif" /></a> <!-- /Creative Commons License --></td>

      <td>Unless noted otherwise, all materials available for
      download from my site are copyrighted by Bill Pringle, and
      are licensed under a <a rel="license" href=
      "http://creativecommons.org/licenses/by-nc-sa/2.0/" target=
      "CC">Creative Commons License</a>.</td>
    </tr>
  </table>
 * 
 * @author Bill Pringle
 *
 */
public class WrpScrollFrame extends JFrame
	implements ActionListener, ComponentListener
{
	/** serialization identity */
	final static long serialVersionUID = 1;

	// class instance variables
	/**
	 * Vector of strings for this scrolled frame.
	 * Each entry in the list will be inserted
	 * into the scrolled frame as a separate line in the list.
	 * The entries of the list can contain embedded end of line sequences.
	 * 
	 */
	protected Vector<String> scrollData=null;
	
	/**
	 * The scrolled pane.
	 * This pane will contain the information to be displayed in the frame.
	 * 
	 */
	protected JScrollPane scroll=null;
	
	/**
	 * The list that contains the text to be displayed in the frame.
	 * 
	 */
	protected JList lst=null;
	
	/**
	 * Default constructor.
	 * This method will create an empty scrolled frame.
	 * 
	 */
	public WrpScrollFrame()
	{
		buildPage(); // create the page
		resizeScroll(); // set the size of the scroll to fill up the window
		setVisible(true); // mark the frame as visible
		setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE); // dispose when done
	}
	
	/**
	 * Constructor with list of strings to be inserted into the frame.
	 * This constructor will set the list data to the provided list.
	 * 
	 * @param vec a list of the strings to be displayed in the scrolled frame
	 * 
	 */
	public WrpScrollFrame(Vector<String> vec)
	{
		this(); // call default constructor
		setScrollData(vec); // set the list data for the frame
	}
	
	/**
	 * Constructor with list and frame title.
	 * This method will create a scrolled frame,
	 * insert the specified list of strings,
	 * and set the title of the frame window. 
	 * 
	 * @param vec the list of data for the frame
	 * @param title the title of the frame
	 */
	public WrpScrollFrame(Vector<String> vec, String title)
	{
		this(vec); // create frame
		setTitle(title); // set the title
	}

	/**
	 * Build a scrolled frame with an empty list.
	 * This method is called by the default constructor
	 * to create the actual frame with an empty scrolled panel.
	 * 
	 */
	protected void buildPage()
	{
		if (scrollData==null) scrollData = new Vector<String>(); // create vector if not defined yet
		lst = new JList(scrollData); // create a list for the scroll data
		scroll = new JScrollPane(lst); // create a scrolled pane for the list
		// set scroll bar policies to always display both scroll bars
		scroll.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_ALWAYS);
		scroll.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
		add(scroll); // add the scrolled panel to the frame
	}
	
	/**
	 * Set or modify the contents of the frame.
	 * This method will set the contents of the frame to the specified
	 * list of strings.
	 * If the frame already had data, then it will be replaced by this list.
	 * Each entry in the vector will be displayed as a separate line
	 * The strings can have embedded additional newline sequences.
	 * 
	 * @param vec the data to be displayed in the scrolled frame
	 * 
	 */
	public void setScrollData(Vector<String> vec)
	{
		scrollData =  vec; // set the contents of the vector
		lst.setListData(scrollData); // set the list data
		resizeScroll(); // resize the scrolled data
	}
	
	/**
	 * Append the specified list of strings to the end of the scrolled frame.
	 * This method adds the strings in the vector <i>vec</i> to the
	 * vector being displayed in the scrolled frame.
	 * Each entry in the vector will be displayed as a separate line
	 * The strings can have embedded additional newline sequences.
	 * 
	 * @param vec the data to be appended to the bottom of the scrolled frame
	 */
	public void appendScrollData(Vector<String> vec)
	{
		scrollData.addAll(vec); // append the data
		lst.setListData(scrollData); // update the list data
		resizeScroll(); // resize the scrolled data
	}
	
	/**
	 * Set the title of the scrolled frame.
	 * This method will set the title of the frame to the specified string.
	 * 
	 */
	public void setTitle(String title)
	{
		super.setTitle(title); // set the title
	}
	
	/**
	 * Set the size of the scrolled text to the default size if not created.
	 * This method sizes the scroll to the default size if not created yet.
	 * Once the frame has been created, the size will remain unchanged
	 * unless the frame is resized manually.
	 * Adding or setting the content of the frame doesn't affect the size.
	 * 
	 */
	protected void resizeScroll()
	{
		Dimension dim = getSize(); // get current size of frame
		if (dim.width==0)
		{ // not created - pick the default size
			dim.width = 300; // default width
			dim.height = 500; // default height
			setSize(dim); // set the frame to the default size
		}
	}
	
	/**
	 * This method handles any action generated by the user.
	 * This method is not currently being used.
	 * 
	 */
	public void actionPerformed(ActionEvent e)
	{
		
	}
	
	// ComponentListener routines
	/**
	 * This method is called when the frame is hidden.
	 */
    public void componentHidden(ComponentEvent e)
    {
    }
    
    /**
     * This method is called when the frame is moved.
     */
    public void componentMoved(ComponentEvent e)
    {
    }
    
    /**
     * This method is called when the frame is resized.
     * The scrolled pane will be resized to match the frame size.
     * 
     */
    public void componentResized(ComponentEvent e)
    {
    	resizeScroll();
    }
    
    /**
     * This method is called when the frame is made visible.
     */
    public void componentShown(ComponentEvent e)
    {
    }
    
	/**
	 * Test driver for this frame.
	 * This method is typically not used.
	 * It is mainly useful for testing purposes.
	 * 
	 * @param args not used
	 */
	public static void main(String[] args) {
        java.awt.EventQueue.invokeLater(new Runnable() {
            public void run() {
                new WrpScrollFrame();
            }
        });
	}

}
