package com.billpringle.utils.wrputils;

import javax.swing.JFrame;



/**
 * This class displays a result set in a scrolled table.
 * The user can specify what columns should appear in the
 * scrolled table.  
 * The remaining fields will be displayed in the detail area.
 * 
 * 
 *   <table border="0" cellspacing="0" cellpadding="6" align="center"
summary="Creative Commons license">
<tr>
  <td><!-- Creative Commons License -->
  <a rel="license" href=
  "http://creativecommons.org/licenses/by-nc-sa/2.0/" target=
  "CC"><img alt="Creative Commons License" border="0" src=
  "http://creativecommons.org/images/public/somerights20.gif" /></a></td>

  <td><a rel="license" href=
  "http://creativecommons.org/licenses/by-nc-sa/2.0/" target=
  "CC"><img alt="Creative Commons License Symbols" border="0"
  src="cc_lic.gif" /></a> <!-- /Creative Commons License --></td>

  <td>Unless noted otherwise, all materials available for
  download from my site are copyrighted by Bill Pringle, and
  are licensed under a <a rel="license" href=
  "http://creativecommons.org/licenses/by-nc-sa/2.0/" target=
  "CC">Creative Commons License</a>.</td>
</tr>
</table>
 *
 * @author Bill Pringle
 *
 */
public class WrpScrollResultSet extends WrpScrollTable
{
	/**
	 * serial version
	 */
	private static final long serialVersionUID = 1L;
	/** result set vector to be displayed */
	public WrpResultSetVector vec;
	
	/**
	 * Constructor with only frame
	 * 
	 * @param win the parent window
	 */
	public WrpScrollResultSet(JFrame win)
	{
		super(win);
	}
	
	/**
	 * Constructor with frame and vector.
	 * 
	 * @param win parent window
	 * @param v result set vector
	 */
	public WrpScrollResultSet(JFrame win, WrpResultSetVector v)
	{
		super(win);
		setVector(v);
	}
	
	/**
	 * Set the result set vector and create model.
	 * This method is called by the constructor accepting a vector.
	 * If a different constructor was used, then the caller
	 * should call this method explicitly.
	 * 
	 * @param v the result set vector
	 */
	public void setVector(WrpResultSetVector v)
	{
		vec = v;
		WrpScrollResultModel model = new WrpScrollResultModel(vec);
		setModel(model);
	}
	
	/**
	 * The table model for the scrolled result set.
	 * 
	 * @author Bill Pringle
	 *
	 */
	public class WrpScrollResultModel extends WrpTableModel
	{
		/** sequencing serial version */
		private static final long serialVersionUID = 1L;
		/** result set vector */
		private WrpResultSetVector vec;
		/** number of columns */
		private int ncols;
		/** column names */
		private String[] colNames;
		/** column types */
		private String[] colTypes;
		/** columns in scroll */
		private int[] colScroll;

		/**
		 * Constructor for table model.
		 * 
		 * @param v the result set vector
		 */
		public WrpScrollResultModel(WrpResultSetVector v)
		{
			// set vector
			setVector(v);
		}
		
		/**
		 * Set the vector and extract information for model.
		 * This method handles null being passed for the vector.
		 * 
		 * @param v result set vector
		 */
		public void setVector(WrpResultSetVector v)
		{
			vec = v;
			if (vec!=null)
			{
				colNames = vec.getColNames();
				colTypes = vec.getColTypes();
				buildColScroll(vec.colScroll);
				ncols = colScroll.length;
			}
		}
		
		/**
		 * Builds list of columns to display in scroll.
		 * If null is passed, then all columns are displayed.
		 * 
		 * @param cols the column indices to display
		 */
		public void buildColScroll(int[] cols)
		{
			if (cols == null)
			{ // default - display all columns
				colScroll = new int[ncols];
				for (int i=0; i<ncols; i++)
				{
					colScroll[i] = i;
				}
			}
			else
			{ // use given columns
				colScroll = cols;
			}
		}
		
		/* (non-Javadoc)
		 * @see com.billpringle.utils.WrpScrollTable.WrpTableModel#getDescription(int)
		 */
		public String getDescription(int nrow)
		{
			if (vec==null || vec.rows==null) return ""; // no rows
			if (nrow < 0 || nrow > vec.rows.size()) return null; // invalid index
			
			WrpResultSetRow row = vec.getRow(nrow);
			StringBuffer sbuf = new StringBuffer(); // work string

			// loop through columns and build rows of name = value
			for (int i=0; i<colNames.length; i++)
			{
				sbuf.append(colNames[i] 
				       + " = " + row.getColString(i) 
				       + " [" + colTypes[i] + "]"
				       + "\n");
			}
			
			String ret = sbuf.toString();
			return ret;
		}
		
		
		public int getRowCount()
		{
			if (vec==null || vec.rows==null) return 0;
			return vec.rows.size();
		}
		
		public int getColumnCount()
		{
			return colScroll.length;
		}

		/* (non-Javadoc)
		 * @see javax.swing.table.DefaultTableModel#getValueAt(int, int)
		 */
		public String getValueAt(int row, int col)
		{
			if (row<0 || row>= getRowCount()) return ""; // invalid
			if (col<0 || col>= getColumnCount()) return "";
			WrpResultSetRow r = vec.getRow(row);
			int ndx = colScroll[col];
			String ret = r.getColString(ndx);
			if (ret==null) ret = "";
			return ret;
		}
		
		
		/* (non-Javadoc)
		 * @see javax.swing.table.DefaultTableModel#getColumnName(int)
		 */
		public String getColumnName(int col)
		{
			if (col<0 || col>=colScroll.length) return "[unknown]";
			return colNames[colScroll[col]];
		}
		
		/* (non-Javadoc)
		 * @see javax.swing.table.DefaultTableModel#isCellEditable(int, int)
		 */
		public boolean isCellEditable(int row, int col) { return false; }
		
		/* (non-Javadoc)
		 * @see javax.swing.table.AbstractTableModel#getColumnClass(int)
		 */
		public Class<?> getColumnClass(int col)
		{ // all columns are strings
			return new String().getClass();
		}

	}
}
