package com.billpringle.utils.wrputils;

import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Rectangle;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;

import javax.swing.BoxLayout;
import javax.swing.JFrame;
import javax.swing.JInternalFrame;
import javax.swing.JPanel;
import javax.swing.JScrollBar;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.JTextArea;
import javax.swing.ListSelectionModel;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.event.TableModelEvent;
import javax.swing.event.TableModelListener;
import javax.swing.table.DefaultTableModel;
import javax.swing.table.TableRowSorter;


/**
 * Create a scrolled window divided into two parts.
 * The top part contains a grid with a list.
 * The bottom part can contain a detailed description
 * for the selected entry. 
 * </p>
 * <p>
 * This class can be extended to create customized tables.
 * Subclasses should override certain methods and class variables,
 * and define a table model that extends WrpTableModel.
 * The Javadoc comments indicate what should be overwritten,
 * and what changes would be needed.
 * </p> 
 * 
 * 
 *   <table border="0" cellspacing="0" cellpadding="6" align="center"
summary="Creative Commons license">
<tr>
  <td><!-- Creative Commons License -->
  <a rel="license" href=
  "http://creativecommons.org/licenses/by-nc-sa/2.0/" target=
  "CC"><img alt="Creative Commons License" border="0" src=
  "http://creativecommons.org/images/public/somerights20.gif" /></a></td>

  <td><a rel="license" href=
  "http://creativecommons.org/licenses/by-nc-sa/2.0/" target=
  "CC"><img alt="Creative Commons License Symbols" border="0"
  src="cc_lic.gif" /></a> <!-- /Creative Commons License --></td>

  <td>Unless noted otherwise, all materials available for
  download from my site are copyrighted by Bill Pringle, and
  are licensed under a <a rel="license" href=
  "http://creativecommons.org/licenses/by-nc-sa/2.0/" target=
  "CC">Creative Commons License</a>.</td>
</tr>
</table>
 * 
 * @author Bill Pringle
 *
 */

public class WrpScrollTable extends JInternalFrame 
	implements TableModelListener, ListSelectionListener, 
		ComponentListener, ActionListener
{
	private static final long serialVersionUID = 1L;
	/** main window of application */
	private JFrame win;

	//
	// GUI objects
	//
	/** text area for detail description */
	JTextArea txtDescription;
	/** panel for table */
	JPanel pnlEntries;
	/** panel for description */
	JPanel pnlDesc;
	/** the table to display */
	private JTable tblEntries;
	/** table model for display */
	private WrpTableModel model;
	/** scroll pane for table */
	JScrollPane scrEntries;
	/** scroll panel for description */
	JScrollPane scrDesc;

	/**
	 * Basic constructor.
	 * Subclasses should call the super constructor
	 * and then set the Title.
	 * 
	 * @param win main window
	 */
	public WrpScrollTable(JFrame win)
	{
		super("Browse Entries",
				true,	// resize
				true,	// close
				true, 	// maximize
				true);	// icon
		this.win = win;
		init(); // set up frame
	}

	/**
	 * Constructor with table.
	 * Subclasses should call the super constructor
	 * and then set the Title.
	 * 
	 * @param win main window
	 * @param tbl table to display
	 */
	public WrpScrollTable(JFrame win, JTable tbl)
	{
		super("Browse Entries",
				true,	// resize
				true,	// close
				true, 	// maximize
				true);	// icon
		this.win = win;
		this.tblEntries = tbl;
		init(); // set up frame
	}
	
	/**
	 * Constructor with table and model
	 * 
	 * @param win main window
	 * @param model table model
	 */
	public WrpScrollTable(JFrame win, WrpTableModel model)
	{
		super("Browse Entries",
				true,	// resize
				true,	// close
				true, 	// maximize
				true);	// icon
		this.win = win;
		this.model = model;
		init(); // set up frame
	}
	
	/**
	 * set the table model
	 * 
	 * @param model table model
	 */
	public void setModel(WrpTableModel model)
	{
		this.model = model;
		setEntries();
	}
	
	/**
	 * Initialize the window
	 * 
	 */
	private void init()
	{
		setDefaultCloseOperation(DISPOSE_ON_CLOSE); // terminate when window closed
		buildPage(); // build the page
		buildMenu();
		setEntries();
		resizeScroll(); // fit scrolled area
		pack();
		show();
	}

	/**
	 * Build the page components
	 * 
	 * This method can be overridden by subclass
	 * 
	 */
	protected void buildPage()
	{
		Dimension dim = win.getSize();
		dim.width -= 20;
		dim.height /= 2;
		setSize(dim);
		dim.height /= 2;
		dim.height -= 40;
		JPanel pnlPage = new JPanel();
		pnlPage.setBackground(Color.YELLOW);
		pnlPage.setLayout(new BoxLayout(pnlPage, BoxLayout.PAGE_AXIS));
		pnlEntries = new JPanel();
		pnlPage.add(pnlEntries);
		pnlDesc = new JPanel();
		// pnlDesc.setBackground(Color.pink);
		txtDescription = new JTextArea();
		txtDescription.setSize(dim);
		txtDescription.setLineWrap(true);
		txtDescription.setWrapStyleWord(true);
		scrDesc = new JScrollPane(txtDescription);
		scrDesc.scrollRectToVisible(new Rectangle());
		pnlDesc.add(scrDesc);
		pnlPage.add(pnlDesc);
		add(pnlPage);
		addComponentListener(this);
		setEntries();
	}

	/**
	 * Build menu for frame
	 * 
	 * This method can be overridden by subclass
	 * 
	 */
	protected void buildMenu()
	{
		// TODO Auto-generated method stub
		
	}

	/**
	 * Set the table entries
	 * 
	 */
	protected void setEntries()
	{
		if (model==null) return; // no model yet
		model.addTableModelListener(this);
		tblEntries = new JTable(model);
		tblEntries.setSelectionMode(ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
		tblEntries.setRowSorter(new TableRowSorter<WrpTableModel>(model));
		ListSelectionModel lsm = tblEntries.getSelectionModel();
		lsm.addListSelectionListener(this);

		scrEntries = new JScrollPane(tblEntries);
		pnlEntries.removeAll();
		pnlEntries.add(scrEntries);
		setColumnWidth();
		resizeScroll();
	}

	private void setColumnWidth()
	{
		// TODO Auto-generated method stub
		
	}

	/**
	 * Rearrange window after resizing
	 * 
	 */
	public void resizeScroll()
	{
		if (model==null) return;
		Component win = (Component) this.getParent();
		if (win==null) return;
		Dimension dim = win.getSize();
		// Dimension dim = getSize();
		dim.width -= 20;
		dim.height /= 2;
		dim.height -= 50;
		txtDescription.setLineWrap(true);
		scrEntries.setPreferredSize(dim);
		txtDescription.setSize(dim);
		// scrDesc.setLocation(0, dim.height);
		scrDesc.setPreferredSize(dim);

		JScrollBar sb = scrDesc.getVerticalScrollBar();
		// sb.setValue(-20);
		sb.setValue(0);
		repaint();
	}

	@Override
	public void setTitle(String title)
	{
		this.setTitle(title);
	}
	
	

	@Override
	public void actionPerformed(ActionEvent arg0)
	{
		// TODO Auto-generated method stub
		
	}

	@Override
	public void tableChanged(TableModelEvent arg0)
	{
		// TODO Auto-generated method stub
		
	}

	@Override
	public void valueChanged(ListSelectionEvent e)
	{
		if (e.getValueIsAdjusting()) return; // not done yet
		ListSelectionModel lsm = (ListSelectionModel)e.getSource();
		String txt = "";
		if (!lsm.isSelectionEmpty())
		{
			int sel = lsm.getMinSelectionIndex();
			int ndx = tblEntries.convertRowIndexToModel(sel);
			txt = model.getDescription(ndx);
			txtDescription.setText(txt);
			txtDescription.setLineWrap(true);
			JScrollBar sb = scrDesc.getVerticalScrollBar();
			sb.setValue(-20);
			resizeScroll();
		}
		repaint();
	}
	

	@Override
	public void componentHidden(ComponentEvent e)
	{
		// TODO Auto-generated method stub
		
	}

	@Override
	public void componentMoved(ComponentEvent e)
	{
		// TODO Auto-generated method stub
		
	}

	@Override
	public void componentResized(ComponentEvent e)
	{
		resizeScroll();
	}

	@Override
	public void componentShown(ComponentEvent e)
	{
		// TODO Auto-generated method stub
		
	}

	public class WrpTableModel extends DefaultTableModel
	{

		/**
		 * 
		 */
		private static final long serialVersionUID = 1L;
		
		/**
		 * Returns description for specified entry.
		 * This method should be overridden by a subclass
		 * 
		 * @param ndx the location of the entry within the list
		 * @return the description to appear in the lower part of window
		 */
		public String getDescription(int ndx)
		{
			return "";
		}
	}

}
