package com.billpringle.utils.wrputils;

import java.util.HashMap;

/**
 * This class encapsulates a singleton session object.
 * </p><p>
 * This is a singleton class that should be created 
 * using the static method WrpSession.initInstance()
 * when a user authenticates with the system.
 * Any subsequent calls to WrpSession.initInstance()
 * will be ignored.
 * </p><p>
 * After that, a method can test to determine if the session
 * has expired by calling the static WrpSession.getUserId() method,
 * which will return -1 if expired, or the user ID that was passed
 * to WrpSession.initInstance() at creation time.
 * </p><p>
 * This class also provides a HashMap for storing session level
 * objects.  The HashMap is defined using a key string and an object
 * value, so any Java object (including arrays and complex objects)
 * can be stored by one method and retrieved by another method.
 * 
 * <p><b>&copy; 2011 by Bill Pringle, all rights reserved.</b></p>
 * 
 * 
 * @author Bill Pringle
 *
 */
public class WrpSession
{
	/** the actual singleton object */
	private static WrpSession ref = null;
	/** the user id for session - initialized */
	private static int userId = -1;
	/** HashMap for storing session information */
	private static HashMap<String, Object> sessionMap;

	/**
	 * Private constructor to prevent multiple instances.
	 * This method is called from the initInstance() method.
	 */
	private WrpSession()
	{
		// prevent more than one instance
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#clone()
	 */
	public Object clone() throws CloneNotSupportedException
	{
		throw new CloneNotSupportedException();
	}

	/**
	 * Get the singleton instance.
	 * If the singleton has not been initialized, this will return null.
	 * 
	 * @return the singleton, or null if not initialized
	 */
	public static WrpSession getInstance()
	{
		return ref;
	}

	/**
	 * Initialize the singleton.
	 * This method should be called first, and will create the singleton
	 * instance and initialize it.
	 * 
	 * @param userId the current logged-in user ID
	 */
	public static synchronized WrpSession initInstance(int userId)
	{
		if (ref!=null) return ref; // only one instance
		
		// first time
		ref = new WrpSession();
		WrpSession.userId = userId; // save authenticated user
		return ref; // return singleton
	}

	/**
	 * @return the userid
	 */
	public static int getUserid() {
		if (ref==null) return -1; // should not happen
		return userId;
	}

	/**
	 * Get session attribute.
	 * 
	 * @param key attribute name
	 */
	public static Object getMap(String key)
	{
		Object value = sessionMap.get(key);
		return value;
	}
	
	/**
	 * Store a session attribute.
	 * 
	 * @param key attribute name
	 * @param value attribute value
	 * @return old attribute value, null if none
	 */
	public static synchronized Object setMap(String key, Object value)
	{
		Object oldval = sessionMap.put(key, value);
		return oldval;
	}


}
