package com.billpringle.utils.wrputils;


import java.io.PrintStream;
import java.nio.ByteBuffer;
import java.nio.CharBuffer;
import java.nio.charset.CharacterCodingException;
import java.nio.charset.Charset;
import java.nio.charset.CharsetDecoder;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Vector;

/**
 * This class implements an Application Specific Language (ASL)
 * that can be used to validate a CSV file.
 * </p>
 * <ul>
 * <li><a href="#intro">Introduction</a></li>
 * <li><a href="#syntax">Command Syntax</a></li>
 * <li><a href="#commands">Commands</a>
 * <ul>
 * <li><a href="#cmd_control">Control Commands</a></li>
 * <li><a href="#cmd_datadef">Data Definition Commands</a></li>
 * <li><a href="#cmd_assignment">Assignment Commands</a></li>
 * <li><a href="#cmd_string">String manipulation and validation commands</a></li>
 * <li><a href="#cmd_numeric">Numeric test and manipulation commands</a></li>
 * <li><a href="#cmd_misc">Miscellaneous commands</a></li>
 * </ul>
 * </li>
 * <li><a href="#patt">Patterns</a></li>
 * <li><a href="#norm">Normalization</a></li>
 * </ul>
 * 
 * <p id="intro">
 * Validation consists of examining the individual fields
 * within a CSV file, applying various rules to determine
 * if the content of the CSV file is syntactically and
 * possibly semantically correct.
 * The validation routines can generate a series of messages
 * that can be used to communicate the validation results.
 * Typically these messages are only generated in the case
 * of errors or warnings, but they could also be used
 * for other purposes. 
 * </p><p>
 * Examples of validation rules are testing the number of fields
 * in the CSV file against a minimum or maximum value, testing
 * the length of a given field against a minimum or maximum value,
 * etc.  The language also provides the ability to test the pattern
 * of the field values, such as all numeric, or following a more
 * specific format.  For example, a US Zip Code could be tested
 * using the patterns Z5 or Z5-9999, where the first pattern
 * indicates numeric string of at least five digits with leading
 * zeros supplied if needed, while the second pattern could be
 * used to test for the nine-digit Zip Code.
 * </p><p id="syntax">
 * Commands have the following components,
 * which are optional unless noted otherwise:
 * </p>
 * <dl>
 * <dt>Sequence number</dt>
 * <dd>
 * This is a required parameter,
 * containing the unique line number of the command.
 * When the rules are retrieved, they are sorted by sequence number 
 * </dd>
 * <dt>Label</dt>
 * <dd>
 * This is an optional name for the statement,
 * which can be used as a <i>GoTo</i> target.
 * </dd>
 * <dt>InVar</dt>
 * <dd>
 * This is the input variable for the command.
 * If present, the command will perform an operation
 * on this variable,
 * and possibly store the result in the <i>OutVar</i> variable.
 * </dd>
 * <dt>Cmd</dt>
 * <dd>
 * This is a required parameter,
 * and contains the name of the command to be executed.
 * </dd>
 * <dd>
 * Each command has the potential to "fire",
 * usually indicating an exception.
 * For example, the <code>minlen</code> command fires
 * if the field length is less than the specified value.
 * When a command fires, a message is displayed (if present),
 * the status of the record is set to <i>Result</i> (if present),
 * and if <i>GoTo Destination</i> is present, it is executed next.
 * </dd>
 * <dt>Arg1</dt>
 * <dd>
 * This is an optional literal argument for the command.
 * The type of this literal is determined by its content:
 * If it is a numeric string, then it is marked as an Integer
 * or Float, depending on the presence of a period;
 * otherwise it is marked as a String.
 * To force a numeric string to be considered a string,
 * enclose it with double quoted (e.g., "9").
 * </dd>
 * <dd>
 * There should be only one literal entry for any given value.
 * If more than one command uses the same literal value,
 * each command should be pointing to the same literal value.
 * </dd>
 * <dt>Arg2</dt>
 * <dd>
 * This is an optional literal argument for those commands
 * that require more than one argument.
 * See <i>Arg1</i> for more details.
 * If the command only uses one literal argument,
 * then <i>Arg1</i> will be used.
 * </dd>
 * <dt>OutVar</dt>
 * <dd>
 * This is an optional variable name.
 * If the command returns a value,
 * it will be stored in <i>OutVar</i>.
 * The type of this variable depends on the command.
 * </dd>
 * <dt>Result</dt>
 * <dd>
 * This is an optional argument that specifies
 * the status of the record.
 * Valid values are:
 * </p>
 * <ul>
 * <li><b>GOOD</b> The record has no errors.</li>
 * <li><b>WARN</b> A warning message - it might be an error</li>
 * <li><b>BAD</b> The record has serious errors</li>
 * </ul>
 * <p>
 * Typically, this field will be left blank
 * or set to <i>BAD</i>, indicating an error was encountered.
 * The record status is affected if the specified status
 * is worse than the current value.
 * (For example, if the current status is "GOOD", and a command
 * fires with a status of "BAD", then the status of the record
 * will be changed from "GOOD" to "BAD".
 * The status would not be changed from "BAD" to "GOOD".
 * </dd>
 * <dt>Message</dt>
 * <dd>
 * This is an optional component containing an error message
 * for the command.
 * If the command fires, this message is displayed.
 * </dd>
 * <dd>
 * This string can contain variables of the form
 * <code>$[<i>name</i>]</code>, where <i>name</i> is the
 * name of the variable to be displayed.
 * </dd><dd>
 * The string can also be of the form <code>{{litval}}</code>,
 * where <i>litval</i> is a named literal value.
 * (See the method <b>litVal</b>.)
 * </dd>
 * <dt>GoTo Destination</dt>
 * <dd>
 * This is an optional component that is only used if the
 * command fires.
 * It contains the label of the statement to be executed next
 * when the command fires.
 * If the command doesn't fire (or the destination isn't present), 
 * then the next statement is executed.
 * </dd>
 * <dt>Comment</dt>
 * <dd>
 * This is an optional component that can be used to explain
 * the purpose of the command, etc.
 * </dd>
 * </dl>
 * <p id="commands">
 * The following commands are supported.
 * All string comparisons are case sensitive.
 * </p>
 * <table border="1" cellspacing="0" width="90%">
 * <tr><th>Command</th><th>Syntax</th><th>Description</th></tr>
 * <tr id="cmd_control"><th colspan="3"><i>Control commands</i></th></tr>
 * <tr><td>NOP</td><td><code>nop</code></td><td>A do-nothing command that never fires</td></tr>
 * <tr><td>ALWAYS</td><td><code>always</code></td>
 * 		<td>This command always fires, so any message is displayed,
 * 			and if a <i>GoTo Destination</i> is present,
 * 			it will be the next command to be executed.</td></tr>
 * <tr><td>DONE</td><td><code>done</code></td>
 * 		<td>This command terminates processing normally.
 * 			The status of the record will be set to the current status value set by <i>Result</i>.</td></tr>
 * <tr><td>ABORT</td><td><code>abort</code></td>
 * 	<td>This command terminates with an error condition.
 * 		It is typically used for logic errors, and issues a <i>Message</i>.</td></tr>
 * <tr><td>STAT</td><td><code>stat arg1</code></td>
 * 		<td>This command fires if the current status of the record matches <i>arg1</i></td></tr>
 * <tr><td>WRITE</td><td><code>write</code></td>
 * 		<td>This command causes a normalized form of the record to be written
 * 			to the previously specified output file</td></tr>
 * <tr id="cmd_datadef"><th colspan="3"><i>Data definition commands</i></th></tr>
 * <tr><td>MAP</td><td><code>map arg1 outvar</code></td>
 * 		<td>Map field number <i>arg1</i> to the variable <i>outvar</i>.
 * 		Each time a new record is read, this field is copied to this variable.
 *      <br />The argument <i>arg1</i>can also be:
 *      <b>NFLDS</b>, indicating the number of fields in the input line
 *      or <b>REC</b> for the entire input line
 * 		</td></tr>
 * <tr><td>DECLARE</td><td><code>invar declare arg1 [arg2]</code></td>
 * 		<td>Declare variable <i>invar</i> with data type <i>arg1</i>.
 * 			All variables must be declared, but the declarations don't need
 * 			to appear before they are referenced.
 * 			If <i>arg2</i> is present, the initial value for
 * 			the variable is set to <i>arg2</i></td></tr>
 * <tr id="cmd_assignment"><th colspan="3"><i>Assignment commands</i></th></tr>
 * <tr><td>COPY</td><td><code>invar copy outvar</code></td>
 * 		<td>Copy the contents of the variable <i>invar</i>
 * 			into the variable <i>outvar</i>.
 * 			Both <i>invar</i> and <i>outvar</i> must be the same data type.</td></tr>
 * <tr><td>SET</td><td><code>set arg1 outvar</code></td>
 * 		<td>Copy the value of the literal <i>arg1</i> to the variable <i>outvar</i>.
 * 			The data types of <i>arg1</i> and <i>outvar</i> must be the same.</td></tr>
 * <tr id="cmd_string"><th colspan="3"><i>String manipulation and validation commands</i></th></tr>
 * <tr><td>ASCII</td><td><code>invar ascii</code></td>
 * 		<td>This command fires if any non-ASCII characters are
 * 			found in the string variable <i>invar</i></td></tr>
 * <tr><td>NODATA</td><td><code>nodata <i>arg1</i> <i>arg2</i></code></td>
 * 		<td>This command fires if all of the fields in the specified range are empty
 * 			or contain only whitespace.
 * 			The data types of <i>arg1</i> and <i>arg2</i> must be integer</td></tr>
 * <tr><td>MATCH</td><td><code>invar match arg1</code></td>
 * 		<td>This command fires if the string value of <i>invar</i>
 * 			is the same as the string literal <i>arg1</i>.</td></tr>
 * <tr><td>STARTS</td><td><code>invar starts arg1</code></td>
 * 		<td>This command fires if the string in <i>invar</i>
 * 			begins with the literal string <i>arg1</i>.</td></tr>
  * <tr><td>ENDS</td><td><code>invar ends arg1</code></td>
 * 		<td>This command fires if the string in <i>invar</i>
 * 			ends with the literal string <i>arg1</i>.</td></tr>
 * <tr><td>CONTAINS</td><td><code>invar contains arg1 [outvar]</code></td>
 * 		<td>This command fires if the string in <i>invar</i>
 * 			contains the literal string <i>arg1</i>.
 * 			If <i>outvar</i> is present, the index of the literal
 * 			into the string <i>invar</i> is stored in
 * 			the numeric variable <i>outvar</i>.</td></tr>
 * <tr><td>PATT</td><td><code>invar patt arg1 outvar</code></td>
 * 		<td>This command applies the <a href="#pattern">pattern</a> indicated by <i>arg1</i>
 * 			on the string variable <i>invar</i> and places the resulting
 * 			string in <i>outvar</i>.</td></tr>
 * <tr><td>SUBSTR</td><td><code>invar substr arg1 [arg2] outvar</code></td>
 * 		<td>This command extracts the substring starting at location
 * 			<i>arg1</i> (where zero is the start of the string).
 * 			If <i>arg2</i> is present, the substring terminates at
 * 			that location (the character at that position is not included
 * 			in the output string).  If <i>arg2</i> is not present, then the
 * 			rest of the string is included.</td></tr>
 * <tr><td>APPEND</td><td><code>invar append outvar</code></td>
 * 		<td>The string variable <i>invar</i> is appended to the
 * 			current value of <i>outvar</i></td></tr>
 * <tr><td>REPLACE</td><td><code>invar replace arg1 arg2 outvar</code></td>
 * 		<td>Replace <i>arg1</i> with <i>arg2</i> in <i>invar</i>
 * 			storing the results in <i>outvar</i></td></tr>
 * <tr><td>TRIM</td><td><code>invar trim outvar</code></td>
 * 		<td>Trim leading and trailing spaces in <i>invar</i>
 * 			and store the results in <i>outvar</i>.</td></tr>
 * <tr><td>TRIMALL</td><td><code>trimall arg1</code></td>
 * 		<td>If <i>arg1</i> is "y" or "t", then all fields will be
 * 			trimmed; otherwise none of them will.</td></tr>
 * <tr><td>NORM</td><td><code>invar norm arg1 outvar</code></td>
 * 		<td>This command normalizes the string value of <i>invar</i>
 * 			using the pattern indicated by <i>arg1</i> and places
 * 			the result in <i>outvar</i>.
 * 			If the string value doesn't match the normalization pattern,
 * 			the command will fire.</td></tr>
 * <tr id="cmd_numeric"><th colspan="3"><i>Numeric test and manipulation commands</i></th></tr>
 * <tr><td>MINLEN</td><td><code>invar minlen arg1</code></td>
 * 		<td>This command fires if the length of the string in <i>invar</i>
 * 			is less than the numeric literal <i>arg1</i></td></tr>
 * <tr><td>MAXLEN</td><td><code>invar maxlen arg1</code></td>
 * 		<td>This command fires if the length of the string in <i>invar</i>
 * 			is greater than the numeric literal in <i>arg1</i></td></tr>
 * <tr><td>EQ</td><td><code>invar eq arg1</code></td>
 * 		<td>This command fires if the numeric value of <i>invar</i>
 * 			equals the numeric literal <i>arg1</i></td></tr>
 * <tr><td>LT</td><td><code>invar lt arg1</code></td>
 * 		<td>This command fires if the numeric value of <i>invar</i>
 * 			is less than the numeric literal <i>arg1</i></td></tr>
 * <tr><td>GT</td><td><code>invar gt arg1</code></td>
 * 		<td>This command fires if the numeric variable <i>invar</i>
 * 			is greater than the  numeric literal <i>arg1</i></td></tr>
 * <tr><td>LEN</td><td><code>invar len outvar</code></td>
 * 		<td>This command stores the length of the string variable <i>invar</i>
 * 			into the numeric variable <i>outvar</i></td></tr>
 * <tr><td>INTVAL</td><td><code>invar intval outvar</code></td>
 * 		<td>This command converts the numeric string in <i>invar</i>
 * 			into the numeric variable <i>outvar</i>.
 * 			The command fires if the string is non-numeric.</td></tr>
 * <tr id="cmd_misc"><th colspan="3"><i>Miscellaneous commands</i></th></tr>
 * <tr><td>ZIP</td><td><code>invar zip outvar</code></td>
 * 		<td>This command validates the city/state/zip 
 * 			combination in <i>invar</i>, in the form
 * 			"city|state|zip", with the fields separated by vertical bars.
 * 			The command fires if the zipcode combination 
 * 			is not valid, and <i>outvar</i> 
 * 			will contain an error message</td></tr> 
 * <tr><td>ZIPCITY</td><td><code>invar zipcity outvar</code></td>
 * 			<td>This command finds the cities that are defined for
 * 			the zipcode in <i>invar</i> and places them in <i>outvar</i>.</td></tr>
 * 
 * </table>
 * 
 * <p id="pattern">
 * The following patterns are supported for the <i>match</i> command:
 * </p>
 * <dl>
 * <dt>p</dt>
 * <dd>
 * This is the "normal" pattern, 
 * which replaces characters in the input string with
 * the following characters in the output string
 * <ul>
 * <li>a - any alphabetic character</li>
 * <li>9 - any numeric digit</li>
 * <li>_ - (underscore) any white space character</li>
 * <li>- - (hyphen) any other type of character</li>
 * </ul>
 * The lengths of the input string and output strings
 * will be identical.
 * </dd>
 * <dd>
 * For example, the pattern for the input string "(123)555-1234"
 * will be "-999-999-9999".
 * </dd>
 * <dt>9</dt>
 * <dd>
 * Only the numeric digits in the input string
 * are replaced by a "9". Non-numeric characters are not copied.
 * </dd><dd>
 * For example, "(123)555-1234)" will be changed to "9999999999".
 * </dd>
 * <dt>n</dt>
 * <dd>
 * This pattern will copy only numeric digits to the output string,
 * dropping any non-numeric character.
 * </dd><dd>
 * For example, the pattern "n" on (123)555-1234"
 * will result in "1235551234".
 * </dd>
 * <dt>z<i>N</i></dt>
 * <dd>
 * Similar to pattern "9", except that additional "9" characters
 * are added if the input string is less that "N" digits.
 * </dd><dd>
 * For example, a pattern of "z5" on "123" will result in "99999".
 * The pattern "z5" for the string "123456" will result in "999999".
 * </dd>
 * <dt>n<i>N</i></dt>
 * <dd>
 * Similar to pattern "n", except that leading zeros are supplied
 * to result in a numeric string with at least "N" digits.
 * </dd><dd>
 * For example, the pattern "n5" for the string "123"
 * will result in "00123".
 * The pattern "n5" for the string "123456" will result in "123456".
 * </dd>
 * <dt>a</dt>
 * <dd>
 * This pattern will copy only letters from the input string
 * to the output string, dropping any other type of character.
 * </dd><dd>
 * For example, the pattern of "a" for the string "U.S."
 * will be "US".
 * </dd>
 * <dt>u</dt>
 * <dd>
 * This is similar to the "a" pattern, except that it will
 * replace any lower case character
 * with its upper case character.
 * </dd><dd>
 * For example, the pattern "u" on input string "pa"
 * will result in "PA".
 * </dd>
 * </dl>
 * <p id="norm">
 * The following patterns are supported for the <i>norm</i> command:
 * </p>
 * <dl>
 * <dt>*</dt>
 * <dd>
 * Copy the entire (or remainder) input string directly to the output string
 * </dd>
 * <dt>X</dt>
 * <dd>Copy the current character.</dd>
 * <dt>9</dt>
 * <dd>Copy a numeric digit from the input string</dd>
 * <dd>If the next input character is not a digit, an error is triggered.</dd>
 * <dt>a</dt>
 * <dd>Copy an alphabetic character from the input string.</dd>
 * <dd>If the next input character is not a letter, an error is triggered.</dd>
 * <dt>Any other character</dt>
 * <dd>The pattern character is copied to the output string</dd>
 * </dl>
 * <p>
 * This program assumes that the CSV file has been parsed
 * and stored in an instance of the WrpCsv class.
 * </p>
 * 
 * @author Bill Pringle
 *
 */
public class WrpValidateCsv {

	///////////////////////////////////
	// Class variables
	//////////////////////////////////
	
	// variable type definitions
	/** data type string */
	public final static int TYPE_STR = 1;
	/** data type integer */
	public final static int TYPE_INT = 2;
	/** data type float */
	public final static int TYPE_FLT = 3;
	/** data type boolean */
	public final static int TYPE_BOOL = 4;
	/** data type in error */
	public final static int TYPE_ERR = -3;
	
	
	// field map definitions
	/** entire record */
	public final static int MAP_REC = 0;
	/** number of fields */
	public final static int MAP_NFLDS = -1;
	/** invalid map */
	public final static int MAP_ERR = -2;
	
	// record status
	/** record status is good */
	public final static int REC_GOOD = 0;
	/** record has warnings */
	public final static int REC_WARN = 1;
	/** record is suspicious */
	public final static int REC_FUZZY = 2;
	/** record has errors */
	public final static int REC_BAD = 3;
	/** record is missing */
	public final static int REC_NONE = 4;
	/** record status unknown */
	public final static int REC_ERROR = -1;
	
	/** Used as String representation of unknown / error values */
	public final static String UNKNOWN = "**Unknown**";

	//
	// Class instance variables
	//
	/** the current CSV file that is being validated */
	public WrpCsv csv = null;
	/** flag to trim all fields (default true) */
	public boolean trimall=true;
	/** pattern building object */
	public WrpPattern patt = null;
	/** the set of validation rules to be used for validation */
	public Vector<ValidationRule> rules = null;
	/** array of normalized fields for this record */
	public Vector<String> normalized = new Vector<String>();
	/** output file for normalized record */
	public String normString = null;
	/** line number of this CSV */
	private int csvLine = -1;
	/** previous error line (if any) */
	private int prevErr = -1;
	/** expected header line */
	private String headerLine = null;
	/** parsed header line */ 
	private WrpCsv headerCsv = null;
	
	////////////////////////////////////////
	// compile time variables
	////////////////////////////////////////
	
	/** 
	 * list of statement labels- created during compile time.
	 * This variable contains a list of statement labels
	 * that were found during the first pass of the compile
	 * step.  It is used to reset the IC (instruction counter)
	 * for a GOTO command.
	 * This variable is also tested to determine if the rules
	 * were compiled.  If the value is null, then the rules
	 * must be compiled before they can be executed.
	 * 
	 */
	protected Vector<GoToLoc> labels = null;
	/** list of declared variables */
	protected Vector<Variable> variables=null;
	/** field mapping for each record */
	protected Vector<Variable> maps=null;
	/** defined literals (arg1 and arg2) */
	protected Vector<Variable> literals=null;
	/** compilation errors for rules */
	protected int compileErrors = -1;
	/** list of error messages */
	Vector<String> errMsgs = null;
	/**
	 * previous sequence number.
	 * Used to check for duplicate sequence numbers.
	 * 
	 */
	protected int lastseq = -1;
	
	protected WrpZipCode zipCodes = null;
	
	/**
	 * Log file, if logging is enabled.
	 * This variable can be set by the client using the 
	 * <a href="#setLog(java.io.PrintStream)">setLog(PrintStream)</a> method.
	 */
	PrintStream prtLog = null;
	
	///////////////////////////////////////
	// run time variables
	///////////////////////////////////////
	
	/** run flag - zero means done; non-zero, keep running */
	protected int runFlg = 0;
	/** the number of run-time errors for this CSV validation */
	protected int runErrors = -1;
	/** the total number of run-time errors for this compilation */
	protected int runTot = 0;
	/** instruction counter - location of next instruction to be executed */
	protected int IC=-1;
	/** current status of the current record */
	protected int Stat=REC_ERROR;
	/** 
	 * breakpoint sequence number (place actual breakpoint in executeCommand).
	 * In the executeCommand() routine, there is an if statement
	 * that tests if the current sequence number is equal to the
	 * value of the following variable.
	 * If the two are equal, a debug line is printed.
	 * You can use your own debugger to place a breakpoint
	 * at the print statement, which will pause the program
	 * when the specified line is reached.
	 * You can then use the debugger to examine variable,
	 * begin stepping instructions, etc.
	 * </p><p>
	 *  You can modify the source code so that breakpt isn't
	 *  initialized to -1 (which effectively turns off this feature)
	 *  or you can place a breakpoint at some place within the
	 *  program, and then use the debugger to change the value
	 *  of breakpt.
	 * 
	 */
	protected int breakpt=-1; // break point for debugger in executeCommand


	
	////////////////////////////////////
	// methods
	////////////////////////////////////
	
	/**
	 * Default constructor.
	 * This constructor creates an object without
	 * any rules or CSV string defined.
	 * </p><p>
	 * Once created, the object can be used to validate
	 * CSV strings by first specifying a list of rules
	 * with the <b>setRules</b> method,
	 * and then calling the <b>validate</b> method
	 * for each string.
	 * 
	 */
	public WrpValidateCsv()
	{
		csv = new WrpCsv();
		patt = new WrpPattern();
		rules = new Vector<ValidationRule>();
	}
	
	/**
	 * Constructor with validation rules.
	 * After this constructor creates the validation object,
	 * any number of CSV strings can be validated with
	 * the <b>validate</b> method
	 * 
	 * @param rules the validation rules to use
	 */
	public WrpValidateCsv(Vector<ValidationRule> rules)
	{
		this();
		setRules(rules);
	}
	
	/**
	 * Create a validation object with the given rules and then
	 * validate the given CSV string.
	 * The <i>csvLine</i> parameter is used to determine if the
	 * text of the line should be added to the list of validation messages
	 * See the method <b>logError</b>.
	 * 
	 * @param rules a vector containing the validation rules.
	 * @param csvLine a sequential line number within the CSV line for this string
	 * @param csv the CSV object to be validated
	 * 
	 */
	public WrpValidateCsv(Vector<ValidationRule> rules, int csvLine, WrpCsv csv)
	{
		this();
		validate(rules, csvLine, csv);
	}
	
	/**
	 * A constructor for the ValidationRule subclass.
	 * </p><p>
	 * This method can be used to create a new empty validation rule.
	 * It is most likely used while reading validation rules from
	 * the database (or other source).
	 * 
	 * @return an empty ValidationRule object
	 * 
	 */
	public ValidationRule newRule() { return new ValidationRule(); }
	
	/**
	 * Specify the CSV object to be validated.
	 * </p><p>
	 * This method is called by the <i>validate(WrpCsv csv)</i> 
	 * method, followed by a call to <i>validate()</i>.
	 * It isn't clear why a user would want to call this method,
	 * but at the same time there is no reason to prevent them.
	 * 
	 * @param csv the WrpCsv object that is to be validated
	 * 
	 */
	public void setCsv(WrpCsv csv)
	{
		this.csv = csv;
		prevErr = -1; // reset line number of previous error
	}
	
	/**
	 * Return the current CSV object to be (or was) validated
	 * 
	 * @return the current WrpCsv object for validation
	 * 
	 */
	public WrpCsv getCsv() { return csv; }
	
	/**
	 * Set the CSV line number.
	 * </p><p>
	 * The csvLine variable is used to determine if the original input line
	 * is to be included in the validation messages.
	 * The previous value is stored internally so that the original line
	 * is only included the first time that logError is called with that line number.
	 * This prevents the input line from being repeated if a line contains more
	 * than one validation error (or generates more than one valiation message).
	 * 
	 * @param csvLine
	 */
	public void setCsvLine(int csvLine) 
	{ 
		this.csvLine = csvLine; 
		prevErr = -1; // reset line number of previous error
	}
	
	/**
	 * Get the expected header line.
	 * The first line of the CSV file should contain
	 * the fields contained in this string.
	 * 
	 * @return the expected header line
	 */
	public String getHeaderLine()
	{
		return this.headerLine;
	}
	
	/**
	 * Set the expected header line.
	 * The first line of the CSV file should contain
	 * the same fields as are contained in
	 * the expected header line.
	 * 
	 * @param headerLine
	 */
	public void setHeaderLine(String headerLine)
	{
		this.headerLine = headerLine;
	}
	
	/**
	 * Get the validation status for the record that was
	 * most recently validated.
	 * </p><p>
	 * The status value is normally returned from the 
	 * variout <i>validate</i> methods.
	 * 
	 * @return the saved valiation message
	 * 
	 */
	public int getStat() { return Stat; }
	
	/**
	 * Specify the validation rules to be used when validating lines
	 * 
	 * @param rules a vector of validation rules
	 * 
	 */
	public void setRules(Vector<ValidationRule> rules)
	{
		this.rules = rules;
		labels = null; // must recompile before execution
	}
	
	/**
	 * This method imports a list of ZipCodes that can be used
	 * to validate city/state/zipcode combinations.
	 * 
	 * @param zc a ZipCode structure that has been loaded with ZipCodes.
	 * 
	 */
	public void setZipCodes(WrpZipCode zc)
	{
		zipCodes = zc; // save zip codes
	}
	
	/**
	 * Set the log file for debugging.
	 * This method can be used to toggle debugging on and off.
	 * If the argument is a printstream object, debugging information
	 * will be written to this file.
	 * If the argument is null, then debugging will be disabled.
	 * 
	 * @param log a printstream or null to disable debugging
	 * 
	 */
	public void setLog(PrintStream log)
	{
		prtLog = log;
	}
	
	/**
	 * Get the list of normalized strings.
	 * This method can be used if you want to access the individual
	 * strings.
	 * You could also obtain the same information by getting
	 * the normalized string, which is a CSV line with the same info.
	 * 
	 * @return the normalized
	 */
	public Vector<String> getNormalized()
	{
		return normalized;
	}

	/**
	 * Get the normalized form of the record.
	 * This method can be used to extract a normalized form of
	 * a CSV line.
	 * The normalized form has certain fields formatted in a standardized
	 * manner.
	 * For example, a variety of phone number formats can be accepted
	 * (e.g., 123-456-7890, (123)456-7890, 123.456.7890, etc.)
	 * but the normalized form can be made into a specific format
	 * (e.g., 123-456-7890).
	 * 
	 * @return the output file for a normalized record
	 */
	public String getNormString() {
		return normString;
	}

	/**
	 * Return the number of compile errors that were encountered
	 * during the most recent compilation.
	 * 
	 * @return the number of compile errors, or -1 if no compile
	 * was performed since the validation rules were obtained.
	 * 
	 */
	public int numErrors() { return compileErrors; }
	
	/**
	 * Return the validation messages from the most recent validation.
	 * 
	 * @return a vector of strings containing the validation messages.
	 * 
	 */
	public Vector<String> getErrMsgs() { return errMsgs; }
	
	/**
	 * Set the error messages for this validation.
	 * This method is used when validating a new format lists.
	 * The first line is not a shipper line, but rather a header
	 * We don't want to validate, but need to have a place-holder
	 * for the messages.
	 * 
	 * @param msgs the list of error messages
	 */
	public void setErrMsgs(Vector<String> msgs) { errMsgs = msgs; }
	
	/** get the number of compile-time errors for the most recent compile */
	public int getCompileErrors() { return compileErrors; }
	/** get the number of run-time errors for the latest execution */
	public int getRunErrors() { return runErrors; }
	/** get the total number of run-time errors since compilation */
	public int getRunTot() { return runTot; }
	
	/**
	 * Obtain rules from database and load into list using the given ResultSet.
	 * </p><p>
	 * This method loops through the ResultSet and reads the resulting valiation
	 * rules into the instance vector <b>rules</b>.  Any previous set of rules
	 * are replace by these rules.
	 * 
	 * @param rs RecordSet containing rules query
	 * 
	 */
	public void loadRules(ResultSet rs)
		throws SQLException
	{
		rules = new Vector<ValidationRule>();
		while (rs.next())
		{ // extract fields from query and load into new rule
			ValidationRule rule = new ValidationRule();
			String s;
			int n;
			n = rs.getInt("id");
			rule.setId(n);
			n = rs.getInt("seq");
			rule.setSeq(n);
			s = rs.getString("label");
			if (s!=null) rule.setLabel(s);
			s = rs.getString("invar");
			if (s!=null) rule.setInVar(s);
			s = rs.getString("cmd");
			if (s!=null) rule.setCmd(s);
			s = rs.getString("arg1");
			if (s!=null) rule.setArg1(s);
			s = rs.getString("arg2");
			if (s!=null) rule.setArg2(s);
			s = rs.getString("outvar");
			if (s!=null) rule.setOutVar(s);
			s = rs.getString("result");
			if (s!=null) rule.setResult(s);
			s = rs.getString("msg");
			if (s!=null) rule.setMsg(s);
			s = rs.getString("goto");
			if (s!=null) rule.setGoTo(s);
			s = rs.getString("comment");
			if (s!=null) rule.setComment(s);
			
			rules.add(rule); // add to list
		}
	}
	
	/**
	 * Return the current list of validation rules.
	 * 
	 * @return the current validation rules
	 */
	public Vector<ValidationRule> getRules() { return rules; }
	
	/**
	 * Add a new rule to the current list of rules.
	 * The given rule is added to the instance vector <b>rules</b>.
	 * These additions must be in sequential order, since validation
	 * will be performed sequentially by executing the rules,
	 * starting with the first rule.
	 * 
	 * @param rule
	 */
	public void addRule(ValidationRule rule)
	{
		rules.add(rule);
		labels = null;	// need to recompile rules
	}
	
	/**
	 * Get the current number of rules.
	 * This method returns how many rules are currently stored
	 * for execution.
	 * 
	 * @return the number of rules
	 * 
	 */
	public int getNumberRules()
	{
		return rules.size();
	}
	
	/**
	 * check rules for syntax, build list of destinations
	 * and variables, validate GoTo destinations
	 * </p><p>
	 * Compilation takes place in two passes.
	 * The first pass is performed on all the rules,
	 * followed by the second pass.
	 * <ul>
	 * <li>
	 * The first pass is a syntactical test,
	 * and identifies what labels and variables
	 * have been declared.
	 * </li>
	 * <li>
	 * The second pass validates the semantics of those rules
	 * by checking the data types of arguments and that
	 * all parameters have been declared
	 * </li>
	 * </ul>
	 * 
	 */
	public int compileRules()
	{
		/* initialize lists of labels and variables */ 
		labels = new Vector<GoToLoc>(); // create empty list of destination labels
		variables = new Vector<Variable>(); // create empty list of variables
		literals = new Vector<Variable>(); // create empty list of literal values
		maps = new Vector<Variable>(); // create empty list of mapped fields
		compileErrors = 0;	// no errors yet
		runErrors = -1; // signal execution not started
		lastseq = -1; // no previous statement
		errMsgs = new Vector<String>(); // will hold error messages

		/* first pass creates lists of labels and variables */
		for (int i=0; i<rules.size(); i++)
		{
			firstPass(i);
		}

		/* second pass validates commands and references */
		for (int i = 0; i < rules.size(); i++) {
			/* validate rules and resolve locations */
			secondPass(i);
		}

		return compileErrors;
	}

	/**
	 * Perform first pass of compile
	 * </p><p>
	 * This pass will build lists of statement labels
	 * and declared variables.
	 * It will generate error messages for invalid declare
	 * statements
	 * 
	 * @param ndx index into rules for this rule
	 * 
	 */
	public void firstPass(int ndx)
	{
		ValidationRule rule = rules.get(ndx);
		
		// check for duplicate sequence numbers
		int seq = rule.getSeq();
		if (seq < 0)
		{ // invalid sequence number
			String msg = "Invalid sequence number (" + seq + ")";
			logError(ndx, msg);
		} else if (seq==lastseq)
		{ // duplicate found
			String msg = "Duplicate sequence numbers (" + seq + ")";
			logError(ndx, msg);
		}
		lastseq = seq;

		// check for statement label
		String lbl = rule.getLabel();
		if (lbl.length()>0) storeLabel(ndx);
		String cmd = rule.getCmd().toLowerCase();

		// declare any literals in arg1 and arg2
		String s; // name (and value) of literal
		int loc; // index location of literal in literals

		// resolve literal arg1
		s = rule.getArg1();
		if (s.length()>0)
		{
			loc = findLiteral(s); // locate (or create) literal value
			rule.setArg1Loc(loc);
		}

		// resolve literal arg2
		s = rule.getArg2();
		if (s.length()>0)
		{
			loc = findLiteral(s); // locate (or create) literal value
			rule.setArg2Loc(loc);
		}
		
		// first pass through commands that declare variables
		if (cmd.equals("declare")) declareVariable(ndx);
		if (cmd.equals("map")) mapField(ndx);
	}
	
	/**
	 * Add statement label of this rule to list of labels
	 * 
	 * @param ndx index into rules of the rule
	 */
	public void storeLabel(int ndx)
	{
		ValidationRule rule = rules.get(ndx); // get rule
		String s = rule.getLabel(); // get the statement label
		if (s.length()==0) return; // no label to store - return
		// create a new label object, set the fields,
		// and add it to the list of labels.
		GoToLoc lbl = new GoToLoc(); 
		lbl.setName(s);
		int n = rule.getSeq();
		lbl.setSeq(n);
		lbl.setLoc(ndx);
		labels.add(lbl);
	}
	
	/**
	 * Add variable declaration to list of variables
	 * 
	 * @param ndx index into rules for this declaration
	 * 
	 */
	public void declareVariable(int ndx)
	{
		ValidationRule rule = rules.get(ndx); // get the rule
		String name = rule.getInVar(); // get the variable name
		String type = rule.getArg1(); // get the variable type
		// validate data type
		int n = varType(type);	// convert type to integer
		if (n==TYPE_ERR)
		{ // invalid variable type
			String msg = "Invalid variable type (" + type + ")";
			logError(ndx, msg);
		}
		// check for duplicate declaration
		int loc = findVariable(name);
		if (loc>=0)
		{ // variable already declared
			String msg = "Duplicate variable declaration at line "
				+ ndx + " (" + name + ")";
			logError(ndx, msg);
		}
		
		// declaration valid - create object for variable and add to list
		Variable v = new Variable();
		// set field values for declaration
		v.setName(name);
		v.setType(n);
		
		// an initial value can be specified with arg2
		String s = rule.getArg2();	// initial value
		if (s.length()>0) v.setVal(s);
		
		// add to list
		variables.add(v);
	}
	
	/**
	 * Map a field into a local variable.
	 * </p><p>
	 * This routine is called during the first pass
	 * when a <code>map</code> command is encountered.
	 * Create a new local variable with the specified name
	 * (if it doesn't already exist),
	 * and map it to field <code>arg1</code>, 
	 * with type <code>arg2</code>.
	 * The default field type is String, except for NFLDS,
	 * which is type Integer.
	 * The following special field names are reserved words:
	 * </p>
	 * <ul>
	 * <li>REC</li> &ndash; the entire input string</li>
	 * <li>NFLDS</li> &ndash; the number of input fields</li>
	 * </ul>
	 * <p>
	 * For each new string to be validated, the string is parsed
	 * (if not a WrpCsv object), and the input fields are copied
	 * to the mapped variables.
	 * </p>
	 * 
	 * @param ndx the location of the map statement
	 * 
	 */
	public void mapField(int ndx)
	{
		ValidationRule rule = rules.get(ndx);
		String vname = rule.getOutVar(); // name of variable
		String fname = rule.getArg1().toLowerCase(); // field number
		int fnum = MAP_ERR;
		if (fname.equals("rec"))
		{
			fnum = MAP_REC;
		}
		else if (fname.equals("nflds"))
		{
			fnum = MAP_NFLDS;
		}
		else
		{ // should be numeric value
			try {
				fnum = Integer.parseInt(fname); // convert to field number
			} catch (NumberFormatException e) 
			{ // invalid field number
				fnum = MAP_ERR;
			}
		}
		
		// either create a new variable or use an existing one
		int loc = findVariable(vname); // see if outvar already defined
		Variable v;
		if (loc<0)
		{ // new variable
			v = new Variable();
			v.setName(vname);
			String tname = rule.getArg2();
			// determine field type
			if (tname.length()>0)
			{ // type specified
				int ftype = varType(tname);
				if (ftype==TYPE_ERR)
				{ // invalid type
					String msg = "Invalid field type (" + tname + "}";
					logError(ndx, msg);
				}
				v.setType(ftype);
			}
			else if (fnum == MAP_NFLDS)
			{ // number of fields is type integer
				v.setType(TYPE_INT);
			}
			else
			{ // default type is string
				v.setType(TYPE_STR);
			}
			variables.add(v);	// add variable to list
		}
		else
		{ // variable already defined - use it
			v = variables.get(loc);
		}
		v.setField(true);
		
		//
		// create mapping from field to variable
		//
		Variable p = new Variable(); // variable for mapping
		p.setName(vname);
		p.setIntVal(fnum);
		
		maps.add(p);	// add to the list
	}
	
	/**
	 * Find specified variable in list of variables for the given variable
	 * </p><p>
	 * 
	 * @param vname name of variable to find
	 * @return index of variable, or -1 if not found
	 * 
	 */
	public int findVariable(String vname)
	{
		/*
		 * Loop through list of variables and find a match
		 * If no match found, return -1
		 * 
		 */
		for (int loc=0; loc<variables.size(); loc++)
		{ // loop through the list of variables
			Variable v = variables.get(loc); // get this variable
			String name = v.getName(); // get the name of this vareiable
			if (name.equals(vname)) return loc; // if the same, return location
		}
		
		return -1; // not found, return -1
	}
	
	/**
	 * Find (or create) literal for the given value
	 * </p><p>
	 * Search for the specified literal value.
	 * If not found, create a new literal for it.
	 * Return the location of a variable for this literal
	 * </p><p>
	 * This implies that literals are constant values.
	 * There is currently no command that will allow the user
	 * to modify these values.
	 * </p><p>
	 * Literal values can have encoded values of the form
	 * <code>{{value}}</code>, where <i>value</i> is the
	 * name of the literal code.
	 * See the <b>litVal</b> method.
	 * 
	 * @param litval literal value to find
	 * @return the location in literals for this value
	 * 
	 */
	public int findLiteral(String litval)
	{
		String lval = litVal(litval); // convert literal notations, if any
		int loc; // will contain the location for this literal
		for (loc=0; loc<literals.size(); loc++)
		{ // search for existing literal value
			Variable v = literals.get(loc);
			String lname = v.getName(); // name is value
			if (lval.equals(lname)) return loc; // value found
		}
		
		// first time for literal - create new variable for it
		Variable v = new Variable();
		
		v.setName(lval); // set name of variable
		
		// determine type of literal
		Character c = lval.charAt(0); // get starting character
		if (Character.isDigit(c))
		{ // probably numeric value
			try 
			{
				if (lval.contains("."))
				{ // float value
					v.setType(TYPE_FLT);
					float f = Float.parseFloat(lval);
					v.setFltVal(f);
				}
				else
				{ // integer value
					v.setType(TYPE_INT);
					int n = Integer.parseInt(lval);
					v.setIntVal(n);
				}
			} catch (NumberFormatException e) 
			{ // invalid number - must be string starting with digit
				v.setType(TYPE_STR);
				v.setStrVal(lval);
			}
		} else if (c=='"')
		{ // quoted literal - drop quotes
			String s;
			if (lval.length()==1)
			{
				s = lval;
			}
			else
			{
				// entire string should be quotes, but just in case
				Character c2 = lval.charAt(lval.length()-1);
				if (c2=='"')
				{ // string enclosed in quotes - strip them
					s = lval.substring(1, lval.length()-1);
				}
				else
				{ // retain leading quote, since no trailing one found
					s = lval; 
				}
			}
			v.setType(TYPE_STR);
			v.setStrVal(s);
		}
		else
		{ // string literal
			v.setType(TYPE_STR);
			v.setStrVal(lval);
		}
		literals.add(v);	// add to list
		loc = literals.size() - 1;	// location of last (newly added) element

		return loc; // location of this new literal
	}
	
	/**
	 * Resolve any  variable expressions.
	 * Strings of the form <i>{{str}}</i> are
	 * converted into a character represented by <i>str</i>
	 * </p><p>
	 * Special literal values can be encoded by enclosing
	 * a key word between curly braces.  These notations
	 * are then replaced by the string they represent.
	 * Currently defined literals are:
	 * </p>
	 * <ul>
	 * <li><code>{{quote}}</code> &ndash; a double quote</li>
	 * <li><code>{{apos}}</code> &ndash; a single quote or apostrophe</li>
	 * </ul>
	 * 
	 * @param litval the value of the literal
	 * @return the input argument, if no encoding literals;
	 * 			otherwise, the translated string
	 */
	public String litVal(String litval)
	{
		int loc = litval.indexOf("{{"); // see if any literal values are present
		if (loc<0) return litval; // no encoded value - return argument
		StringBuffer sb = new StringBuffer(); // will be used to build the return value
		
		// literal encoding found
		if (loc > 0)
		{ // copy any characters before the opening braces
			sb.append(litval.substring(0, loc)); // string before ampersand
		}
		loc += 2; // point to start of keyword
		while (loc < litval.length())
		{ // loop processing literal values until the string is exhausted
			int loc2 = litval.indexOf("}}", loc); // end of expression
			if (loc2<0)
			{ // no closing character - return string
				return litval;
			}
			// extract the keyword and insert its value into the output string
			String s = litval.substring(loc, loc2); 
			if (s.equals("quote"))
			{ // double quote
				sb.append("\"");
			}
			else if (s.equals("apos"))
			{ // single quote
				sb.append("'");
			}
			else
			{ // unknown - just append string
				sb.append(s);
			}
			loc = loc2 + 2; // next character after the closing braces
			// if there are no more literal values, just copy the rest of the string
			loc2 = litval.indexOf("{{", loc);
			if (loc2<0)
			{ // no more sequences
				if (loc<litval.length())
				{ // copy remainder of string
					sb.append(litval.substring(loc));
				}
				break;
			}
			// another literal value still exists - continue loop
			loc = loc2;
		}
		// sb should now contain the resulting string
		String ret = sb.toString();
		return ret;
	}
	
	/**
	 * Find the map entry for the specified field
	 * 
	 * @param name the name of the map field to find
	 * @return the index into maps for the field, or -1 if not found
	 * 
	 */
	public int findMap(String name)
	{
		for (int i=0; i<maps.size(); i++)
		{ // look for this field name
			Variable v = maps.get(i); // get variable
			if (v.name.equals(name)) return v.intVal; // return field number
		}
		return -1; // not found
	}
	
	/**
	 * Find target destination of GoTo in list of statement labels
	 * </p><p>
	 * 
	 * @param label the statement label to find
	 * @return the location of the label in labels, or -1 if not found
	 * 
	 */
	public int findGoTo(String label)
	{
		for (int loc=0; loc<labels.size(); loc++)
		{ // loop through list of lables looking for a match
			GoToLoc g = labels.get(loc); // get the label
			String s = g.getName(); // get the label name
			if (label.equals(s))
			{ // label found - return index into rules for this statement
				int ret = g.getLoc();
				return ret;
			}
		}
		
		return -1; // not found
	}
	
	/**
	 * Perform second compile pass
	 * </p><p>
	 * This pass will validate commands
	 * and resolve any references (variables or goto)
	 * by inserting the index to the variable or rule.
	 * It will also detect any undeclared variables
	 * or undefined goto targets.
	 * It will also generate errors for invalid syntax.
	 * 
	 * @param ndx index into rules for the rule to be compiled
	 * 
	 */
	public void secondPass(int ndx)
	{
		ValidationRule rule = rules.get(ndx);
		
		//
		// resolve any variables or locations
		//
		int loc;
		
		// resolve input variable
		String s = rule.getInVar();
		if (s.length()>0)
		{ // resolve input variable
			loc = findVariable(s);	// locate variable
			if (loc<0)
			{ // undefined variable
				String msg = "Undefined input variable (" + s + ")";
				logError(ndx, msg);
			}
			rule.setInVarLoc(loc);
		}
		
		// resolve output variable
		s = rule.getOutVar();
		if (s.length()>0)
		{ // resolve input variable
			loc = findVariable(s);	// locate variable
			if (loc<0)
			{ // undefined variable
				String msg = "Undefined output variable (" + s + ")";
				logError(ndx, msg);
			}
			rule.setOutVarLoc(loc);
		}
		
		// resolve goto destination (if present)
		s = rule.getGoTo();
		if (s.length()>0)
		{ // resolve goto destination
			loc = findGoTo(s);
			if (loc<0)
			{ // undefined destination
				String msg = "Undefined GoTo label (" + s + ")";
				logError(ndx, msg);
			}
			rule.setGoToLoc(loc);
		}
		checkSyntax(ndx);
	}
	
	/**
	 * Check the syntax of the specified command
	 * There is a validation segment for each instruction / command.
	 * 
	 * @param ndx the index into rules for the specified command
	 * 
	 */
	public void checkSyntax(int ndx)
	{
		ValidationRule rule = rules.get(ndx); // get this rule
		String cmd = rule.getCmd().toLowerCase(); // get command name
		
		/*
		 * check for negation.
		 * If a command starts with an exclamation point,
		 * it indicates that the results should be negated.
		 * For syntax checking, ignore this flag.
		 * During executation, the results will be toggled
		 * before being returned.
		 * 
		 */
		if (cmd.startsWith("!"))
		{ // drop leading character */
			cmd = cmd.substring(1); // the actual command
		}
		
		/////////////////////////////////////////
		// control commands
		/////////////////////////////////////////

		// nop - no syntax (always false)
		if (cmd.equals("nop")) return;
		
		// always - no syntax (always true)
		if (cmd.equals("always")) return;
		
		// done - normal termination
		if (cmd.equals("done")) return;
		
		// error - abnormal termination (logic error)
		if (cmd.equals("abort")) return;
		
		// stat - test current status of record
		if (cmd.equals("stat"))
		{ // stat arg1
			chkCmdArg1(ndx);
			return;
		}
		
		// write - create normalized record
		if (cmd.equals("write"))
		{ // write - no arguments
			return;
		}
		
		////////////////////////////////////////
		// data definition commands
		////////////////////////////////////////
		
		// map - map field from record to a variable
		if (cmd.equals("map"))
		{ // map fldnum outvar
			String arg1 = rule.getArg1().toLowerCase(); // field number (0 -> entire record)
			String fldname = rule.getOutVar(); // name of field
			if (fldname.length()==0)
			{ // missing parameter name
				String msg = "Missing parameter name (outvar)";
				logError(ndx, msg);
			}
			if (arg1.length()==0)
			{ // missing field number
				String msg = "Missing field number (arg1)";
				logError(ndx, msg);
			}
			int fldnum = -1;
			if (arg1.equals("rec"))
			{ // map to record
				fldnum = MAP_REC;
			}
			else if (arg1.equals("nflds"))
			{ // number of fields
				fldnum = MAP_NFLDS;
			}
			else
			{ // type should be a number
				try {
					fldnum = Integer.parseInt(arg1);
				} catch (NumberFormatException e) 
				{
					String msg = "Invalid field number (" + arg1 + ")";
					logError(ndx, msg);
				}
			}
			// now check field number
			if (fldnum < MAP_NFLDS)
			{
				String msg = "Invalid field number (" + arg1 + ")";
				logError(ndx, msg);
			}

			return;
		}

		// declare variable - invar declare ftype
		if (cmd.equals("declare"))
		{
			String invar = rule.getInVar();
			String arg1 = rule.getArg1();
			if (invar.length()==0)
			{ // missing variable name
				String msg = "Missing variable name (invar)";
				logError(ndx, msg);
			}
			if (arg1.length()==0)
			{ // missing field type
				String msg = "Missing variable type (arg1)";
				logError(ndx, msg);
			}
			int ftype = varType(arg1);
			if (ftype==TYPE_ERR)
			{ // invalid variable type
				String msg = "Invalid variable type (" + arg1 + ")";
				logError(ndx, msg);
			}
			return;
		}
		
		////////////////////////////////////////
		// assignment commands
		/////////////////////////////////////////
		
		// copy variable - invar copy outvar
		if (cmd.equals("copy"))
		{ // invar copy outvar
			chkInvarCmdOutvar(ndx);
			// invar and outvar should be the same type
			int loc1 = rule.getInVarLoc();
			int loc2 = rule.getOutVarLoc();
			if (loc1>=0 && loc2>=0)
			{ // both present, check type
				Variable v1 = variables.get(loc1);
				int t1 = v1.getType();
				String s1 = varType(t1);
				Variable v2 = variables.get(loc2);
				int t2 = v2.getType();
				String s2 = varType(t2);
				if (t1 != t2)
				{
					String msg = "InVar (" + s1 + ") and OutVar (" + s2 + ") are different data types";
					logError(ndx, msg);
				}
			}
			return;
		}
		
		// assignment - set outvar to literal arg1
		if (cmd.equals("set"))
		{ // set val outvar
			chkCmdArg1Outvar(ndx);
			return;
		}
		
		//////////////////////////////////////
		// String validation and manipulation commands
		////////////////////////////////////////
		
		// ascii - check for non-ASCII characters
		if (cmd.equals("ascii"))
		{ // invar ascii
			String invar = rule.getInVar();
			if (invar.length()==0)
			{ // missing input variable
				String msg = "Missing input variable";
				logError(ndx, msg);
			}
			typeInvar(ndx, TYPE_STR);
			return;
		}
		
		// nodata - no syntax
		if (cmd.equals("nodata"))
		{
			typeArg1(ndx, TYPE_INT);
			typeArg2(ndx, TYPE_INT);
			return;
		}
		
		// match - invar equals string arg1
		if (cmd.equals("match"))
		{ // invar match literal
			chkInvarCmdArg1(ndx);
			// both invar and match literal should be strings
			typeInvar(ndx, TYPE_STR);
			typeArg1(ndx, TYPE_STR);
			return;
		}
		
		// starts - invar starts with string arg1
		if (cmd.equals("starts"))
		{
			chkInvarCmdArg1(ndx);
			typeInvar(ndx, TYPE_STR);
			typeArg1(ndx, TYPE_STR);
			return;
		}
		
		// ends - invar ends with string arg1
		if (cmd.equals("ends"))
		{
			chkInvarCmdArg1(ndx);
			typeInvar(ndx, TYPE_STR);
			typeArg1(ndx, TYPE_STR);
			return;
		}
		
		// contains - invar contains string arg1, location in outvar
		if (cmd.equals("contains"))
		{ // invar cmd patt [outvar]
			chkInvarCmdArg1(ndx); // location in outvar optional
			typeInvar(ndx, TYPE_STR);
			typeArg1(ndx, TYPE_STR);
			// if outvar is used,it must be an integer
			int loc = rule.getOutVarLoc();
			if (loc>=0)
			{
				Variable var = variables.get(loc);
				if (var.type!=TYPE_INT)
				{
					String msg = "OutVar must be integer";
					logError(ndx, msg);
				}
			}
			return;
		}
		
		// patt - apply pattern for invar into outvar
		if (cmd.equals("patt"))
		{ // invar patt code outvar
			chkInvarCmdArg1Outvar(ndx); // arg1 is optional
			// all should be string arguments
			typeInvar(ndx, TYPE_STR);
			typeOutvar(ndx, TYPE_STR);
			int loc = rule.getArg1Loc();
			if (loc>=0)
			{ // pattern given
				Variable v = literals.get(loc);
				if (v.type!=TYPE_STR)
				{
					String msg = "Pattern literal (Arg1) should be type String";
					logError(ndx, msg);
				}
			}
			return;
		}
		
		// substr - build substring of invar using arg1 and arg2
		if (cmd.equals("substr"))
		{ // invar substr start [end] outvar
			chkInvarCmdArg1Outvar(ndx); // arg2 is optional
			typeInvar(ndx, TYPE_STR);
			typeArg1(ndx, TYPE_INT);
			typeOutvar(ndx, TYPE_STR);
			// if arg2 present, it should be integer
			int loc = rule.getArg2Loc();
			if (loc>=0)
			{
				Variable var = literals.get(loc);
				if (var.type!=TYPE_INT)
				{
					String msg = "Arg2 literal should be type Integer";
					logError(ndx, msg);
				}
			}
			return;
		}
		
		// append - append input string to output string
		if (cmd.equals("append"))
		{ // invar append outvar
			typeInvar(ndx, TYPE_STR);
			typeOutvar(ndx, TYPE_STR);
			return;
		}
		
		// replace - change arg1 to arg2 in invar to outvar
		if (cmd.equals("replace"))
		{ // invar replace arg1 arg2 outvar
			typeInvar(ndx, TYPE_STR);
			typeArg1(ndx, TYPE_STR);
			typeArg2(ndx, TYPE_STR);
			typeOutvar(ndx, TYPE_STR);
			// lengths of arg1 and arg2 should be the same
			Variable v1 = getArg1(rule);
			Variable v2 = getArg2(rule);
			String s1 = v1.strVal;
			String s2 = v2.strVal;
			if (s1.length() != s2.length())
			{
				String msg = "Replacement character strings must be the same length";
				logError(ndx, msg);
			}
			return;
		}
		
		// trim - strip leading and trailing spaces
		if (cmd.equals("trim"))
		{ // invar trim outvar
			typeInvar(ndx, TYPE_STR);
			typeOutvar(ndx, TYPE_STR);
			return;
		}
		
		// trimall - global trim flag
		if (cmd.equals("trimall"))
		{ // trimall [arg1]
			int loc = rule.Arg1Loc;
			if (loc >= 0)
			{ // arg1 present, should be boolean
				typeArg1(ndx, TYPE_BOOL);
			}
			return;
		}
		
		// norm - normalize string using pattern
		if (cmd.equals("norm"))
		{ // invar norm patt outvar
			chkInvarCmdArg1Outvar(ndx);
			typeInvar(ndx, TYPE_STR);
			typeArg1(ndx, TYPE_STR);
			typeOutvar(ndx, TYPE_STR);
			// output variable should be mapped to a field
			int loc = rule.getOutVarLoc();
			if (loc>=0)
			{
				Variable var = variables.get(loc);
				if (!var.isField())
				{
					String msg = "Output Variable should be mapped to a field";
					logError(ndx, msg);
				}
			} // if outvar is missing, it was reported by typeOutvar above
			return;
		}
		
		///////////////////////////////////
		// numeric test commands
		////////////////////////////////////
		
		// minlen - invar length less than arg1
		if (cmd.equals("minlen"))
		{ // invar minlen len
			chkInvarCmdArg1(ndx);
			typeInvar(ndx, TYPE_STR);
			typeArg1(ndx, TYPE_INT);
			return;
		}
		
		// maxlen - invar length greater than arg1
		if (cmd.equals("maxlen"))
		{ // invar maxlen length
			chkInvarCmdArg1(ndx);
			typeInvar(ndx, TYPE_STR);
			typeArg1(ndx, TYPE_INT);
			return;
		}
		
		// eq - invar numerically equal to arg1
		if (cmd.equals("eq"))
		{ // invar eq val
			chkInvarCmdArg1(ndx);
			typeInvar(ndx, TYPE_INT);
			typeArg1(ndx, TYPE_INT);
			return;
		}
		
		// lt - invar numerically less than arg1
		if (cmd.equals("lt"))
		{ // invar lt val
			chkInvarCmdArg1(ndx);
			typeInvar(ndx, TYPE_INT);
			typeArg1(ndx, TYPE_INT);
			return;
		}
		
		// gt - invar numerically greater than arg1
		if (cmd.equals("gt"))
		{ // invar gt val
			chkInvarCmdArg1(ndx);
			typeInvar(ndx, TYPE_INT);
			typeArg1(ndx, TYPE_INT);
			return;
		}
		
		// len - store length of invar into outvar
		if (cmd.equals("len"))
		{
			chkInvarCmdOutvar(ndx);
			typeInvar(ndx, TYPE_STR);
			typeOutvar(ndx, TYPE_INT);
			return;
		}
		
		// intval - convert string to integer
		if (cmd.equals("intval"))
		{ // invar intval outvar
			chkInvarCmdOutvar(ndx);
			typeInvar(ndx, TYPE_STR);
			typeOutvar(ndx, TYPE_INT);
			return;
		}
		
		/////////////////////////////
		// Misc. Commands
		/////////////////////////////
		if (cmd.equals("zip") || cmd.equals("zipcity"))
		{ // city|state|zip zip
			typeInvar(ndx, TYPE_STR);
			typeOutvar(ndx, TYPE_STR);
			return;
		}		
		
		//
		// unknown command
		//
		String msg = "Unknown command (" + cmd + ")";
		logError(ndx, msg);
	}

	/////////////////////////////////////
	// routines to check command syntax
	////////////////////////////////////////


	/**
	 * Check syntax for form: invar cmd arg1
	 * This command can be used for any type of literal.
	 * If the argument should be numeric or string,
	 * then test for the specific forms
	 * 
	 * @param ndx the index into rules for this rule 
	 * 
	 */
	public void chkInvarCmdArg1(int ndx)
	{
		ValidationRule rule = rules.get(ndx);
		String invar = rule.getInVar();
		String arg1 = rule.getArg1();
		if (invar.length()==0)
		{ // missing input variable
			String msg = "Missing input variable";
			logError(ndx, msg);
		}
		if (arg1.length()==0)
		{ // missing length
			String msg = "Missing literal (arg1)";
			logError(ndx, msg);
		}
	}
	
	/**
	 * Check syntax for form: invar cmd outvar
	 * 
	 * @param ndx index into rules for this rule
	 * 
	 */
	public void chkInvarCmdOutvar(int ndx)
	{
		ValidationRule rule = rules.get(ndx);
		String invar = rule.getInVar();
		String outvar = rule.getOutVar();
		if (invar.length()==0)
		{ // missing input variable
			String msg = "Missing input variable";
			logError(ndx, msg);
		}
		if (outvar.length()==0)
		{ // missing output variable
			String msg = "Missing output variable";
			logError(ndx, msg);
		}
	}
	
	/**
	 * Check syntax for form: invar cmd arg1 outvar
	 * 
	 * @param ndx index into rules for this rule
	 *  
	 */
	public void chkInvarCmdArg1Outvar(int ndx)
	{
		chkInvarCmdOutvar(ndx);
		ValidationRule rule = rules.get(ndx);
		String arg1 = rule.getArg1();
		if (arg1.length()==0)
		{ // missing arg1
			String msg = "Missing Arg1 parameter";
			logError(ndx, msg);
		}
	}
	
	/**
	 * Check syntax for the form: cmd arg1
	 * 
	 * @param ndx index into rules for this rule
	 * 
	 */
	public void chkCmdArg1(int ndx)
	{
		ValidationRule rule = rules.get(ndx);
		String arg1 = rule.getArg1();
		if (arg1.length()==0)
		{ // missing arg1
			String msg = "Missing Arg1 parameter";
			logError(ndx, msg);
		}
	}
	
	/**
	 * Check syntax for form: cmd arg1 outvar
	 * 
	 * @param ndx index into rules for this rule
	 * 
	 */
	public void chkCmdArg1Outvar(int ndx)
	{
		ValidationRule rule = rules.get(ndx);
		String arg1 = rule.getArg1();
		if (arg1.length() == 0)
		{
			String msg = "Literal (Arg1) missing";
			logError(ndx, msg);
		}
		String var = rule.getOutVar();
		if (var.length() == 0)
		{
			String msg = "OutVar is missing";
			logError(ndx, msg);
		}
	}
	
	//////////////////////////////////////////////////
	// routines to check data type
	//////////////////////////////////////////////////

	/**
	 * Check data type for InVar
	 * 
	 * @param ndx index into variables for this variable
	 * @param type expected data type
	 * 
	 */
	public void typeInvar(int ndx, int type)
	{
		ValidationRule rule = rules.get(ndx);
		int loc = rule.getInVarLoc();
		if (loc<0)
		{
			String msg = "InVar missing";
			logError(ndx, msg);
			return;
		}
		Variable var = variables.get(loc);
		int t = var.getType();
		if (t!=type)
		{
			String msg = "InVar should be type " + varType(type)
				+ ", not " + varType(t);
			logError(ndx, msg);
		}
	}
	
	/**
	 * Check data type for literal Arg1
	 * 
	 * @param ndx index into literals for element
	 * @param type expected data type
	 * 
	 */
	public void typeArg1(int ndx, int type)
	{
		ValidationRule rule = rules.get(ndx);
		int loc = rule.getArg1Loc();
		if (loc<0)
		{
			String msg = "Arg1 missing";
			logError(ndx, msg);
			return;
		}
		Variable var = literals.get(loc);
		int t = var.getType();
		if (t!=type)
		{
			String msg = "Literal Arg1 should be type "
				+ varType(type) + ", not " + varType(t);
			logError(ndx, msg);
		}
	}
	
	/**
	 * Check data type for literal Arg2
	 * 
	 * @param ndx index into literals for this element
	 * @param type expected data type
	 * 
	 */
	public void typeArg2(int ndx, int type)
	{
		ValidationRule rule = rules.get(ndx);
		int loc = rule.getArg2Loc();
		if (loc<0)
		{
			String msg = "Arg2 missing";
			logError(ndx, msg);
			return;
		}
		Variable var = literals.get(loc);
		int t = var.getType();
		if (t!=type)
		{
			String msg = "Literal Arg2 should be type "
				+ varType(type) + ", not " + varType(t);
			logError(ndx, msg);
		}
	}
	
	/**
	 * Check data type for OutVar
	 * 
	 * @param ndx index into variables for element
	 * @param type expected data type
	 * 
	 */
	public void typeOutvar(int ndx, int type)
	{
		ValidationRule rule = rules.get(ndx);
		int loc = rule.getOutVarLoc();
		if (loc<0)
		{
			String msg = "OutVar missing";
			logError(ndx, msg);
			return;
		}
		Variable var = variables.get(loc);
		int t = var.getType();
		if (t!=type)
		{
			String msg =  "OutVar should be type "
				+ varType(type) + ", not " + varType(t);
			logError(ndx, msg);
		}
	}
	
	/**
	 * Validate supplied CSV using supplied rules
	 * </p><p>
	 * The rules should be compiled before this routine is called.
	 * The supplied rules are executed against the supplied CSV
	 * 
	 * @return the number of run-time validation errors
	 * @throws ValidationRuntimeException
	 */
	private int executeRules() throws ValidationRuntimeException
	{
		/****************
		errMsgs = new Vector<String>();
		normString = "";
		*****************/

		if (labels==null)
		{
			compileRules(); // compile rules before execution
		}
		if (compileErrors>0) return -1; // don't execute if compile errros
		runErrors = 0; // reset error count
		
		// map fields to variables
		mapFields();
		
		//
		// main execution loop
		//
		IC = 0; // entry point is starting location
		runFlg = 1; // run flag - keep running
		Stat = REC_GOOD; // reset record flag
		while (runFlg>0)
		{
			int thisCmd = IC; // current instruction to execute
			if (thisCmd < 0)
			{ // serious error
				String s = "Execution error - IC negative.";
				String info = "[" + thisCmd + "]";
				logError(thisCmd, s);
				throw new ValidationRuntimeException(info, s);
			}
			else if (thisCmd >= rules.size())
			{ // ran off end of program - terminate quietly
				break;
			}
			IC++; // assume following instruction is next
			Boolean res = executeCommand(thisCmd); // execute current command
			if (res) fireCommand(thisCmd);
		}

		runTot += runErrors;
		return runErrors; // number of errors for this CSV
	}
	
	/**
	 * Execute the current list of rules, with the current line number
	 * </p><p>
	 * The line number can be used if the same compilation will be used
	 * to execute against multiple CSV entries.
	 * The line number is used to determine if the input line
	 * should be displayed.
	 * 
	 * @param csvLine the line number for the current CSV entry
	 * @return the record status of the CSV
	 * 
	 */
	public int executeRules(int csvLine)
	{
		this.csvLine = csvLine;
		int ret = executeRules();
		return ret;
	}
	
	/**
	 * Map the defined fields in CSV to variables.
	 * Loop through the defined field maps and copy
	 * the corresponding CSV field to the mapped variable.
	 * 
	 */
	public void mapFields()
	{
		// create array for normalized strings
		normalized = new Vector<String>();
		/** TODO - should use max map value? */
		for (int ndx=0; ndx<maps.size(); ndx++)
		{ // initialize to empty strings
			/** TODO - should initialize to input field? */
			normalized.add("");
		}
		
		// loop through maps and store fields from CSV into variables
		int nFlds = csv.getNumFields(); // how many fields in record
		for (int ndx=0; ndx<maps.size(); ndx++)
		{ // loop through defined maps
			Variable var = maps.get(ndx);
			int fldNum = var.intVal;
			String fldName = var.name;
			int loc = findVariable(fldName); // locate variable
			if (loc < 0)
			{ // variable name not found - internal error
				String msg = "Invalid field name [" + fldName + "] in MAP command";
				logError(ndx, msg);
				runErrors++;
				continue;
			}
			Variable fld = variables.get(loc);
			if (fldNum > nFlds)
			{ // not enough fields for this - clear value
				fld.strVal = ""; // empty field
			}
			else if (fldNum == MAP_REC)
			{ // entire record
				String s = csv.getLine();
				fld.strVal = s;
			}
			else if (fldNum == MAP_NFLDS)
			{ // number of fields
				fld.intVal = nFlds;
			}
			else if (fldNum > MAP_REC)
			{ // field within range
				String s = csv.getField(fldNum-1); // get CSV field
				if (trimall) s = s.trim(); // trim field if required
				fld.strVal = s; // store field
				normalized.setElementAt(s, fldNum-1);
			}
			else
			{ // error
				String msg = "MAP - invalid field number (" + fldNum + ")";
				logError(ndx, msg);
			}
		}
	}
	
	/**
	 * Execute the current command
	 * 
	 * @param ndx index into Rules of the command to execute
	 * @return the result of the command (true means the command fired)
	 * 
	 */
	public Boolean executeCommand(int ndx) throws ValidationRuntimeException
	{
		Boolean ret = false; // assume it doesn't fire
		Boolean notFlg = false; // negation flag
		
		ValidationRule rule = rules.get(ndx);
		String cmd = rule.getCmd().toLowerCase();
		int seq = rule.getSeq(); // sequence number
		
		//////////////////////////////////////////////////
		// check for debugging break point
		//
		// If current sequence number matches breakpt (default -1),
		// the following if statement will trigger.
		// place a breakpoint on the "System.out.println" statement
		// 
		// You can update the value of breakpt during runtime
		// using the debugger (rather than the source code)
		//////////////////////////////////////////////////
		if (breakpt == seq)
		{
			System.out.println("Seq " + seq + " (" + cmd + ")");
		}
		
		// check for not flag
		if (cmd.startsWith("!"))
		{ // negation - toogle result
			notFlg = true;
			cmd = cmd.substring(1); // drop ! flag
		} else notFlg = false;
		
		/*
		 * Execute the actual command.
		 * The following code is a series of chained if statements.
		 * 
		 */
		if (cmd.equals("nop"))
		{
			ret = false; // always false
		}
		else if (cmd.equals("always"))
		{
			ret = true; // always true
		}
		else if (cmd.equals("done"))
		{ // terminate execution normally
			runFlg = 0;
			ret = true;
		}
		else if (cmd.equals("abort"))
		{ // abnormal termination
			runFlg = 0;
			logError(ndx, "Abort instruction encountered");
			ret = true; // fire command (issue message if any)
			runErrors++; // considered run-time error
		}
		else if (cmd.equals("stat"))
		{ // test current record status
			String s = rule.getArg1(); // status to check for
			int n = recStat(s);			// convert to binary
			if (n==Stat) ret = true; // status matches
		}
		else if (cmd.equals("write"))
		{ // write normalized form to output string
			StringBuffer sb = new StringBuffer();
			for (int i=0; i<normalized.size(); i++)
			{ // copy SQL-safe form of field to output string
				String s = normalized.get(i); // get string
				if (i>0) sb.append(","); // add delimiter
				sb.append("\"" + s + "\""); // add the field
			}
			normString = sb.toString(); // user can get this with getNormString()
		}
		else if (cmd.equals("map"))
		{
			ret = false; // done at compile time
		}
		else if (cmd.equals("declare"))
		{
			ret = false; // compile instruction
		}
		else if (cmd.equals("copy"))
		{ // copy invar to outvar
			Variable invar = getInVar(rule);
			Variable outvar = getOutVar(rule);
			ret = !outvar.set(invar); // trigger on exception
		}
		else if (cmd.equals("set"))
		{ // set outvar value to arg1
			Variable arg1 = getArg1(rule);
			Variable outvar = getOutVar(rule);
			ret = !outvar.set(arg1); // trigger on exception
		}
		else if (cmd.equals("ascii"))
		{ // check for non-ASCII characters
			Variable invar = getInVar(rule);
			String s = invar.strVal;
			ret = ascii(s); // trigger on non-ascii found
		}
		else if (cmd.equals("nodata"))
		{ // test if any fields contain data
			Variable arg1 = getArg1(rule);
			Variable arg2 = getArg2(rule);
			int n1 = arg1.getIntVal();
			int n2 = arg2.getIntVal();
			if (n1 < 1) n1 = 1; // lowest possible value
			int n3 = 1;
			if (csv!=null) n3 = csv.getNumFields(); // max number of fields
			if (n2 < 1 || n2 > n3) n2 = n3; 
			ret = nodata(n1, n2);
		}
		else if (cmd.equals("match"))
		{ // see if invar matches arg1
			Variable invar = getInVar(rule);
			Variable arg1 = getArg1(rule);
			String s1 = invar.strVal;
			String s2 = arg1.strVal;
			if (s1==null || s2==null)
			{
				nullArg(ndx);
			}
			ret = s1.equals(s2);
		}
		else if (cmd.equals("starts"))
		{ // see if invar starts with arg1
			Variable invar = getInVar(rule);
			Variable arg1 = getArg1(rule);
			String s1 = invar.strVal;
			String s2 = arg1.strVal;
			if (s1==null || s2==null)
			{
				nullArg(ndx);
			}
			ret = s1.startsWith(s2);
		}
		else if (cmd.equals("ends"))
		{ // see if invar ends with arg1
			Variable invar = getInVar(rule);
			Variable arg1 = getArg1(rule);
			String s1 = invar.strVal;
			String s2 = arg1.strVal;
			if (s1==null || s2==null)
			{
				nullArg(ndx);
			}
			ret = s1.endsWith(s2);
		}
		else if (cmd.equals("contains"))
		{ // see if invar contains arg1.  If yes, put loc in outvar
			Variable invar = getInVar(rule);
			Variable arg1 = getArg1(rule);
			String s1 = invar.strVal;
			String s2 = arg1.strVal;
			if (s1==null || s2==null)
			{
				nullArg(ndx);
			}
			int loc = s1.indexOf(s2);
			if (loc>=0)
			{ // string found - save location if outvar is present
				ret = true;
				Variable outvar = getOutVar(rule);
				if (outvar!=null) outvar.intVal = loc;
			}
		}
		else if (cmd.equals("patt"))
		{ // apply pattern arg1 to invar, placing results in outvar
			Variable invar = getInVar(rule);
			Variable arg1 = getArg1(rule);
			Variable outvar = getOutVar(rule);
			String s = invar.strVal;
			String p = arg1.strVal;
			if (s==null || p==null)
			{
				nullArg(ndx);
			}
			String r = WrpPattern.applyPattern(s, p); // apply the pattern
			outvar.strVal = r;
		}
		else if (cmd.equals("substr"))
		{ // apply substring to invar and place results in outvar
			Variable invar = getInVar(rule);
			Variable arg1 = getArg1(rule);
			Variable arg2 = getArg2(rule);
			Variable outvar = getOutVar(rule);
			String strIn = invar.strVal; // input string
			if (strIn==null)
			{
				nullArg(ndx);
			}
			String strOut=null; // output string
			int loc = arg1.intVal;
			if (loc<0)
			{ // invalid starting location
				String msg = "Invalid starting location (" + loc + ")";
				logError(ndx, msg);
				strOut = ""; // return empty string
				ret = true; // fire if invalid parms
			}
			else
			{ // starting location should be less than the length of the string
				if (loc>= strIn.length()) strOut = ""; // string too short
				else if (arg2==null)
				{ // rest of string from start
					strOut = strIn.substring(loc);
				}
				else
				{ // start and end locations given
					int loc2 = arg2.intVal;
					if (loc2 >= strIn.length())
					{ // end location past end of string - copy rest of string
						strOut = strIn.substring(loc);
					}
					else
					{ // extract substring
						strOut = strIn.substring(loc, loc2);
					}
				}
			}
			outvar.strVal = strOut;
		}
		else if (cmd.equals("append"))
		{ // append the contents of invar to the end of outvar
			Variable invar = getInVar(rule);
			Variable outvar = getOutVar(rule);
			String s1 = invar.strVal;
			String s2 = outvar.strVal;
			outvar.strVal = s2 + s1; // append the two values
		}
		else if (cmd.equals("replace"))
		{ // replace chars in arg1 with chars in arg2 in invar, results in outvar
			Variable invar = getInVar(rule);
			Variable outvar = getOutVar(rule);
			Variable arg1 = getArg1(rule);
			Variable arg2 = getArg2(rule);
			String s1 = invar.strVal;
			String cs1 = arg1.strVal;
			String cs2 = arg2.strVal;
			if (s1==null || cs1==null || cs2==null)
			{
				nullArg(ndx);
			}
			// both strings must be the same length
			if (cs1.length()==cs2.length())
			{
				for (int i=0; i<cs1.length(); i++)
				{ // replace c1 with c2
					char c1 = cs1.charAt(i);
					char c2 = cs2.charAt(i);
					s1 = s1.replace(c1, c2);
				}
			}
			else
			{ // invalid replacement strings - return input string
				/*
				 * we should never get here.
				 * if the two strings are not the same length,
				 * a compile error is generated, so execution shouldn't happen
				 * 
				 */
				ret = true; // trigger exception
			}
			// store results in outvar
			outvar.strVal = s1;
		}
		else if (cmd.equals("trim"))
		{ // trim leading and trailing spaces
			Variable invar = getInVar(rule);
			String s = invar.strVal;
			if (s==null)
			{
				nullArg(ndx);
			}
			String t = s.trim();
			Variable outvar = getOutVar(rule);
			outvar.strVal = t;
		}
		else if (cmd.equals("trimall"))
		{ // set global flag to trim all fields
			Variable arg1 = getArg1(rule);
			if (arg1==null) trimall = true; // default is true
			else trimall = arg1.boolVal;
		}
		else if (cmd.equals("norm"))
		{ // normalize invar using arg1 and store results in outvar
			Variable invar = getInVar(rule);
			Variable arg1 = getArg1(rule);
			Variable outvar = getOutVar(rule);
			String strIn = invar.strVal;
			String strPatt = arg1.strVal;
			if (strIn==null || strPatt==null)
			{
				nullArg(ndx);
			}
			String strOut=norm(ndx, strIn, strPatt);
			outvar.strVal = strOut;
			int mapFld = findMap(invar.name); // find map for this field
			if (mapFld>=0) normalized.setElementAt(strOut, mapFld);
		}
		else if (cmd.equals("minlen"))
		{ // trigger command if invar less than arg1
			Variable invar = getInVar(rule);
			String s = invar.strVal;
			int n;
			if (s==null)
			{
				n = 0;
			}
			else
			{
				n = s.length();
			}
			Variable arg1 = getArg1(rule);
			int m = arg1.intVal;
			if (n < m) ret = true; // trigger
		}
		else if (cmd.equals("maxlen"))
		{ // trigger command if invar greater than arg1
			Variable invar = getInVar(rule);
			String s = invar.strVal;
			int n;
			if (s==null)
			{
				n = 0;
			}
			else
			{
				n = s.length();
			}
			Variable arg1 = getArg1(rule);
			int m = arg1.intVal;
			if (n > m) ret = true; // trigger
		}
		else if (cmd.equals("eq"))
		{ // trigger if invar and arg1 are numerically equals
			Variable invar = getInVar(rule);
			int n1 = invar.intVal;
			Variable arg1 = getArg1(rule);
			int n2 = arg1.intVal;
			if (n1 == n2) ret = true; // trigger
		}
		else if (cmd.equals("lt"))
		{ // trigger if invar is numerically less than arg1
			Variable invar = getInVar(rule);
			int n1 = invar.intVal;
			Variable arg1 = getArg1(rule);
			int n2 = arg1.intVal;
			if (n1 < n2) ret = true; // trigger
		}
		else if (cmd.equals("gt"))
		{ // trigger if invar numerically greater than arg1
			Variable invar = getInVar(rule);
			int n1 = invar.intVal;
			Variable arg1 = getArg1(rule);
			int n2 = arg1.intVal;
			if (n1 > n2) ret = true; // trigger
		}
		else if (cmd.equals("len"))
		{ // store length of invar in outvar
			Variable invar = getInVar(rule);
			String s = invar.strVal;
			Variable outvar = getOutVar(rule);
			int n = s.length(); // length of invar
			outvar.intVal = n;
		}
		else if (cmd.equals("intval"))
		{ // store integer value of invar in outvar
			Variable invar = getInVar(rule);
			Variable outvar = getOutVar(rule);
			String s = invar.strVal;
			try {
				int n = Integer.parseInt(s);
				outvar.intVal = n;
			} catch (NumberFormatException e) {
				logError(ndx, "Invalid numeric string [" + s + "]");
				ret = true; // trigger command
			}
		}
		else if (cmd.equals("zip"))
		{ // valiate city/state/zipcode in invar, send message to outvar
			Variable invar = getInVar(rule);
			Variable outvar = getOutVar(rule);
			if (zipCodes==null)
			{ // zip codes not loaded - can't validate
				outvar.strVal = "ZipCodes not initalized - use setZipCodes()";
				ret = true;
			}
			else
			{
				String s = invar.strVal;
				if (s==null || s.length()==0)
				{
					outvar.strVal = "Zip: No ZipCode combination given";
					ret = true;
				}
				else
				{
					ret = validateZipCode(s, outvar);
				}
			}
		}
		else if (cmd.equals("zipcity"))
		{
			Variable invar = getInVar(rule);
			Variable outvar = getOutVar(rule);
			if (zipCodes==null)
			{ // zip codes not loaded - can't validate
				outvar.strVal = "ZipCodes not initalized - use setZipCodes()";
				ret = true;
			}
			else
			{
				String s = invar.strVal;
				if (s==null || s.length()==0)
				{
					outvar.strVal = "Zip: No ZipCode given";
					ret = true;
				}
				else
				{
					WrpZipCode zc = zipCodes.findZipCode(s);
					if (zc==null)
					{
						outvar.strVal = "ZipCode (" + s + ") not defined.";
						ret = true;
					}
					else
					{
						StringBuffer sb = new StringBuffer("ZipCode " + s + " defined in " 
								+ zc.getState() + " for city(s): ");
						Vector<String> cities = zc.getCities();
						for (int i=0; i<cities.size(); i++)
						{
							String c = cities.get(i);
							if (i>0) sb.append(", ");
							sb.append(c);
						}
						String outstr = sb.toString();
						outvar.strVal = outstr;
						ret = true;
					}
				}
			}			
		}
		else
		{ // unknown command
			logError(ndx, "Unknown command [" + cmd + "]");
			runErrors++;
		}
		if (notFlg) ret = !ret; // toggle results
		if (ret && prtLog !=  null)
		{
			prtLog.println(rule + ", ret=" + ret);
		}
		return ret;
	}
	
	/**
	 * Fire the current command
	 * 
	 * If a command returns true, then the command fired.
	 * When this happens, the following happens:
	 * <ul>
	 * <li>
	 * If a result is present, the record status is updated
	 * (if the result is worse than the current status)
	 * </li>
	 * <li>any message that is present is displayed</li>
	 * <li>
	 * If a GoTo Destination is present, then that will be
	 * the next instruction to execute.
	 * </li>
	 * </ul>
	 * 
	 */
	public void fireCommand(int ndx)
	{
		ValidationRule rule = rules.get(ndx);
		
		// set status if worse than current
		String stat = rule.getResult(); // record status (if any)
		if (stat.length()>0)
		{ // set status if worse than current status
			int n = recStat(stat);
			if (n>Stat) Stat = n; // set record status
		}
		
		// display message if present
		String msg = rule.getMsg();
		if (msg.length()>0)
		{
			String s = decodeMsg(ndx, msg);
			logError(ndx, s);
		}
		
		// set next location if GoTo destination present
		int n = rule.getGoToLoc(); // location of next instruction
		if (n>=0) IC = n; // valid destination - set it
	}
	
	/**
	 * Replace variable names in a string with their values.
	 * </p><p>
	 * If a string of the form "$[varname]" is present,
	 * replace it with the current value of the variable.
	 * 
	 * @param ndx index into rules for the current rule
	 * @param msg text of the message string to be decoded
	 * @return the resulting string, with values replaces
	 * 
	 */
	public String decodeMsg(int ndx, String msg)
	{
		ValidationRule rule = rules.get(ndx);
		int loc1=0;
		int loc2 = msg.indexOf("$[");
		if (loc2<0 && loc1 < msg.length())
		{ // no insert - return input string
			return msg;
		}
		
		// variable expression found - loop through and process
 		StringBuffer sb = new StringBuffer();
 		while (loc2>=0 && loc1 < msg.length())
 		{ // variable found - substitute value
 			sb.append(msg.substring(loc1, loc2)); // segment before variable
 			loc1 = loc2 + 2; // start of variable name
 			loc2 = msg.indexOf("]", loc1); // end of variable name
 			if (loc2<0)
 			{ // not found - problems
 				logError(ndx, "Variable name not formatted correctly [" + msg + "]");
 				runErrors++;
 				return msg; // return entire text
 			}
 			String varname = msg.substring(loc1, loc2).toLowerCase(); // variable name
 			Variable v=null;
 			// check for argument names (e.g., invar, arg1, etc.)
 			if (varname.equals("invar"))
 			{
 				v = getInVar(rule);
 			}
 			else if (varname.equals("arg1"))
 			{
 				v = getArg1(rule);
 			}
 			else if (varname.equals("arg2"))
 			{
 				v = getArg2(rule);
 			}
 			else if (varname.equals("outvar"))
 			{
 				v = getOutVar(rule);
 			}
 			else
 			{ // not an argument - try a variable
 				int n = findVariable(varname);
 				if (n>=0) v = variables.get(n);
 			}
 			if (v==null)
 			{ // unknown variable - leave unchanged
 				String s = "Unknown variable [" + varname + "]";
 				logError(ndx, s);
 				sb.append("$[" + varname + "]");
 			}
 			else
 			{ // append string value of variable
 				int t = v.type;
 				String s;
 				switch (t)
 				{
 				case TYPE_STR: s = v.strVal; break;
 				case TYPE_INT: s = "" + v.intVal; break;
 				case TYPE_FLT: s = "" + v.fltVal; break;
 				case TYPE_BOOL: s = (v.boolVal ? "y" : "n"); break;
 				default: s = UNKNOWN; break;
 				}
 				sb.append(s);
 			}
 			
 			// variable substituted - search for another one
 			loc1 = loc2+1; // next character after variable
 			loc2 = msg.indexOf("$[", loc1);
 			if (loc2<0)
 			{ // not more variables - append rest of message
 				sb.append(msg.substring(loc1));
 			}
 		}
		
		String ret = sb.toString();
		return ret;
	}
	
	///////////////////////////////////////////
	// run-time routines
	/////////////////////////////////////////////
	
	/**
	 * Obtain the InVar variable for this rule
	 * 
	 * @param rule the rule containing the specified invar
	 * @return Variable for InVar
	 * 
	 */
	private Variable getInVar(ValidationRule rule)
	{
		int loc = rule.getInVarLoc();
		if (loc<0) return null;
		Variable ret = variables.get(loc);
		return ret;
	}
	
	/**
	 * Obtain the OutVar variable for the specified rule
	 * 
	 * @param rule the rule containing the OutVar
	 * @return Variable for OutVar
	 * 
	 */
	private Variable getOutVar(ValidationRule rule)
	{
		int loc = rule.getOutVarLoc();
		if (loc<0) return null;
		Variable ret = variables.get(loc);
		return ret;
	}
	
	/**
	 * Get the Arg1 variable for the specified rule
	 * 
	 * @param rule the rule containing the Arg1 literal
	 * @return Variable for Arg1
	 */
	private Variable getArg1(ValidationRule rule)
	{
		int loc = rule.getArg1Loc();
		if (loc<0) return null;
		Variable ret = literals.get(loc);
		return ret;
	}
	
	/**
	 * Get the Arg2 variable for the specified rule
	 * 
	 * @param rule the rule containing the Arg2 literal
	 * @return Variable for Arg2
	 */
	private Variable getArg2(ValidationRule rule)
	{
		int loc = rule.getArg2Loc();
		if (loc<0) return null;
		Variable ret = literals.get(loc);
		return ret;
	}
	
	/**
	 * Test the input string for non-ASCII characters
	 * 
	 * @param fld input string to be tested
	 * @return true if a non-ASCII character found; false otherwise
	 * 
	 */
	public boolean ascii(String fld)
	{
		if (fld==null) return false; // nothing in field

		Boolean ret = false;	// assume all ascii
		Boolean debugSw = false;

		byte[] b = fld.getBytes();
		CharsetDecoder d = Charset.forName("US-ASCII").newDecoder();
		try
		{ // attempt to decode as ASCII
			CharBuffer r = d.decode(ByteBuffer.wrap(b));
			if (debugSw) System.out.println("Output string [" + r + "]");
		} catch (CharacterCodingException e)
		{ // non-ASCII character found
			if (debugSw) System.err.println("applyASCII: non-ASCII character [" + fld + "]");
			ret = true;
		}

		return ret;
	}
	
	/**
	 * Determine if the specified fields are empty or contain only white space
	 * 
	 * @param n1 lowest field to check (base 1)
	 * @param n2 highest field to check (base 1)
	 * @return true if no data present; false otherwise
	 */
	public boolean nodata(int n1, int n2)
	{
		if (csv==null) return false; // should never happen
		for (int i=n1; i<=n2; i++)
		{
			String s = csv.getField(i-1); // csv is base 0
			if (s.length() == 0) continue; // empty
			for (int j=0; j<s.length(); j++)
			{ // check for non-white space
				Character c = s.charAt(j); // get character
				if (!Character.isWhitespace(c)) return false; // data found
			}
		}
		return true; // no data found
	}
	
	/**
	 * Apply the specified pattern to the specified string
	 * 
	 * @param str input string
	 * @param pattern the pattern code to be applied to the string
	 * @return the resulting output string with pattern codes
	 * 
	 */
	public String patt(String str, String pattern)
	{
		String ret = WrpPattern.applyPattern(str, pattern);
		return ret;
	}
	
	/**
	 * Normalize the input string using the specified pattern.
	 * Each character in the pattern string is compared with
	 * the next character from the input string.
	 * If the pattern character is a match character, 
	 * then the input character is either copied or transformed
	 * to the output string.
	 * Some match characters apply to a single character in
	 * the input string, while others match the entire string.
	 * If the pattern character is not a match character,
	 * then the pattern character is copied to the output
	 * string, and the input character is compared with
	 * the next pattern character.
	 * For example, the pattern "999-999-9999" would normalize
	 * the input string "1234567890" to "123-456-7890".
	 * Also, the pattern "u" will normalize the input string "usa"
	 * to "USA".
	 * </p><p>
	 * Supported match characters:
	 * </p>
	 * <p>
	 * Notice that the characters "X", "9", and "a" can't be inserted
	 * into the output string using the pattern.
	 * </p>
	 * 
	 * @param ndx index into rules for current rule
	 * @param strIn input string
	 * @param pattern normalization pattern
	 * @return the resulting string
	 * 
	 */
	public String norm(int ndx, String strIn, String pattern)
	{
		if (pattern.equals("*"))
		{ // copy pattern as it
			return strIn;
		}
		// apply pattern rule to input string
		StringBuffer sb =  new StringBuffer();
		int loc = 0; // pointer for input string
		for (int i=0; i<pattern.length(); i++)
		{ // compare pattern against input string
			if (loc >= strIn.length())
			{ // not enough characters in input string
				String msg = "Not enough characters in input string (" + strIn + ") for pattern (" + pattern + ")";
				logError(ndx, msg);
				break;
			}
			Character cin = strIn.charAt(loc);
			Character cpatt = pattern.charAt(i);
			if (cpatt == 'X')
			{ // matches any character from input string
				sb.append(cin);
			}
			else if (cpatt == 'a')
			{ // input character should be a letter
				if (Character.isLetter(cin))
				{
					sb.append(cin);
				}
				else
				{ // input character doesn't match
					String msg = "Input string character (" + cin + ") doesn't match pattern character (a)";
					logError(ndx, msg);
				}
			}
			else if (cpatt == '9')
			{ // input character should be a digit
				if (Character.isDigit(cin))
				{
					sb.append(cin);
				}
				else
				{ // input string doesn't match normalization pattern
					String msg = "Input string character (" + cin + ") doesn't match pattern character (9)";
					logError(ndx, msg);
				}
			}
			else
			{ // copy pattern character, and don't update input character pointer
				sb.append(cpatt); // append pattern
				loc--; // compensate for upcoming increment
			}
			loc++;
		}
		
		String ret = sb.toString(); // normalized string
		return ret;
	}
	
	/**
	 * Raise exception if a null argument is passed for a required argument.
	 * This function should never be called if the rules are correct.
	 * This method reports that a null argument was passed to a command,
	 * but the command required to argument to function properly.
	 * The method throws a runtime error with the appropriate error message.
	 * 
	 * @param ndx the index into Rules of the command being executed.
	 * @throws ValidationRuntimeException
	 */
	protected void nullArg(int ndx) throws ValidationRuntimeException
	{
		ValidationRule rule = rules.get(ndx);
		String s = "Null argument passed to command";
		String info = "Rule " + ndx + "=[" + rule.toString() + "]";
		ValidationRuntimeException exp = new ValidationRuntimeException(info, s);
		throw exp;
	}
	
	/**
	 * Validate the city/state/zip combination.
	 * This method determins if the zip code is valid for the
	 * city and state combination.
	 * This method uses the USPS database.
	 * 
	 * @param str input string: "<i>city</i>|<i>state</i>|<i>zip</i>"
	 * @param outvar error message, if any
	 * @return true if valid; false otherwise (with error message in <i>outvar</i>)
	 */
	protected boolean validateZipCode(String str, Variable outvar)
	{
		boolean ret = false;
		String[] flds = str.split("\\|"); // split on vertical bars
		if (flds.length!=3)
		{ // wrong number of fields
			outvar.strVal = "Invalid ZipCode string - use 'city|state|zip'";
			return false;
		}
		String strCity = flds[0];
		String strState = flds[1];
		String strZip = flds[2];
		if (strZip.length() > 5)
		{
			strZip = flds[2].substring(0, 5); // 5 digit zip code
		}
		ret = !zipCodes.isZipCodeValid(strZip, strState, strCity);
		outvar.strVal = "Invalid zip code combination [" + str + "]";
		return ret;
	}
	
	/**
	 * Add error message to list of messages
	 * 
	 * @param msg an error message explaining the problem
	 * @param ndx location of invalid rule
	 */
	public void logError(int ndx, String msg)
	{
		boolean compile;
		if (runErrors < 0)
		{ // compile stage
			compile = true;
			compileErrors++; // compilation has errors
		}
		else
		{ // execution mode - update runtime errors
			compile = false;
			runErrors++;
		}
		
		/**
		 * Get current sequence number and rule.
		 * If a compile error, display the rule.
		 * If a run-time error, display the CSV string
		 */
		ValidationRule rule = rules.get(ndx);
		int seq = rule.getSeq();
		if (seq==breakpt)
		{ // debugging break point
			System.out.println("logError: Seq=" + seq);
		}
		if (compile || csvLine != prevErr)
		{ // first error for this line
			errMsgs.add(""); // insert blank line separator
			String s;
			if (compile)
			{
				s = "Rule " + ndx + " - " + rule.toString();
				prevErr = seq;
			}
			else
			{
				s = "Line " + csvLine + ": " + csv.getLine();
				prevErr = csvLine; // don't display again
			}
			errMsgs.add(s);
		}
		// add message for this error
		errMsgs.add("        " + msg);
	}
	
	/**
	 * Validate the given CSV record using the given rules
	 * 
	 * @param rules the validation rules to use
	 * @param csvLine the line number to use for this validation
	 * @param csv the CSV record to validate
	 * 
	 * @return the number of run-time validation errors
	 */
	public int validate(Vector<ValidationRule> rules, int csvLine, WrpCsv csv)
	{
		this.csvLine = csvLine;
		setRules(rules);
		setCsv(csv);
		return validate();
	}
	
	/**
	 * Validate the current record using the specified line number.
	 * This method validates the current record using existing rules,
	 * but first setting the line number to the specified value.
	 * 
	 * @param csvLine the line number to use for this validation
	 * @return the number of run-time errors
	 * 
	 */
	public int validate(int csvLine)
	{
		this.csvLine = csvLine;
		int ret = validate();
		return ret;
	}
	
	/**
	 * Validate the given CSV record using the specified rules
	 * 
	 * @return the number of compilation or run-time validation errors
	 * 
	 */
	public int validate()
	{
		int ret = 0;

		if (labels==null) 
		{
			ret = compileRules(); // sets compileErrors
			if (ret>0) return -ret; // compile errors
		}

		// rules compiled correctly - execute them to validate records
		prevErr = -1; // reset line number of previous error
		runErrors = 0; // reset error count
		errMsgs = new Vector<String>();
		normString = "";

		try {
			if (csv!=null)
			{
				if (csvLine==1 && headerLine!=null)
				{ // validate header line
					ret = validateHeader();
				}
				else
				{
					ret = executeRules(); // sets runErrors
				}
			}
		} catch (ValidationRuntimeException e) {
			String info = e.getDebugInfo(); // debug message
			String msg = e.getMessage();
			errMsgs.add("Internal error (" + msg + ") info=[" + info + "]");
			ret = 1; // invalid record
			Stat = REC_ERROR; // mark list bad
		}
		return ret;
	}
	
	/**
	 * Validate the specified CSV using the previously supplied rules
	 * 
	 * @param csv CSV line to be validated
	 * @return the number of run-time errors
	 * 
	 */
	public int validate(WrpCsv csv)
	{
		setCsv(csv);
		int ret = validate();
		
		return ret;
	}
	
	/**
	 * Validate the header line.
	 * This method compares the current line with the header string.
	 * It should be called by validate() when the first line is read,
	 * if the headerLine string has been set.
	 * 
	 * @return Record status (0 if good - see constants REC_*)
	 */
	public int validateHeader()
	{
		int ret = REC_ERROR;
		if (headerCsv==null)
		{ // first time - parse header line
			headerCsv = new WrpCsv(headerLine);
		}
		int nflds = csv.getNumFields();
		int hflds = headerCsv.getNumFields();
		String sfld = null;
		String hfld = null;
		if (nflds==hflds)
		{ // number of fields match - compare them
			ret = REC_GOOD; // assume they match
			for (int i=0; i<nflds; i++)
			{ // compare each field
				sfld = csv.getField(i);
				hfld = headerCsv.getField(i);
				if (sfld==null || hfld==null)
				{ // should never happen
					ret = REC_BAD;
				}
				else
				{ // compare fields
					if (!sfld.equalsIgnoreCase(hfld))
					{
						logError(1, "Header field mismatch.  Expected ["
								+ hfld + "], found ["
								+ sfld + "]"
								);
						ret = REC_BAD;
					}
				}
			}
		}
		else
		{ // number of fields don't match
			logError(1, "Wrong number of fields in the header.  Expected [" + hflds 
					+ "], found [" + nflds + "]"
			);
			ret = REC_BAD;
		}

		return ret;
	}
	
	/**
	 * Inner class for validation rule
	 * </p><p>
	 * This class instantiates a rule used to validate a CSV record.
	 * A rule can "fire", which typically indicates a sucessful test
	 * or an exception to the operation.
	 * </p><p>
	 * A validation rule consists of various arguments.
	 * Which arguments are required depends on the command itself.
	 * The rule can have a statement label as well as a GoTo destination.
	 * (Execution continues at the GoTo label if the command fires.)
	 * It can also have a message to be displayed if the command fires,
	 * as well as a record status if the command fires.
	 * Two variables can be specified - an input and an output variable 
	 * (<i>invar</i> and <i>outvar</i>) (which can be the same variable), 
	 * and two literals (<i>arg1</i> and <i>arg2</i>), which are constants.
	 * </p><p>
	 * The data type of a literal depends on its name:
	 * </p>
	 * <ul>
	 * <li>If the name starts with a letter, it is type String</li>
	 * <li>If the name starts with a double quote, 
	 * then open and closing quotes are removed and the type is String</li>
	 * <li>If the name starts with a digit, it is considered numeric,
	 * unless the string contains a non-numeric character.</li>
	 * <li>If a numeric string contains a decimal point, it is type Float</li>
	 * <li>If a numeric string has no decimal point, it is type Integer</li>
	 * </ul>
	 * <p>
	 * To reduce overhead during run time, the index into <i>variables</i>
	 * or <i>literals</i> for each argument is stored for each referenced
	 * variable or literal.
	 * 
	 * @author Bill Pringle
	 *
	 */
	public class ValidationRule
	{
		/** unique row ID (from the database) */
		public int Id=0;
		/** sequence number for rule */
		public int Seq=0;
		/** destination label */
		public String Label="";
		/** input variable */
		public String InVar="";
		/** compiled location of input variable */
		public int InVarLoc=-1;
		/** command for rule */
		public String Cmd="";
		/** argument 1 */
		public String Arg1="";
		/** location of literal in arg1 */
		public int Arg1Loc=-1;
		/** argument 2 */
		public String Arg2="";
		/** location of literal in arg2 */
		public int Arg2Loc=-1;
		/** output variable - where results are stored */
		public String OutVar="";
		/** compiled location of output variable */
		public int OutVarLoc=-1;
		/** validation results for entry */
		public String Result="";
		/** output message for user */
		public String Msg="";
		/** GoTo destination */
		public String GoTo="";
		/** compiled index into rules for destination */
		public int GoToLoc=-1;
		/** comment */
		public String Comment="";
		
		
		/**
		 * default constructor
		 * 
		 */
		public ValidationRule()
		{
			
		}

		/**
		 * Provide a string representation of the validation rule.
		 * The command (cmd) parameter is always displayed;
		 * other missing fields are not displayed.
		 * 
		 */
		public String toString()
		{
			StringBuffer sb = new StringBuffer("WrpValidationRule: ");
			int n = getSeq();
			sb.append("Seq=" + n);
			String s = getLabel();
			if (s.length()>0) sb.append(", Lbl=[" + s + "]");
			s = getInVar();
			if (s.length()>0) sb.append(", InVar=[" + s + "]");
			s = getCmd();
			sb.append(", Cmd=[" + s + "]");
			s = getArg1();
			if (s.length()>0) sb.append(", Arg1=[" + s + "]");
			s = getArg2();
			if (s.length()>0) sb.append(", Arg2=[" + s + "]");
			s = getOutVar();
			if (s.length()>0) sb.append(", OutVar=[" + s + "]");
			s = getResult();
			if (s.length()>0) sb.append(", Result=[" + s + "]");
			s = getMsg();
			if (s.length()>0) sb.append(", Msg=[" + s + "]");
			s = getGoTo();
			if (s.length()>0) sb.append(", GoTo=[" + s + "]");
			s = getComment();
			if (s.length()>0) sb.append(", Comment=[" + s + "]");
			
			String ret = sb.toString();
			return ret;
		}

		/**
		 * Get the sequence number for the current rule
		 * 
		 * @return the sequence number
		 */
		public int getSeq() {
			return Seq;
		}


		/**
		 * Set the sequence number for this rule
		 * 
		 * @param seq the sequence number to set
		 */
		public void setSeq(int seq) {
			Seq = seq;
		}


		/**
		 * Get the statement label for this rule
		 * 
		 * @return the statement label
		 */
		public String getLabel() {
			return Label;
		}


		/**
		 * Set the statement label for this rule
		 * 
		 * @param label the statement label to set
		 */
		public void setLabel(String label) {
			Label = label;
		}


		/**
		 * Get the input variable for this rule
		 * 
		 * @return the input variable
		 */
		public String getInVar() {
			return InVar;
		}


		/**
		 * Set the input variable for this rule
		 * 
		 * @param inVar the input variable to set
		 */
		public void setInVar(String inVar) {
			InVar = inVar;
		}


		/**
		 * Return the command name for this rule
		 * 
		 * @return the command
		 */
		public String getCmd() {
			return Cmd;
		}


		/**
		 * set the command name for this rule
		 * 
		 * @param cmd the command to set
		 */
		public void setCmd(String cmd) {
			Cmd = cmd;
		}


		/**
		 * Get the first literal value
		 * 
		 * @return the first argument
		 */
		public String getArg1() {
			return Arg1;
		}


		/**
		 * Set the first literal value
		 * 
		 * @param arg1 the first argument to set
		 */
		public void setArg1(String arg1) {
			Arg1 = arg1;
		}

		/**
		 * Save location of literal for arg1 of this command
		 * 
		 * @param loc the index into literals for arg1 of this command
		 */
		public void setArg1Loc(int loc)
		{
			Arg1Loc = loc;
		}
		
		/**
		 * Obtain the location of the literal for arg1
		 * 
		 * @return the index into literals for the literal in arg1
		 * 
		 */
		public int getArg1Loc()
		{
			return Arg1Loc;
		}

		/**
		 * Get the second literal value for this rule
		 * 
		 * @return the second argument
		 */
		public String getArg2() {
			return Arg2;
		}

		/**
		 * Set the second literal value for this rule
		 * 
		 * @param arg2 the second argument to set
		 */
		public void setArg2(String arg2) {
			Arg2 = arg2;
		}

		/**
		 * Save location of literal for arg2 of this command
		 * 
		 * @param loc the index into literals for arg2 of this command
		 */
		public void setArg2Loc(int loc)
		{
			Arg2Loc = loc;
		}

		/**
		 * Obtain the location of the literal for arg2
		 * 
		 * @return the index into literals for arg2
		 * 
		 */
		public int getArg2Loc()
		{
			return Arg2Loc;
		}

		/**
		 * Get the output variable for this rule
		 * 
		 * @return the output variable
		 */
		public String getOutVar() {
			return OutVar;
		}


		/**
		 * Set the second literal value for this rule
		 * 
		 * @param outVar the output variable to set
		 */
		public void setOutVar(String outVar) {
			OutVar = outVar;
		}


		/**
		 * Get the optional result value for this rule
		 * 
		 * @return the validation result
		 */
		public String getResult() {
			return Result;
		}


		/**
		 * Set the result string for this rule
		 * 
		 * @param result the validation result to set
		 */
		public void setResult(String result) {
			Result = result;
		}


		/**
		 * Get the user message for this rule
		 * 
		 * @return the user message
		 */
		public String getMsg() {
			return Msg;
		}


		/**
		 * Set the user message for this rule
		 * 
		 * @param msg the user message to set
		 */
		public void setMsg(String msg) {
			Msg = msg;
		}


		/**
		 * Get the "GoTo" destination for this rule
		 * 
		 * @return the destination label
		 */
		public String getGoTo() {
			return GoTo;
		}


		/**
		 * Set the "GoTo" destination for this rule
		 * 
		 * @param goTo the destination label to set
		 */
		public void setGoTo(String goTo) {
			GoTo = goTo;
		}


		/**
		 * Get the comment for this rule
		 * 
		 * @return the comment
		 */
		public String getComment() {
			return Comment;
		}


		/**
		 * Set the comment for this rule
		 * 
		 * @param comment the comment to set
		 */
		public void setComment(String comment) {
			Comment = comment;
		}


		/**
		 * Get the unique database ID for this rule
		 * 
		 * @return the id
		 */
		public int getId() {
			return Id;
		}


		/**
		 * Set the unique database ID for this rule
		 * 
		 * @param id the id to set
		 */
		public void setId(int id) {
			Id = id;
		}


		/**
		 * Get the location in <i>variables</i> for the input variable for this rule
		 * 
		 * @return the inVarLoc
		 */
		public int getInVarLoc() {
			return InVarLoc;
		}


		/**
		 * Set the location in <i>variables</i> of the input variable for this rule
		 * 
		 * @param inVarLoc the inVarLoc to set
		 */
		public void setInVarLoc(int inVarLoc) {
			InVarLoc = inVarLoc;
		}


		/**
		 * Get the location in <i>variables</i> of the output variable for this rule
		 * 
		 * @return the outVarLoc
		 */
		public int getOutVarLoc() {
			return OutVarLoc;
		}


		/**
		 * Set the location in <i>variables</i> of the outvariable for this rule
		 * 
		 * @param outVarLoc the outVarLoc to set
		 */
		public void setOutVarLoc(int outVarLoc) {
			OutVarLoc = outVarLoc;
		}


		/**
		 * Get the location in <i>rules</i> for the GoTo destination of this rule
		 * 
		 * @return the goToLoc
		 */
		public int getGoToLoc() {
			return GoToLoc;
		}


		/**
		 * Set the location in<i>rules</i> for the GoTo destination label of this rule
		 * 
		 * @param goToLoc the goToLoc to set
		 */
		public void setGoToLoc(int goToLoc) {
			GoToLoc = goToLoc;
		}
		
	}
	
	
	/**
	 * Inner class for variable.
	 * </p><p>
	 * This class represents variables during the execution
	 * of the validation rules.
	 * All variables must be declared, although the declarations
	 * can appear after it is referenced.
	 * A variable exists throughout the execution stage.
	 * The <i>map</i> command can be used to define a relationship
	 * between a CSV field and a variable.
	 * Each time a new record is processed, the CSV fields are
	 * copied to the mapped variables.
	 * Non-mapped variables retain their values until they are changed.
	 * Because of this, variables can retain their values across
	 * different CSV records, which means you can accumulate totals
	 * for the entire execution phase. 
	 * </p><p>
	 * To increase efficiency, the index into variables
	 * for a referenced variable is stored in the rule structure
	 * so that a search for the variable is not required during run time.
	 * 
	 * @author Bill Pringle
	 *
	 */
	public class Variable
	{
		/** variable name */
		public String name;
		/** variable type (TYPE_something) */
		public int type;
		/** flag indicating variable is mapped to a field */
		public Boolean fld;
		/** value of variable, if string */
		public String strVal;
		/** value of variable, if integer */
		public int intVal;
		/** value of variable, if float */
		public float fltVal;
		/** value of variable, if boolean */
		public boolean boolVal;
		
		/**
		 * Default constructor
		 * 
		 */
		public Variable()
		{
			name = null;
			type = TYPE_ERR;
			fld = false;
		}
		
		/**
		 * Set the value based on the type of variable
		 * 
		 * @param val the new value of the variable
		 * @return true if the assignment succeeds; false otherwise
		 * @throws NumberFormatException if the value doesn't match the data type
		 * 
		 */
		public boolean setVal(String val)
			throws NumberFormatException
		{
			boolean ret = true; // set valid
			switch (type)
			{
			case TYPE_INT:
				int n = Integer.parseInt(val);
				intVal = n; // integer value
				break;
			case TYPE_STR:
				strVal = val; // just copy value
				break;
			case TYPE_FLT:
				float f = Float.parseFloat(val);
				fltVal = f;
				break;
			case TYPE_BOOL:
				if (val.startsWith("t") || val.startsWith("y")) boolVal = true;
				else if (val.startsWith("f") || val.startsWith("n")) boolVal = false;
				else ret = false;
				break;
			default:
				ret = false;
			}
			return ret;
		}
		
		/**
		 * Set the value to that of the specified variable.
		 * Both variables should be the same data type.
		 * 
		 * @param v the Variable containing the new value
		 * 
		 * @return true if operation suceeded
		 * 
		 */
		public boolean set(Variable v)
		{
			int type = getType();
			int type2 = v.getType();
			if (type!=type2) return true; // error - wrong data types
			
			switch (type)
			{
			case TYPE_INT: intVal = v.intVal; break;
			case TYPE_STR: strVal = v.strVal; break;
			case TYPE_FLT: fltVal = v.fltVal; break;
			case TYPE_BOOL: boolVal = v.boolVal; break;
			default: return false;
			}
			return true;
		}

		/**
		 * Get the current value of the variable.
		 * </p><p>
		 * The data type that is returned depends on
		 * the type of the variable.
		 * 
		 * @return an object with the current value of the variable
		 * 
		 */
		public Object getVal()
		{
			if (type==TYPE_STR)
			{
				return strVal;
			}
			else if (type==TYPE_INT)
			{
				return intVal;
			}
			else if (type==TYPE_FLT)
			{
				return fltVal;
			}
			else if (type==TYPE_BOOL)
			{
				return boolVal;
			}
			else return "*Error*";
		}
		
		/**
		 * Get the name of this variable
		 * 
		 * @return the name
		 */
		public String getName() {
			return name;
		}
		/**
		 * Set the name of this variable
		 * 
		 * @param name the name to set
		 */
		public void setName(String name) {
			this.name = name;
		}
		/**
		 * Get the data type of this variable
		 * 
		 * @return the type
		 */
		public int getType() {
			return type;
		}
		/**
		 * Set the data type of this variable
		 * 
		 * @param type the type to set
		 */
		public void setType(int type) {
			this.type = type;
		}
		
		/**
		 * Set the data type of this variable
		 * 
		 * @param ftype the name of the data type to be set
		 */
		public void setType(String ftype)
		{
			int type = varType(ftype);
			this.type = type;
		}
		
		/**
		 * Get the string value of this variable
		 * 
		 * @return the strVal
		 */
		public String getStrVal() {
			return strVal;
		}
		/**
		 * Set the string value of this variable
		 * 
		 * @param strVal the strVal to set
		 */
		public void setStrVal(String strVal) {
			this.strVal = strVal;
		}
		/**
		 * Get the integer value of this variable
		 * 
		 * @return the intVal
		 */
		public int getIntVal() {
			return intVal;
		}
		/**
		 * Set the integer variable of this variable
		 * 
		 * @param intVal the intVal to set
		 */
		public void setIntVal(int intVal) {
			this.intVal = intVal;
		}
		/**
		 * Get the floating point value of this variable
		 * 
		 * @return the fltVal
		 */
		public float getFltVal() {
			return fltVal;
		}
		/**
		 * Set the floating point variable of this variable
		 * 
		 * @param fltVal the fltVal to set
		 */
		public void setFltVal(float fltVal) {
			this.fltVal = fltVal;
		}
		/**
		 * Get the boolean value for this variable
		 * 
		 * @return the boolVal
		 */
		public boolean isBoolVal() {
			return boolVal;
		}
		/**
		 * Set the boolean value of this variable
		 * 
		 * @param boolVal the boolVal to set
		 */
		public void setBoolVal(boolean boolVal) {
			this.boolVal = boolVal;
		}
		
		/**
		 * Determines if the current variable is mapped to a CSV field.
		 * 
		 * @return true if the variable is mapped; otherwise false
		 * 
		 */
		public boolean isField()
		{
			return fld;
		}
		
		/**
		 * Specify if the variable is mapped to a CSV field.
		 * 
		 * @param b true if the variable is mapped; false otherwise
		 * 
		 */
		public void setField(Boolean b)
		{
			fld = b;
		}
		
		/**
		 * Build string representation of Variable,
		 * including the values and locations
		 * 
		 */
		public String toString()
		{
			StringBuffer sb = new StringBuffer("Variable: ");
			sb.append(" Name=[" + name + "]");
			String s = varType(type); // get text version
			sb.append(", Type=" + type + " [" + s + "]");
			switch (type)
			{
			case TYPE_INT: sb.append(", Val=[" + intVal + "]"); break;
			case TYPE_STR: sb.append(", Val=[" + strVal + "]"); break;
			case TYPE_FLT: sb.append(", Val=[" + fltVal + "]"); break;
			case TYPE_BOOL: sb.append(", Val=[" + boolVal + "]"); break;
			case TYPE_ERR: sb.append(", Error type"); break;
			default: sb.append(", Unknown data type"); break;
			}
			
			if (isField()) sb.append(", Variable is mapped to a Field");
			String ret = sb.toString();
			return ret;
		}
	}
	
	/**
	 * Inner class for statement label
	 * </p><p>
	 * A statement can have a statement label,
	 * which can then be used as a destination if a rule fires.
	 * A statement label can also be used for documentation,
	 * and do not require that a rule specifies the statement
	 * as a destination.
	 * </p><p>
	 * For efficiency sake, the index into the labels vector
	 * is stored so that the program doesn't have to search
	 * for the statement during run time.
	 * 
	 * @author Bill Pringle
	 *
	 */
	public class GoToLoc
	{
		/** the statement label */
		String name;
		/** the statement sequence number */
		int seq;
		/** index into rules for statement */
		int loc;
		
		/**
		 * Get the name of this statement label
		 * 
		 * @return the name
		 */
		public String getName() {
			return name;
		}
		/**
		 * Set the name of this statement label.
		 * This name should be unique.
		 * 
		 * @param name the name to set
		 */
		public void setName(String name) {
			this.name = name;
		}
		/**
		 * Return the sequence number of the rule associated
		 * with this statement label.
		 * 
		 * @return the seq
		 */
		public int getSeq() {
			return seq;
		}
		/**
		 * Set the statement number 
		 * for the rule associated with this statement label
		 * 
		 * @param seq the seq to set
		 */
		public void setSeq(int seq) {
			this.seq = seq;
		}
		/**
		 * Get the location in <i>rules</i> for the rule
		 * associated with this statement label
		 * 
		 * @return the location of the rule
		 */
		public int getLoc() {
			return loc;
		}
		/**
		 * Set the location within <i>rules</i> of the rule
		 * associated with this statement label.
		 * 
		 * @param loc the loc to set
		 */
		public void setLoc(int loc) {
			this.loc = loc;
		}
		
		/**
		 * Build string representation for this label
		 * 
		 */
		public String toString()
		{
			StringBuffer sb = new StringBuffer("GoTo Label: ");
			sb.append("Name=[" + name + "]");
			sb.append(", Seq=[" + seq + "]");
			sb.append(", Loc=[" + loc + "]");
			
			String ret = sb.toString();
			return ret;
		}
	}
	
	/**
	 * Inner class for validation run time exception
	 * </p><p>
	 * This class can be used to throw and handle
	 * exceptions during run time.
	 * 
	 * @author Bill Pringle
	 *
	 */
	public class ValidationRuntimeException extends RuntimeException
	{
		final static long serialVersionUID = 1;

		/** 
		 * additional Exception field used to store debugging information
		 * (e.g., values of variables, and indicating of where exception was thrown, etc.)
		 * 
		 */
		String DebugInfo;
		
		/**
		 * Default constructor with no debugging information
		 * 
		 */
		public ValidationRuntimeException()
		{
			super("Validation Runtime Exception");
			DebugInfo = null;
		}
		
		/**
		 * constructor with debugging information specified
		 * 
		 * @param DebugInfo string containing debugging information
		 */
		public ValidationRuntimeException(String DebugInfo)
		{
			super("Validation Runtime Exception");
			this.DebugInfo = DebugInfo;
		}
		
		/**
		 * Constructor with message and debugging information
		 * 
		 * @param DebugInfo string containing debugging information
		 * @param Message exception message
		 * 
		 */
		public ValidationRuntimeException(String DebugInfo, String Message)
		{
			super(Message);
			this.DebugInfo = DebugInfo;
		}

		/**
		 * Get the debugging information for this exception
		 * 
		 * @return the debugInfo
		 */
		public String getDebugInfo() {
			return DebugInfo;
		}

		/**
		 * Set the debugging information for this exception
		 * 
		 * @param debugInfo the debugInfo to set
		 */
		public void setDebugInfo(String debugInfo) {
			DebugInfo = debugInfo;
		}
		
	}
	
	//////////////////////////////////////////////////
	// public static routines
	///////////////////////////////////////////////////
	
	/**
	 * Convert type number to name
	 * 
	 * @param num type number (TYPE_something)
	 * @return a string representation of the type
	 * 
	 */
	public static String varType(int num)
	{
		switch (num)
		{
		case TYPE_STR: return "String";
		case TYPE_INT: return "Integer";
		case TYPE_FLT: return "Float";
		case TYPE_BOOL: return "Boolean";
		case TYPE_ERR: return "Type Error";
		case MAP_REC: return "Entire Record";
		case MAP_NFLDS: return "Number of Fields";
		case MAP_ERR: return "Map Error";
		default: return UNKNOWN;
		}
	}
	
	/**
	 * Convert type name to integer value
	 * 
	 * @param name the name of the type
	 * @return the integer value for the type (TYPE_something)
	 * 
	 */
	public static int varType(String name)
	{
		String s = name.toLowerCase();
		if (s.startsWith("i")) return TYPE_INT;
		if (s.startsWith("s")) return TYPE_STR;
		if (s.startsWith("f")) return TYPE_FLT;
		if (s.startsWith("b")) return TYPE_BOOL;
		return TYPE_ERR;
	}
	
	/**
	 * Convert record status from string to integer
	 * 
	 * @param name the record status name
	 * @return an integer representation for the record status
	 * 
	 */
	public static int recStat(String name)
	{
		if (name==null) return REC_ERROR;
		String s = name.toUpperCase();
		if (s.equals("GOOD")) return REC_GOOD;
		else if (s.equals("BAD")) return REC_BAD;
		else if (s.equals("WARN")) return REC_WARN;
		else if (s.equals("FUZZY")) return REC_FUZZY;
		else if (s.equals("MISSING")) return REC_NONE;
		else return REC_ERROR; // invalid status
	}
	
	/**
	 * convert record status number to a string
	 * 
	 * @param num a numerical representation of the record status
	 * @return the string representation for the record status
	 * 
	 */
	public static String recStat(int num)
	{
		switch (num)
		{
		case REC_GOOD: return "GOOD";
		case REC_WARN: return "WARN";
		case REC_FUZZY: return "FUZZY";
		case REC_BAD: return "BAD";
		case REC_NONE: return "MISSING";
		default: return UNKNOWN;
		}
	}
	
	/**
	 * Example driver for this class.
	 * It is assumed that the user will create their own main program
	 * that will use their own database rather than the example UtilDb. 
	 * 
	 * @param args not used
	 */
	public static void main(String[] args) 
	{
		ResultSet rs=null;
		try {
			UtilDb db = new UtilDb();
			String query = "Call GetWrpValidationRules";
			rs = db.executeQuery(query);
		} catch (ClassNotFoundException e) {
			System.err.println("Unable to find MySQL driver class");
			e.printStackTrace();
		} catch (IllegalAccessException e) {
			System.err.println("Unauthorized to create new MySQL Driver class");
			e.printStackTrace();
		} catch (InstantiationException e) {
			System.err.println("Unable to instantiate new MySQL Driver class");
			e.printStackTrace();
		} catch (SQLException e) {
			System.err.println("SQL Exception while opening connection");
			e.printStackTrace();
		}
		
		if (rs==null)
		{
			System.err.println("No ResultSet returned from query!");
			System.exit(1);
		}
		WrpValidateCsv val = new WrpValidateCsv();
		try {
			val.loadRules(rs);
		} catch (SQLException e) {
			System.err.println("SQL Exception while loading rules");
			e.printStackTrace();
		}
		
		// rules loaded - compile them		
		int numErrors = val.compileRules();
		if (numErrors==0)
		{
			System.out.println("No compile errors found");
		}
		else
		{
			Vector<String> msgs = val.getErrMsgs();
			System.err.println(numErrors + " compile errors found");
			for (int i=0; i<msgs.size(); i++)
			{
				String s = msgs.get(i);
				System.err.println(s);
			}
		}
		
		System.exit(0);
	}

}
