package com.billpringle.utils.wrputils;

import java.io.BufferedReader;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintStream;
import java.util.Vector;

/**
 * This class identifies what words appear in a string and how many.
 * 
 * A key string is used to identify the specific entity that is beingn counted.
 * For example, you can count words in several text documents,
 * and report on how many times a given word appears in both documents.
 * 
 *   <table border="0" cellspacing="0" cellpadding="6" align="center"
  summary="Creative Commons license">
    <tr>
      <td><!-- Creative Commons License -->
      <a rel="license" href=
      "http://creativecommons.org/licenses/by-nc-sa/2.0/" target=
      "CC"><img alt="Creative Commons License" border="0" src=
      "http://creativecommons.org/images/public/somerights20.gif" /></a></td>

      <td><a rel="license" href=
      "http://creativecommons.org/licenses/by-nc-sa/2.0/" target=
      "CC"><img alt="Creative Commons License Symbols" border="0"
      src="cc_lic.gif" /></a> <!-- /Creative Commons License --></td>

      <td>Unless noted otherwise, all materials available for
      download from my site are copyrighted by Bill Pringle, and
      are licensed under a <a rel="license" href=
      "http://creativecommons.org/licenses/by-nc-sa/2.0/" target=
      "CC">Creative Commons License</a>.</td>
    </tr>
  </table>
 *
 * @author Bill Pringle
 */

public class WrpWordCount {
	/** unique key */
	protected String key;
	/** reader for file */
	protected BufferedReader rdr;
	/** collection of words */
	protected Vector<WordCount> words = new Vector<WordCount>();
	
	/**
	 * Default constructor
	 * 
	 */
	public WrpWordCount()
	{
		
	}
	
	/**
	 * Constructor with key string
	 * 
	 * @param key the key string
	 */
	public WrpWordCount(String key)
	{
		this.key = key; // save key
	}
	
	/**
	 * Parse the specified file and count words
	 * 
	 * @param fname the name of the file to be read
	 */
	public void parseFile(String fname)
	{
		try {
			rdr = new BufferedReader(new FileReader(new File(fname)));
			parseFile();
		} catch (FileNotFoundException e) {
			System.err.println("File not found [" + e.getMessage() + "]");
			e.printStackTrace(System.err);
		}
	}
	
	/**
	 * Parse a file using the specified reader
	 * 
	 * @param rdr the BufferedReader to be used
	 */
	public void parseFile(BufferedReader rdr)
	{
		this.rdr = rdr;
		parseFile();
	}
	
	/**
	 * Parse the file using the predefined BufferedReader
	 * 
	 */
	public void parseFile()
	{
		if (rdr==null)
		{ // no reader defined
			System.err.println("WrpWordCount: No reader specified");
			return;
		}
		/*
		 * Parse the file using spaces and/or tabs
		 * to separate words
		 * 
		 */
		WrpCsv csv = new WrpCsv();
		String delims = " \t";
		csv.setDelims(delims);
		/*
		 * Loop reading lines and counting words
		 */
		try {
			while (true)
			{
				String str = rdr.readLine();
				if (str==null) break; // EOF
				if (str.length()==0) continue; // ignore blank lines
				parseString(str);
			}
		} catch (IOException e) {
			System.err.println("WrpWordCount: I/O Error [" + e.getMessage() + "]");
			e.printStackTrace(System.err);
		}
	}
	
	/**
	 * Parse the specified strings into words
	 * 
	 * @param str the string to parse
	 */
	public void parseString(String str)
	{
		WrpCsv csv = new WrpCsv();
		String delims = " \t";
		csv.setDelims(delims);
		int nflds = csv.parseString(str); // parse line
		for (int i=0; i<nflds; i++)
		{
			String s = csv.getField(i).toLowerCase(); // get word
			String w = lettersOnly(s);
			// we ignore 1 and 2 letter words
			if (w.length()>0) updateCount(key, w);
		}

	}
	
	/**
	 * Extract only letters and digits from string,
	 * removing punctuation, etc.
	 * 
	 * @param str the string to extract from
	 * @return string with only letters or digits
	 */
	public String lettersOnly(String str)
	{
		StringBuffer sbuf = new StringBuffer();
		for (int i=0; i<str.length(); i++)
		{
			char c = str.charAt(i);
			if (Character.isLetterOrDigit(c)) sbuf.append(c);
		}
		String ret = sbuf.toString();
		return ret;
	}
	
	/**
	 * Update count of words.
	 * If word is present, increment count
	 * If word not present, set count to one
	 * 
	 * @param key key string
	 * @param word the word whose count is to be updated
	 * 
	 */
	public void updateCount(String key, String word)
	{
		for (int i=0; i<words.size(); i++)
		{
			WordCount wc = words.get(i);
			if (!key.equals(wc.key)) continue; // wrong key
			if (word.equals(wc.word))
			{ // word found - increment count
				wc.count++;
				return;
			}
		}
		// word not found - create new entry
		WordCount wc = new WordCount(); // entry to be added
		wc.id = 0; // no ID yet - will be defined by database
		wc.key = key; // initialize key
		wc.word = word; // initialize word
		wc.count = 1; // initialize count
		words.add(wc); // add to list
	}
	
	/**
	 * Get the count associated with the given key / word pair.
	 * 
	 * @param key key string
	 * @param word desired word
	 * @return the number of times the word was found
	 */
	public int getCount(String key, String word)
	{
		for (int i=0; i<words.size(); i++)
		{
			WordCount wc = words.get(i);
			if (!key.equals(wc.key)) continue; // not same key
			if (word.equals(wc.word))
			{ // found word - return the count
				return wc.count;
			}
		}
		// word not found
		return 0;
	}
	
	/**
	 * Get the number of words that have been found
	 * 
	 * @return the number of words
	 */
	public int getNumberWords()
	{
		return words.size();
	}
	
	/**
	 * Get the word count information associated with a specific word
	 * 
	 * @param cnt which words to extract
	 * @return word count information for the specified word
	 */
	public WordCount getWordCount(int cnt)
	{
		if (cnt < 0 || cnt >= words.size()) return null;
		WordCount wc = words.get(cnt);
		return wc;
	}
	
	/**
	 * Get the specified word
	 * 
	 * @param cnt the word number
	 * @return the actual word
	 */
	public String getWord(int cnt)
	{
		WordCount wc = getWordCount(cnt);
		if (wc==null) return null;
		String ret = wc.word;
		return ret;
	}
	
	/**
	 * Get the word count information for specified word
	 * 
	 * @param cnt the number of the word
	 * @return the word count information for that word
	 */
	public int getCount(int cnt)
	{
		WordCount wc = getWordCount(cnt);
		if (wc==null) return -1;
		int ret = wc.count;
		return ret;
	}
	
	/**
	 * Get the current key string
	 * 
	 * @return the key string
	 */
	public String getKey()
	{
		return key;
	}
	
	/**
	 * Get the key associated with the specified word
	 * 
	 * @param cnt the number of the word to extract
	 * @return the word count information
	 */
	public String getKey(int cnt)
	{
		WordCount wc = getWordCount(cnt);
		if (wc==null) return null; // no entry, no key
		String ret = wc.key;
		return ret;
	}
	
	/**
	 * Dump the word count information
	 * 
	 * This method is useful for debugging.
	 * 
	 */
	public void dump()
	{
		dump(System.out);
	}
	
	/**
	 * Dump the word count information using the specified stream
	 * 
	 * @param prt the PrintStream to use for the dump
	 */
	public void dump(PrintStream prt)
	{
		int nwords = words.size();
		
		for (int i=0; i<nwords; i++)
		{
			String k = getKey(i);
			String word = getWord(i);
			int cnt = getCount(i);
			prt.println("Key [" + k + "]"
					+ " Word [" + word + "] count=" + cnt);
		}

	}
	
	/**
	 * Internal class for word count information
	 * </p><p>
	 * This class contains the following information:
	 * </p>
	 * <dl>
	 * <dt>id</dt>
	 * <dd>Database key</dd>
	 * <dt>key</dt>
	 * <dd>The file code</dd>
	 * <dt>word</dt>
	 * <dd>The actual word</dd>
	 * <dt>count</dt>
	 * <dd>The number of times the word appears in file</dd>
	 * </dl>
	 * 
	 * @author Bill Pringle
	 *
	 */
	public class WordCount
	{
		public int id; // database key
		public String key; // file code
		public String word; // actual word
		public int count; // number of times in file
		
		/**
		 * Default constructor
		 * 
		 */
		public WordCount()
		{
			
		}
	}
	
	/**
	 * Test driver for the WrpWordCount class
	 * 
	 * @param args command line arguments
	 */
	public static void main(String[] args) {

		WrpKeyboard kb = new WrpKeyboard();

		while (true)
		{
			try {
				System.out.print("Enter key: ");
				String key = kb.readLine();
				WrpWordCount wc = new WrpWordCount(key);
				String str;
				while (true) {
					System.out.print("Enter string to index, or ENTER to quit: ");
					str = kb.readLine().trim();
					if (str.length()==0) break;
					wc.parseString(str);
				}
				int nwords = wc.getNumberWords();
				System.out.println(nwords + " unique words entered.");
				wc.dump(System.out);
			} catch (IOException e) {
				System.err.println("I/O Exception [" + e.getMessage() + "]");
				e.printStackTrace();
			}
		}

	}

}
