package com.billpringle.utils.wrputils;

import java.io.File;
import java.io.IOException;
import java.io.PrintStream;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Set;
import java.util.Vector;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.Attributes;
import org.xml.sax.ContentHandler;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;

/**
 * Class that creates a HashMap from a simple XML file
 * </p><p>
 * This class will parse an XML file with simple name / value pairs
 * and create a HashMap where the value(s) can be accessed
 * by the name and the value being a Vector of String values.
 * </p><p>
 * For example, consider the file sample.xml:
 * </p>
 * <pre>
 * &lt;foo&gt;
 * &lt;user&gt;John Smith&lt;/user&gt;
 * &lt;password&gt;foobar&lt/password&gt;
 * &lt;color&gt;red&lt;/color&gt;
 * &lt;color&gt;blue&lt;/color&gt;
 * &lt/foo&gt;
 * </pre>
 * <p>
 * This class will read the file, ignore the root
 * element (foo) and create HashMap entries for user,
 * password, and color.  The user and password entries
 * will return a Vector with a single value, while the color tag
 * will be a Vector with two values.
 * </p><p>
 * The easiest way to use this class is to specify the
 * name of the XML file to be parsed in the Constructor. 
 * Another way is to call the parseFile(fname) method.
 * </p><p>
 * An example session:
 * </p>
 * <pre>
 * // parse XML file
 * WrpXmlMap map = new WrpXmlMap("sample.xml"); 
 * // get and display the user value
 * Vector<String> vals = map.getValues("user");
 * System.out.println("User name: " + vals[0]);
 * // get and display the colors
 * vals = map.getValues("color");
 * System.out.print("Colors:");
 * for (int i=0; i&lt;vals.size(); i++)
 * { // print all color values
 * 	  System.out.println(vals.get(i));
 * }
 *  // parse another dictionary type file
 *  // re-using the same map
 *  map.parseFile("dictionary.xml");
 *  // we can now use getValues(key) to get the values
 *  // from the new XML file
 * </pre>
 * <p>
 * This will produce the following output:
 * </p>
 * <pre>
 * User Name: John Smith
 * Colors:
 *     red
 *     blue
 * </pre>
 * 
 *   <table border="0" cellspacing="0" cellpadding="6" align="center"
  summary="Creative Commons license">
    <tr>
      <td><!-- Creative Commons License -->
      <a rel="license" href=
      "http://creativecommons.org/licenses/by-nc-sa/2.0/" target=
      "CC"><img alt="Creative Commons License" border="0" src=
      "http://creativecommons.org/images/public/somerights20.gif" /></a></td>

      <td><a rel="license" href=
      "http://creativecommons.org/licenses/by-nc-sa/2.0/" target=
      "CC"><img alt="Creative Commons License Symbols" border="0"
      src="cc_lic.gif" /></a> <!-- /Creative Commons License --></td>

      <td>Unless noted otherwise, all materials available for
      download from my site are copyrighted by Bill Pringle, and
      are licensed under a <a rel="license" href=
      "http://creativecommons.org/licenses/by-nc-sa/2.0/" target=
      "CC">Creative Commons License</a>.</td>
    </tr>
  </table>

 * 
 * @author Bill Pringle
 *
 */
public class WrpXmlMap {

	/** used by XML parser to save current tag name */
	private String thisKey;
	/** flag to ignore the root element */
	boolean firstElement = true;
	
	/**
	 * XML Event Handler for WrpXmlMap class
	 * 
	 * This class responds as the various XML elements
	 * are being read.
	 * The very first element (the root element) is ignored,
	 * but the remainder of the file should be of the form:
	 * <br />
	 * <pre>
	 * &lt;name&gt;value&lt;/name&gt;
	 * </pre>
	 * Where <i>name</i> is the name (or key) and
	 * <i>value</i> is the value associated with the key.
	 * </p><p>
	 * When the start tag is encountered, the name is saved.
	 * When the value string is read, that value is added to
	 * the vector associated with the saved key string.
	 * </p><p>
	 * Each key will have a Vector of String values.
	 * There can be multiple entries with the same key.
	 * 
	 * 
	 * @author Bill Pringle
	 *
	 */
	public class XmlMapContentHandler implements ContentHandler {

		/* (non-Javadoc)
		 * @see org.xml.sax.ContentHandler#characters(char[], int, int)
		 * 
		 * This method is called when the value string is read.
		 * The value string is added to the Vector associated
		 * with the current key tag.
		 */
		@Override
		public void characters(char[] ch, int start, int length)
				throws SAXException {
			/*
			 * The first element encountered is the root element.
			 * The root element is ignored.  All subsequent tags
			 * are used as keys for the HashMap and the value strings
			 * associated with the key tag are stored in the Vector
			 * for that key.
			 * 
			 */
			if (firstElement)
			{ // first time - ignore root element
				firstElement = false;
				return;
			}
			// ignore new lines and white space
			String s = new String(ch, start, length).trim();
			if (thisKey.length()==0) return; // can't have an empty key
			Vector<String> vals = map.get(thisKey); // try for values
			// if vals is null, then this is the first time
			if (vals==null) vals = new Vector<String>(); // create new Vector
			vals.add(s); // add this value to the Vector
			map.put(thisKey, vals); // store the Vector in the map
			thisKey = ""; // prevent duplication from EOL
		}

		@Override
		public void endDocument() throws SAXException {
			// TODO Auto-generated method stub

		}

		@Override
		public void endElement(String url, String localName, String qualName)
				throws SAXException {
			// System.out.println("End [" + qualName + "]");

		}

		@Override
		public void endPrefixMapping(String arg0) throws SAXException {
			// TODO Auto-generated method stub

		}

		@Override
		public void ignorableWhitespace(char[] arg0, int arg1, int arg2)
				throws SAXException {
			// TODO Auto-generated method stub

		}

		@Override
		public void processingInstruction(String arg0, String arg1)
				throws SAXException {
			// TODO Auto-generated method stub

		}

		@Override
		public void setDocumentLocator(Locator arg0) {
			// TODO Auto-generated method stub

		}

		@Override
		public void skippedEntity(String arg0) throws SAXException {
			// TODO Auto-generated method stub

		}

		/* (non-Javadoc)
		 * @see org.xml.sax.ContentHandler#startDocument()
		 * 
		 */
		@Override
		public void startDocument() throws SAXException {

			// create new map for this file
			map = new HashMap<String, Vector<String>>();

		}

		/* (non-Javadoc)
		 * @see org.xml.sax.ContentHandler#startElement(java.lang.String, java.lang.String, java.lang.String, org.xml.sax.Attributes)
		 * 
		 * The first element encountered is the root element.
		 * The root element is ignored.  
		 * When the start tag is encountered, the tag name is saved.
		 * This information is used when the character (value) string is read.
		 * 
		 * All tags except the root element
		 * are used as keys for the HashMap and the value strings
		 * associated with the key tag are stored in the Vector
		 * for that key.
		 */
		@Override
		public void startElement(String uri, String localName, 
				String qualName, Attributes attribs) 
		throws SAXException {
			// System.out.println("Start [" + qualName + "]");
			/*
			 * The localName argument is the key
			 * The first time, we need to create a Vector for the values
			 * If key appears more than once, we append value to Vector
			 * 
			 */
			thisKey = qualName; // save key name
		}

		@Override
		public void startPrefixMapping(String arg0, String arg1)
				throws SAXException {
			// TODO Auto-generated method stub

		}

	}

	/** name of configuration file */
	protected String xmlFile = "wrpmap.xml";
	/** name / value map */
	HashMap<String, Vector<String>> map = new HashMap<String, Vector<String>>();
	
	/**
	 * Default constructor
	 * </p><p>
	 * If this constructor is used, the parseFile(fname)
	 * method should be used to parse the XML file
	 * 
	 */
	public WrpXmlMap()
	{
		
	}
	
	/**
	 * Constructor that specifies the configuration file
	 * </p><p>
	 * If this method is used, the specified XML file
	 * will be parsed.
	 * 
	 * @param configFile
	 * @throws IOException 
	 */
	public WrpXmlMap(String configFile) throws IOException
	{
		setXmlFile(configFile);
		parseFile(); // parse the XML file
	}
	

	/**
	 * This method can be called to parse the XML file
	 * @throws IOException 
	 * 
	 */
	public void parseFile() throws IOException {
		String fname = convertToFileURL(xmlFile);
		parseXmlFile(fname);

	}

	/**
	 * Convert file name to actual File URL
	 * 
	 * @param filename name of file
	 * @return file URL
	 */
	private String convertToFileURL(String filename) {
		String path = new File(filename).getAbsolutePath();
        if (File.separatorChar != '/') {
            path = path.replace(File.separatorChar, '/');
        }
        if (!path.startsWith("/")) {
            path = "/" + path;
        }
        return "file:" + path;

	}

	private HashMap<String, Vector<String>> parseXmlFile(String fname) throws IOException {

			SAXParserFactory factory = SAXParserFactory.newInstance();
			SAXParser saxParser;
			try {
				saxParser = factory.newSAXParser();
				XMLReader xmlReader = saxParser.getXMLReader();
				xmlReader.setContentHandler(new XmlMapContentHandler());
				xmlReader.parse(fname);
			} catch (ParserConfigurationException e) {
				System.err.println("Error with parser configuration while parsing XML Map file");
				System.err.println(e.getMessage());
				e.printStackTrace();
			} catch (SAXException e) {
				System.err.println("SAX Parser exception while parsing XML Map file");
				System.err.println(e.getMessage());
				e.printStackTrace();
				/******************************
			} catch (IOException e) {
				System.err.println("I/O Exception while parsing XML Map file");
				System.err.println(e.getMessage());
				e.printStackTrace();
				********************************/
			}

		
		
		return map;
	}

	public void parseFile(String xmlFile) throws IOException
	{
		setXmlFile(xmlFile);
		parseFile();
	}

	
	/**
	 * @return the configFile
	 */
	public String getXmlFile() {
		return xmlFile;
	}



	/**
	 * @param xmlFile the configFile to set
	 */
	public void setXmlFile(String xmlFile) {
		this.xmlFile = xmlFile;
	}


	/**
	 * Print the map names and values to the print stream
	 * 
	 * @param out PrintStream to use for display
	 */
	public void printValues(PrintStream out) {
		out.println("Name/Value pairs from XMP Map file [" + xmlFile + "]");
		
		Set<String> keys = map.keySet(); // get keys
		Iterator<String> it = keys.iterator(); // get iterator
		while (it.hasNext())
		{ // more keys
			String key = it.next(); // get key
			Vector<String> vals = getValues(key);
			out.print(key + ":");
			for (int i=0; i<vals.size(); i++)
			{
				out.println("\t" + vals.get(i));
			}
		}
	}


	/**
	 * Get the list of values associated with a key.
	 * </p><p>
	 * Each key can have any number of values associated
	 * with it.
	 * This method will return all the values
	 * associated with the key.
	 * If the key isn't present, null will be returned.
	 * 
	 * @param key string under which values are stored
	 * @return list of values or null if not present
	 */
	public Vector<String> getValues(String key) {
		if (!map.containsKey(key)) return null; // no key/value
		Vector<String> vals = map.get(key);
		return vals;
	}
	
	/**
	 * Generate single value associated with key
	 * </p><p>
	 * This method is a convenience function for
	 * keys that should have a single value.
	 * If more than one value is present,
	 * the first is returned.
	 * 
	 * @param key string key
	 * @return single or first value; null if not present
	 */
	public String getValue(String key)
	{
		String ret = null; // assume not there
		if (map.containsKey(key))
		{ // have key - get first value
			Vector<String> vals = map.get(key);
			if (vals.size()>0) ret = vals.get(0);
		}
		return ret;
	}

	/**
	 * @param args
	 */
	public static void main(String[] args) {
		WrpXmlMap xmlmap = new WrpXmlMap();
		String xmlFile = xmlmap.getXmlFile();
		WrpKeyboard kbd = new WrpKeyboard(); // create object
		String s;
		
		try
		{
			System.out.println("XML Map file is [" + xmlFile + "]");
			System.out.print("Read from this file? (y or n):");
			s = kbd.readLine().toLowerCase(); // read response from user
			if (s.startsWith("n"))
			{
				System.out.print("Enter new XML Map file, or ENTER to use default: ");
				s = kbd.readLine();
				if (s.length()>0) xmlmap.setXmlFile(s);
			}
			// parse file and extract map
			xmlmap.setXmlFile(xmlFile);
			xmlmap.parseFile();
			xmlmap.printValues(System.out);
			
		}
		catch (IOException exp)
		{
			System.err.println("I/O Exception [" + exp.getMessage() + "]");
			exp.printStackTrace();
		}
	}

}
