package com.billpringle.utils.wrputils;

import java.io.IOException;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Vector;

/**
 * This class encapsulates a USPS zip code with its associated state and city.
 * It also provides an interface to store zip code information in memory
 * as an array of WrpZipCode objects.
 * </p>
 * 
 * 
  <table border="0" cellspacing="0" cellpadding="6" align="center"
  summary="Creative Commons license">
    <tr>
      <td><!-- Creative Commons License -->
      <a rel="license" href=
      "http://creativecommons.org/licenses/by-nc-sa/2.0/" target=
      "CC"><img alt="Creative Commons License" border="0" src=
      "http://creativecommons.org/images/public/somerights20.gif" /></a></td>

      <td><a rel="license" href=
      "http://creativecommons.org/licenses/by-nc-sa/2.0/" target=
      "CC"><img alt="Creative Commons License Symbols" border="0"
      src="cc_lic.gif" /></a> <!-- /Creative Commons License --></td>

      <td>Unless noted otherwise, all materials available for
      download from my site are copyrighted by Bill Pringle, and
      are licensed under a <a rel="license" href=
      "http://creativecommons.org/licenses/by-nc-sa/2.0/" target=
      "CC">Creative Commons License</a>.</td>
    </tr>
  </table>
  
 * 
 * @author Bill Pringle
 * 
 */
public class WrpZipCode 
{
	/** stored procedure to retrieve zip codes from database */
	private static final String query = "Call GetZipCodes";

	////////////////////////////////////
	//
	// instance variables
	//
	////////////////////////////////////
	/** unique ID from database */
	private int id;
	/** the zipcode */
	private String zipcode;
	/** the state abbreviation */
	private String state;
	/** a list of cities associated with the zipcode */
	private Vector<String> city;
	
	//////////////////////////////
	// Vector for internal memory storage of zip codes
	////////////////////////////
	/**
	 * An array of all zip codes.
	 * The list is initially created as a vector,
	 * but once all the zip codes have been read,
	 * it is converted to an array for speed and lack of overhead.
	 * </p><p>
	 * This list can be built using the
	 * <a href="#loadZipCodes(wrpUtils.WrpDb)">loadZipCodes</a> method.
	 * The list can then be searched using the 
	 * <a href="#findZipCode(java.lang.String)">findZipCode</a> method
	 * 
	 */
	private WrpZipCode[] zipList = null;

	/**
	 * The default Constructor
	 * 
	 */
	public WrpZipCode()
	{
		init(); // initialize instance variables
	}
	
	/**
	 * Constructor with only zipcode
	 * 
	 * @param zipcode the zipcode
	 * 
	 */
	public WrpZipCode(String zipcode)
	{
		init(); // initialize instance variables
		this.zipcode = zipcode; // save zip code
	}
	
	/**
	 * Constructor with state abbreviation
	 * 
	 * @param zipcode the zipcode
	 * @param state the state abbreviation
	 * 
	 */
	public WrpZipCode(String zipcode, String state)
	{
		init(); // initialize instance variables
		this.zipcode = zipcode; // save zip code
		this.state = state; // save state abbreviation
	}
	
	/**
	 * Constructor with initial state and city combination
	 * 
	 * @param zipcode the zipcode
	 * @param state the state abbreviation associated with this zipcode
	 * @param city a city associated with this zipcode
	 * 
	 */
	public WrpZipCode(String zipcode, String state, String city)
	{
		init(); // initialize instance variables
		this.zipcode = zipcode; // save zip code
		this.state = state; // save state abbreviation
		this.city.add(city); // save city name or abbreviation
	}
	
	/**
	 * Constructor to clone a zip code object
	 * 
	 * @param zc the zip code to clone
	 */
	public WrpZipCode(WrpZipCode zc)
	{
		init(); // initialize instance variables
		// if parameter is null, then nothing to clone
		if (zc!=null)
		{ // copy all values
			this.zipcode = zc.getZipcode(); // copy zip code
			this.state = zc.getState(); // copy state abbreviation
			for (int i=0; i<zc.getNumberCities(); i++)
			{ // copy all the city names or abbreviations
				addCity(zc.getCity(i)); // copy city
			}
		}
	}

	/**
	 * Constructor that will load the current list of zipcodes.
	 * 
	 * @param db database connection to zipcodes
	 * 
	 */
	public WrpZipCode(WrpDb db) throws SQLException
	{
		init(); // initialize instance variables
		loadZipCodes(db); // load the vector with all zip codes from database
	}
	
	/**
	 * Compare this zipcode with another one
	 * </p><p>
	 * This method compares the zipcode with the argument zipcode.
	 * The fields tested are the zipcode, the state, and the city.
	 * (The state should always be the same.)
	 * </p><p>
	 * If this zipcode comes before the other, a negative number is returned.
	 * If this zipcode comes after the other, a positive number is returned.
	 * If the two zipcodes are identical, a zero is returned.
	 * </p>
	 * <p>
	 * The method compares the zip code, the state, and any cities.
	 * If a difference is found, then the appropriate value is returned.
	 * If no differences are found (including the two lists of cities),
	 * then zero is returned.
	 * 
	 * @param zip the other zipcode to be compared
	 * @return negative number if the zipcode less than the argument,
	 * 		a positive number if the zipcode is greater than the argument,
	 * 		a zero if the two entries are identical.
	 * 
	 */
	public int compareTo(WrpZipCode zip)
	{
		// primary key is zip code
		int n = zipcode.compareTo(zip.zipcode);
		if (n!=0) return n;
		n = state.compareTo(zip.state);
		if (n!=0) return n; // should never happen
		// compare city names
		for (int i=0; i<city.size(); i++)
		{ // compare each city with the other list
			String s1 = getCity(i); // this city
			String s2 = zip.getCity(i); // city from the parameter
			if (s2==null) return 1; // more cities than the other
			n = s1.compareTo(s2); // compare the two cities
			if (n!=0) return n; // difference found
		}
		/*
		 * If we get here, then either the two lists are identical,
		 * or the zip code that was passed in has a longer list
		 * of cities than this one.
		 * 
		 */
		if (getNumberCities() < zip.getNumberCities()) return -1;
		return 0; // entries identical
	}
	
	/**
	 * Get the corresponding city from the list of cities
	 * 
	 * @param loc index of city
	 * @return the name of the city or null if the index was invalid
	 * 
	 */
	public String getCity(int loc)
	{
		if (loc < 0 || loc >= city.size()) return null; // bad parms
		return city.get(loc); // return the requested city
	}
	
	/**
	 * Determine if the specified city is defined for zip code
	 * 
	 * @param city the city name to test for
	 * @return true if city in list of cities, false otherwise
	 * 
	 */
	public boolean hasCity(String city)
	{
		if (city==null) return false; // bad argument
		for (int i=0; i<getNumberCities(); i++)
		{ // loop thru cities looking for a match
			String s = this.city.get(i); // get this city
			if (s.equals(city)) return true; // match found
		}
		// if we get here, no match was found
		return false; // no match
	}
	
	/**
	 * Get the unique ID assigned by the database for this zip code.
	 * 
	 * @return the id
	 */
	public int getId() {
		return id;
	}

	/**
	 * Set the unique ID for this zip code.
	 * 
	 * @param id the id to set
	 */
	public void setId(int id) {
		this.id = id;
	}

	/**
	 * Get the zip code.
	 * 
	 * @return the zipcode
	 */
	public String getZipcode() {
		return zipcode;
	}

	/**
	 * Set the zip code.
	 * 
	 * @param zipcode the zipcode to set
	 */
	public void setZipcode(String zipcode) {
		this.zipcode = zipcode;
	}

	/**
	 * Get the state.
	 * 
	 * @return the state
	 */
	public String getState() {
		return state;
	}

	/**
	 * Set the state.
	 * 
	 * @param state the state to set
	 */
	public void setState(String state) {
		this.state = state;
	}

	/**
	 * Add another city for this zipcode.
	 * This method checks to determine if the city is already present
	 * to prevent duplicate entries.
	 * 
	 * @param city the city (or abbreviation) to be added
	 */
	public void addCity(String city)
	{
		for (int i=0; i<this.city.size(); i++)
		{ // check for the city already being in the list
			String s = this.city.get(i); // get next city
			if (s.equals(city)) return; // already there, skip add
		}
		// if we get here, the city wasn't in the list
		this.city.add(city); // add this city to the list
	}

	/**
	 * Get the number of possible city names (or abbreviations) for zipcode
	 * 
	 * @return the number of cities or abbreviations that are valid for this zipcode
	 * 
	 */
	public int getNumberCities()
	{
		return this.city.size();
	}

	/**
	 * Get the list of cities assigned to this zipcode
	 * 
	 * @return the current list of cities
	 * 
	 */
	public Vector<String> getCities()
	{
		return city;
	}

	/**
	 * initialize instance variables
	 * 
	 */
	public void init()
	{
		zipcode = null; // no zip code set
		state = null; // no state set
		city = new Vector<String>(); // create empty list for cities
	}
	
	///////////////////////////////////////////////////
	//
	// routines to handle a list of zipcodes
	//
	//////////////////////////////////////////////////
	
	/**
	 * Load the zip codes from a database into an array.
	 * This method uses the <a href="#query">string query</a>
	 * to read the current list of zip codes from the database
	 * and store them into the vector <a href="#zipList">zipList</a>
	 * </p><p>
	 * The query returns a collection of lines with zipcode,
	 * state, and city entries.  There will be as many entries for
	 * a zip code as there are cities defined for that zip code.
	 * Each zip code should only have one state assocateid with it.
	 * 
	 * @param db database connection that contains zipcode definitions
	 * 
	 */
	public void loadZipCodes(WrpDb db) throws SQLException
	{
		Vector<WrpZipCode> lst = new Vector<WrpZipCode>(); // create new list

		ResultSet rs = db.executeQuery(query); // get the zip codes
		// loop through query records and store in list
		while (rs.next())
		{ // get zip code fields
			String strZip = rs.getString("zipcode"); // get zip code
			String strState = rs.getString("state"); // get state
			String strCity = rs.getString("city"); // get this city
			// see if this zip code was already read from database
			WrpZipCode zc = findLoadingZipCode(lst, strZip);
			if (zc == null) 
			{ // first time for this zip code
				zc = new WrpZipCode(strZip, strState, strCity); // new zip code
				lst.add(zc); // add to list
			}
			else
			{ // zipcode already present, just add this city
				zc.addCity(strCity);
			}
		}
		
		// list built - convert to array for speed
		int n = lst.size(); // number of entries
		zipList = new WrpZipCode[n]; // create new array
		zipList = (WrpZipCode[])lst.toArray(zipList); // convert to array
	}
	
	/**
	 * Get array of zipcodes.
	 * This method returns the current list of zip codes that was
	 * read from the database.
	 * If the list wasn't read yet, then it will return null.
	 * </p><p>
	 * This method can be used if the client wants to create some
	 * customized searches or lists for zip codes.
	 * 
	 * @return zipList, which is null if the zipcodes weren't loaded yet
	 * 
	 */
	public WrpZipCode[] getZipList()
	{
		return zipList; // return the list or null if not loaded yet
	}
	
	/**
	 * Search for zipcode entry in vector.
	 * </p><p>
	 * This routine is called when zipcodes are being read
	 * from a database.  Since the entries are sorted by
	 * zipcode, we check the last entry first.  If the zipcode
	 * matches, we return that entry, but if the zipcode is greater
	 * than the one at the end, we return null.
	 * 
	 * @param lst vector containing zipcodes
	 * @param strZipCode string containing the zipcode
	 * @return WrpZipCode object for zipcode, or null if not found
	 * 
	 */
	private WrpZipCode findLoadingZipCode(Vector<WrpZipCode> lst, String strZipCode)
	{
		int loc = lst.size() - 1; // index of last entry
		if (loc < 0) return null; // list empty - not found
		WrpZipCode zc = lst.get(loc); // get entry
		String s = zc.getZipcode(); // get the last zip code that was read so far
		if (s.equals(strZipCode)) return zc; // entry found

		// the zip code didn't match the last one read,
		// a new zip code object should be created
		return null; // should be new zipcode entry
	}
	
	/**
	 * This routine will search for the specified zipcode
	 * </p><p>
	 * This routine will perform a binary search on the array zipList
	 * to find the corresponding zipcode.
	 * </p><p>
	 * This method requires the
	 * <a href="#loadZipCodes(wrpUtils.WrpDb)">loadZipCodes(WrpDb)</a>
	 * method to have been called.
	 * 
	 * @param strZip the desired zipcode
	 * @return the zipcode object if found, or null if not found
	 * 
	 */
	public WrpZipCode findZipCode(String strZip)
	{
		if (zipList==null) return null; // zip codes not read yet
		if (strZip==null) return null; // invalid argument
		
		// zip code should be 5 digit
		if (strZip.length() == 3) strZip = "00" + strZip;
		else if (strZip.length() == 4) strZip = "0" + strZip;
		else if (strZip.length() > 5) strZip = strZip.substring(0, 5);
		if (strZip.length()!=5) return null;
		
		// parameters valid, perform search
		int low = 0; // low limit for search
		int high = zipList.length - 1; // upper limit
		if (high == 0) return null; // empty list
		int mid = (high + low) / 2; // midpoint to check
		// zero in on entry, then switch to sequential search
		while ((high - low) > 50)
		{ // large range - perform binary search
			String tz = zipList[mid].getZipcode(); // middle zipcode
			int n = tz.compareTo(strZip); // compare to desired zipcode
			if (n==0) return zipList[mid]; // entry found
			// entry not found, search before or after middle zip code
			if (n<0)
			{ // desired zipcode is higher
				low = mid; // search above
			}
			if (n>0)
			{ // desired zipcode is lower
				high = mid; // search below
			}
			mid = (high + low) / 2; // new midpoint
		}
		// we are close to the entry - switch to sequential search
		for (int i=low; i<high; i++)
		{ // loop thru segment of list
			String tz = zipList[i].getZipcode(); // get the zip code
			if (tz.equals(strZip)) return zipList[i]; // entry found
		}
		// if we get here, the zip code wasn't found
		return null; // not found
	}
	
	/**
	 * Get the list of cities associated with the specified zip code.
	 * This method searches the list of zip codes for the specified
	 * value.
	 * If found, the list of cities associated with that zip code
	 * is returned.
	 * If the zip code isn't found, then null is returned.
	 * </p><p>
	 * This method requires the
	 * <a href="#loadZipCodes(wrpUtils.WrpDb)">loadZipCodes(WrpDb)</a>
	 * method to have been called.
	 * 
	 * @param strZip the desired zip code
	 * @return a list of cities associated with the zip code,
	 * 			or null if the zip code wasn't found.
	 * 
	 */
	public Vector<String> findCities(String strZip)
	{
		WrpZipCode tz = findZipCode(strZip); // search for the zip code
		if (tz == null) return null; // not found
		return tz.getCities(); // found - return the list of cities
	}

	
	/////////////////////////////////////////////////
	// Public Validation Routines
	/////////////////////////////////////////////////

	/**
	 * Determine if specified string is a valid zipcode.
	 * </p><p>
	 * This method requires the
	 * <a href="#loadZipCodes(wrpUtils.WrpDb)">loadZipCodes(WrpDb)</a>
	 * method to have been called.
	 * 
	 * @param strZip zipcode string to be validated
	 * @return true is zipcode exists, false otherwise
	 * 
	 */
	public boolean isZipCodeValid(String strZip)
	{
		WrpZipCode zc = findZipCode(strZip); // search for zip code
		if (zc == null) return false; // not found - zip code invalid
		return true; // zip code found, it is valid
	}
	
	/**
	 * Determine if a zipcode is valid for a state.
	 * 
	 * @param strZip string with a zipcode
	 * @param strState string with a state abbreviation
	 * @return true if zipcode defined within state; false otherwise
	 * 
	 */
	public boolean isZipCodeValid(String strZip, String strState)
	{
		return isStateValid(strZip, strState); // see if zip code is in the state
	}

	/**
	 * Determine if city / state / zipcode combination valid
	 * </p><p>
	 * This method requires the
	 * <a href="#loadZipCodes(wrpUtils.WrpDb)">loadZipCodes(WrpDb)</a>
	 * method to have been called.
	 * 
	 * @param strZip the zipcode to be validated
	 * @param strState the state abbreviation to be validated
	 * @param strCity the city name to be validated
	 * @return true if zipcode defined for city within state; false otherwise
	 * 
	 */
	public boolean isZipCodeValid(String strZip, String strState, String strCity)
	{
		if (strZip==null||strState==null||strCity==null) return false; // bad parms
		WrpZipCode zc = findZipCode(strZip); // see if found
		if (zc==null) return false; // not found (or list not read yet)
		String state = strState.toUpperCase();
		if (!zc.state.equals(state)) return false; // invalid state
		String city = strCity.toUpperCase();
		for (int i=0; i<zc.city.size(); i++)
		{ // search for city in current list
			String s = zc.city.get(i); // get this city
			if (s.equals(city)) return true; // if match, return true
		}
		// if we get here, the city didn't match
		return false;
	}

	/**
	 * Determine if zipcode is defined within a state.
	 * This method assumes that a given zip code is defined
	 * within a single state (i.e., a zip code doesn't span
	 * more than one state.)
	 * </p><p>
	 * This method requires the
	 * <a href="#loadZipCodes(wrpUtils.WrpDb)">loadZipCodes(WrpDb)</a>
	 * method to have been called.
	 * 
	 * @param strZip the zipcode to be validated
	 * @param strState the state abbreviation to be validated
	 * @return true if zipcode defined within state; false otherwise
	 * 
	 */
	public boolean isStateValid(String strZip, String strState)
	{
		WrpZipCode zc = findZipCode(strZip); // see if zip code in list
		if (zc == null) return false; // invalid zip code
		String s = strState.toUpperCase(); // make sure upper case
		String t = zc.state.toUpperCase(); // should already be upper case
		return t.equals(s); // see if they match
	}

	/**
	 * Determine if city is valid for a zipcode.
	 * </p><p>
	 * This routine only compares the city name,
	 * not the state.
	 * The caller should also validate that the state
	 * is valid for the zipcode.
	 * </p><p>
	 * This method requires the
	 * <a href="#loadZipCodes(wrpUtils.WrpDb)">loadZipCodes(WrpDb)</a>
	 * method to have been called.
	 * 
	 * @param strZip the zipcode to be validated
	 * @param strCity the city to be validated
	 * @return true if zipcode includes the city; false otherwise
	 * 
	 */
	public boolean isCityValid(String strZip, String strCity)
	{
		if (strZip==null || strCity==null) return false; // bad parms
		WrpZipCode zc = findZipCode(strZip); // search for zip code
		if (zc == null) return false; // invalid zip code
		if (zc.city==null) return false; // invalid city
		String city = strCity.toUpperCase();
		for (int i=0; i<zc.city.size(); i++)
		{ // loop thru cities looking for a match
			String s = zc.city.get(i).toUpperCase(); // get this city
			if (s.equals(city)) return true;  // match found
		}
		// if we get here, we didn't find a match
		return false; // no match found - city invalid for zip code
	}

	/**
	 * Test driver for zip codes.
	 * </p><p>
	 * This driver connects to the database, loads the zipcodes,
	 * and then lets the user validate city/state/zipcodes
	 * 
	 * @param args not used
	 */
	public static void main(String[] args)
	{
		UtilDb db=null; // database connection for the database

		// connect to database
		System.out.println("Connecting to database ...");
		try {
			db = new UtilDb(); // connect to database
		} catch (ClassNotFoundException e) {
			System.err.println("Class not found while creating database connection [" + e.getMessage() + "]");
			e.printStackTrace();
		} catch (IllegalAccessException e) {
			System.err.println("Illegal Access Exception while creating database connection [" + e.getMessage() + "]");
			e.printStackTrace();
		} catch (InstantiationException e) {
			System.err.println("Instantiation Exception while creating database connection [" + e.getMessage() + "]");
			e.printStackTrace();
		} catch (SQLException e) {
			System.err.println("SQL Exception while creating database connection [ " + e.getMessage() + "]");
			e.printStackTrace();
		}
		
		// load zipcodes
		System.out.println("Loading zipcodes ...");
		WrpZipCode zc=null;
		try {
			zc = new WrpZipCode(db);
		} catch (SQLException e) {
			System.err.println("SQL Exception while loading zipcodes [" + e.getMessage() + "]");
			e.printStackTrace();
		}
		
		// set up for interactive loop
		WrpKeyboard kb = new WrpKeyboard();

		while (true)
		{ // loop processing commands
			try {
				// display user menu
				System.out.println("Options are:");
				System.out.println("Z Validate zipcode");
				System.out.println("C Validate city/state/zip combinations");
				System.out.println("Q Quit");
				System.out.print("Enter option: ");
				String opt = kb.readLine().toUpperCase();
				if (opt.equals("Z"))
				{
					while (true)
					{ // loop validating zip codes
						System.out.print("Enter zipcode, or ENTER to quit: ");
						String strZip = kb.readLine(); // get zip code
						if (strZip.length()==0) break; // user done
						Vector<String> cities = zc.findCities(strZip);
						if (cities == null)
						{ // no cities - probably zip code was invalid
							System.out.println("Zipcode [" + strZip + "] not found");
						}
						else
						{ // have list of cities - display information
							int n = cities.size(); // number of cities
							System.out.println("Zipcode [" + strZip + "] has " + n + " cities");
							for (int i=0; i<n; i++)
							{ // loop displaying cities
								String s = cities.get(i); // get this city
								System.out.println("\t" + s); // display indented
							}
						}
					}
				}
				else if (opt.equals("C"))
				{ // validate city/state/zip combination
					while (true)
					{ // loop validating city/state/zip combinations
						System.out.print("Enter zipcode, or ENTER to quit: ");
						String zip = kb.readLine(); // get zip code
						if (zip.length() == 0) break; // user done
						System.out.print("Enter State: ");
						String state = kb.readLine().toUpperCase(); // get state
						System.out.print("Enter City: ");
						String city = kb.readLine().toUpperCase(); // get city
						if (zc.isZipCodeValid(zip, state, city))
						{ // combination valid
							System.out.println("Combination is valid");
						}
						else
						{ // combination not valid - see what went wrong
							System.out.println("Combination is not valid");
							WrpZipCode tz = zc.findZipCode(zip);
							if (tz==null)
							{ // zip code is invalid
								System.out.println("Zipcode [" + zip + "] is not defined");
							}
							else if (!zc.isZipCodeValid(zip, state))
							{ // see if zip code defined within the specified state
								String s = tz.state; // get what state it is defined for
								System.out.println("Zipcode [" + zip + "] is is defined for state [" + s + "], not [" + state + "]");
							}
							else
							{ // combination not valid
								System.out.println("Zipcode [" + zip + "] in state [" + state + "], but not city [" + city + "]");
								System.out.print("Valid cities are: ");
								for (int i=0; i<tz.city.size(); i++)
								{ // list the cities on one line
									if (i>0) System.out.print(", "); // separte cities
									if (i==tz.city.size()-1) System.out.print("and "); // cutesy touch
									System.out.print(tz.city.get(i)); // display the city
								}
								System.out.println("."); // end the line
							}
						}
					}
				}
				else if (opt.equals("Q"))
				{ // user quits
					break;
				}
			} catch (IOException e) {
				System.err.println("I/O Exception while reading from keyboard [" + e.getMessage() + "]");
				e.printStackTrace();
			}
		}
	}

}
